package imcacat.imcabc;

import imcacat.epicsca4j.ca.EpicsCaException;
import imcacat.epicsca4j.model.EpicsMotor;

import java.lang.reflect.InvocationTargetException;

import gov.aps.jca.TimeoutException;

/**
 * I automatically optimize Omega, Horizontal Focus, and Vertical Focus.
 */
public class AllAutoOptimization17bm extends BeamFocusAutoOptimization17bm {
  protected static final long OMEGA_MOVE_TIMEOUT = 60000;

  protected EpicsMotor omegaMotor;

  protected double originalOmegaMotorPosition;

  @Override
  protected void initialize()
  throws InterruptedException, EpicsCaException {
    super.initialize();
    omegaMotor = new EpicsMotor(
      Resources.getDefault().getOmegaMotorPvName(), "\u00B5rad");
    originalOmegaMotorPosition = omegaMotor.getPosition();
  }

  @Override
  public void release() {
    if (omegaMotor != null) omegaMotor.release();
    super.release();
  }

  @Override
  protected void go()
  throws InterruptedException, TimeoutException, EpicsCaException {
    setStatus("Configuring beamline for Omega optimization");
    disableAutobp();
    moveMiddleAndDownstreamAndSmallSlitsToDefaultSize();
    setProgress(5);

    setStatus("Scanning Omega");
    scanOmega();
    setProgress(20);

    setStatus("Fitting Omega scan data to Gaussian function");
    GaussianFunctionCurveFitter fitter;
    try {
      fitter = new GaussianFunctionCurveFitter(measuredXYSeries.toArray());
    } catch (Exception e) {
      throw new RuntimeException("Failed to fit measured data to Gaussian " +
        "function", e);
    }
    try {
      setFitDataSeriesInEventDispatchThreadAndWait(fitter);
    } catch (InvocationTargetException e) {
      throw new RuntimeException("Failed to add fit data to " +
        "XYSeriesCollection", e);
    }
    setProgress(25);

    setStatus("Moving Omega to fit maximum");
    moveOmega(fitter.getC());
    setProgress(30);

    setStatus("Allowing beam position correction");
    allowBeamPositionCorrection();
    setProgress(33);

    setStatus("Configuring beamline for Horizontal Focus optimization");
    moveSlitsForHorizontalFocusOptimization();
    setProgress(35);

    setStatus("Scanning Horizontal Focus");
    scanHorizontalFocus();
    setProgress(50);

    setStatus("Moving Horizontal Focus to maximum");
    moveFocus(getXWithMaximumY(measuredXYSeries));
    setProgress(55);

    setStatus("Allowing beam position correction");
    allowBeamPositionCorrection();
    setProgress(66);

    setStatus("Configuring beamline for Vertical Focus optimization");
    moveSlitsForVerticalFocusOptimization();
    setProgress(70);

    setStatus("Scanning Vertical Focus");
    scanVerticalFocus();
    setProgress(80);

    setStatus("Moving Vertical Focus to maximum");
    moveMirrorBend(getXWithMaximumY(measuredXYSeries));
    setProgress(90);

    setStatus("Configuring beamline for use");
    moveMiddleAndDownstreamAndSmallSlitsToDefaultSize();
    enableAutobp();

    setProgressAndStatusToComplete();
  }

  @Override
  public void cleanUp() {
    Exception caughtException = null;

    try {
      scan.stop();
    } catch (Exception e) {
      if (caughtException == null) caughtException = e;
    }
    try {
      omegaMotor.stop();
    } catch (Exception e) {
      if (caughtException == null) caughtException = e;
    }
    try {
      omegaMotor.setDesiredPosition(originalOmegaMotorPosition);
      omegaMotor.moveAndWait(OMEGA_MOVE_TIMEOUT);
    } catch (Exception e) {
      if (caughtException == null) caughtException = e;
    }

    try {
      super.cleanUp();
    } catch (Exception e) {
      if (caughtException == null) caughtException = e;
    }

    if (caughtException != null) throw new RuntimeException(caughtException);
  }

  protected void scanOmega()
  throws EpicsCaException, InterruptedException {
    performOmegaScan();
    if (!scan.completedSuccessfully())
      throw new RuntimeException("Omega scan did not complete successfully");
    try {
      setMeasuredDataSeriesInEventDispatchThreadAndWait(getData(scan));
    } catch (InvocationTargetException e) {
      throw new RuntimeException("Failed to update measured data series " +
        "with complete data from scan", e);
    }
  }

  protected void performOmegaScan() throws EpicsCaException,
  InterruptedException {
    setXYDatasetTitle("Omega optimization scan");
    setXDataLabel("Omega (" + getOmegaUnitOfMeasure() + ")");
    setYDataLabel("BPM intensity (counts)");
    scanResetter.reset();
    scan.setBeforeScanPvLink(
      Resources.getDefault().getBpmIntensityAcquisitionTimePvName());
    scan.setBeforeScanCommandData(1.0f);
    scan.setPositionerPvLink(0,
      Resources.getDefault().getDesiredOmegaPositionPvName());
    scan.setPositionerReadbackPvLink(0,
      Resources.getDefault().getOmegaPositionPvName());
    scan.setPositionerReadbackDelta(0, 2.0);
    scan.setPositionerCenteredLinearMode(0, getOmegaPosition(), 5.0, 20);
    scan.setPositionerSettlingTime(2.0f);
    scan.setDetectorTriggerPvLink(0,
      Resources.getDefault().getBpmIntensityAcquisitionStartPvName());
    scan.setDetectorTriggerCommandData(0, 1.0f);
    scan.setDetectorPvLink(0,
      Resources.getDefault().getBpmIntensityAcquisitionSumPvName());
    scan.go();
    scan.waitToFinish();
  }

  protected double getOmegaPosition()
  throws InterruptedException, EpicsCaException {
    return omegaMotor.getPosition();
  }

  protected String getOmegaUnitOfMeasure() {
    return omegaMotor.getUnitOfMeasure();
  }

  protected void moveOmega(double desiredPosition)
  throws InterruptedException, TimeoutException, EpicsCaException {
    omegaMotor.setDesiredPosition(desiredPosition);
    omegaMotor.moveAndWait(OMEGA_MOVE_TIMEOUT);
  }
}
