package imcacat.imcabc;

import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.Insets;
import java.awt.LayoutManager;

/**
 * I am a layout manager that enforces a fixed aspect ratio for the sizing of
 * a component.
 * <p>
 * The container for which I am the layout manager must contain no more than
 * one component. By default I determine my aspect ratio at instantiation time
 * from the preferred size of the component as returned by its
 * <code>getPreferredSize()</code> method. If I am constructed with a
 * specified aspect ratio, I will use that instead.
 * <p>
 * I may throw a <code>RuntimeException</code> from any of my methods if the
 * container for which I am the layout manager has more than one component.
 */
public class AspectRatioLayout implements LayoutManager {
  protected static final double DEFAULT_ASPECT_RATIO = 1.0;
  protected static final double NULL_ASPECT_RATIO = 0.0;

  protected double aspectRatio;

  /**
   * I create an instance using the preferred size of the component to
   * determine the aspect ratio.
   */
  public AspectRatioLayout() {
    this(AspectRatioLayout.NULL_ASPECT_RATIO);
  }

  /**
   * I create an instance with the specified aspectRatio.
   *
   * @param aspectRatio  must be >= 0.0; if 0.0, a default aspect ratio will
   *                     be used
   * @throws RuntimeExcpetion if aspectRatio is < 0.0
   */
  public AspectRatioLayout(double aspectRatio) {
    super();
    if (aspectRatio < 0.0) throw new RuntimeException("aspectRatio < 0.0");
    this.aspectRatio = aspectRatio;
  }

  public void addLayoutComponent(String name, Component comp) {
  }

  public void layoutContainer(Container parent) {
    Component component = getComponent(parent);
    if (component == null) return;
    Dimension parentDimension = parent.getSize();
    double desiredAspectRatio = getAspectRatio(parent);
    double width = parentDimension.width;
    double height = width / desiredAspectRatio;
    if (height > parentDimension.height) {
      height = parentDimension.height;
      width = desiredAspectRatio * height;
    }
    width = Math.min(parentDimension.width, Math.rint(width));
    height = Math.min(parentDimension.height, Math.rint(height));
    Insets insets = parent.getInsets();
    component.setBounds(insets.left, insets.top, (int)width, (int)height);
  }

  public Dimension minimumLayoutSize(Container parent) {
    Component component = getComponent(parent);
    if (component == null) return createDimension(parent.getInsets());
    return createDimension(parent.getInsets(), component.getMinimumSize());
  }

  public Dimension preferredLayoutSize(Container parent) {
    Component component = getComponent(parent);
    if (component == null) return createDimension(parent.getInsets());
    return createDimension(parent.getInsets(), component.getPreferredSize());
  }

  public void removeLayoutComponent(Component comp) {
  }

  protected double getAspectRatio(Container parent) {
    if (aspectRatio != AspectRatioLayout.NULL_ASPECT_RATIO)
      return aspectRatio;
    Component component = getComponent(parent);
    if (component == null) return AspectRatioLayout.DEFAULT_ASPECT_RATIO;
    Dimension dimension = component.getPreferredSize();
    aspectRatio = (double)dimension.width / (double)dimension.height;
    return aspectRatio;
  }

  protected Component getComponent(Container parent) {
    Component[] components = parent.getComponents();
    if (components.length == 0) return null;
    if (components.length == 1) return components[0];
    throw new RuntimeException("Container has more than one component");
  }

  protected Dimension createDimension(Insets insets) {
    return new Dimension(insets.left + insets.right,
      insets.top + insets.bottom);
  }

  protected Dimension createDimension(Insets insets, Dimension dimension) {
    return new Dimension(insets.left + insets.right + dimension.width,
      insets.top + insets.bottom + dimension.height);
  }
}
