package imcacat.imcabc;

import java.util.HashMap;
import java.util.Map;

import imcacat.epicsca4j.bean.OneWayPropertyChangeMediator;
import imcacat.epicsca4j.ca.EpicsCaException;
import imcacat.epicsca4j.convert.DoubleToLongEpicsCaValueConverter;
import imcacat.epicsca4j.model.EpicsPv;
import imcacat.epicsca4j.model.TpmacAssemblyMotionCommand;

/**
 * I am a <code>BeamFacade</code> for the 17-ID beamline.
 */
public class BeamFacade17id extends BeamFacade {
  private static final long serialVersionUID = 1;

  public static final String PROPNAME_HARMONIC =
    "harmonic";
  public static final String PROPNAME_GAP =
    "gap";
  public static final String PROPNAME_ENERGY_MOVE_MODE =
    "energyMoveMode";
  public static final String PROPNAME_ENERGY_MOTION_COMMAND =
    "energyMotionCommand";

  public static final String PROPNAME_APERTURE =
    "aperture";
  public static final String PROPNAME_APERTURE_RBV =
    "apertureRbv";
  public static final String PROPNAME_APERTURE_SIZE =
    "apertureSize";
  public static final String PROPNAME_APERTURE_MOVING =
    "apertureMoving";
  public static final String PROPNAME_APERTURE_MOTION_COMMAND =
    "apertureMotionCommand";

  public static final String PROPNAME_MIRROR_BPM_INTENSITY =
    "mirrorBpmIntensity";
  public static final String PROPNAME_SAMPLE_BPM_INTENSITY =
    "sampleBpmIntensity";
  public static final String PROPNAME_ION_CHAMBER_INTENSITY =
    "ionChamberIntensity";
  public static final String PROPNAME_BEAM_STOP_INTENSITY =
    "beamStopIntensity";

  public static final String PROPNAME_ATTENUATION_PERCENTAGE =
    "attenuationPercentage";

  protected EpicsPv harmonic;
  protected OneWayPropertyChangeMediator harmonicMediator;

  protected EpicsPv gap;
  protected OneWayPropertyChangeMediator gapMediator;

  protected TpmacEnergyAxis energyAxis;
  protected OneWayPropertyChangeMediator energyAxisMediator;

  protected Minibeam minibeam;
  protected OneWayPropertyChangeMediator minibeamMediator;

  protected EpicsPv mirrorBpmIntensity;
  protected OneWayPropertyChangeMediator mirrorBpmIntensityMediator;

  protected EpicsPv sampleBpmIntensity;
  protected OneWayPropertyChangeMediator sampleBpmIntensityMediator;

  protected EpicsPv ionChamberIntensity;
  protected OneWayPropertyChangeMediator ionChamberIntensityMediator;

  protected EpicsPv beamStopIntensity;
  protected OneWayPropertyChangeMediator beamStopIntensityMediator;

  protected ImcaFilterRack filterRack;
  protected OneWayPropertyChangeMediator filterRackMediator;

  public BeamFacade17id() throws InterruptedException, EpicsCaException {
    super();
  }

  @Override
  protected void initialize()
  throws InterruptedException, EpicsCaException {
    super.initialize();
    initializeEnergySection();
    initializeBeamDefiningApertureSection();
    initializeIntensitySection();
    initializeAttenuationSection();
  }

  protected void initializeEnergySection()
  throws InterruptedException, EpicsCaException {
    harmonic = new EpicsPv(Resources.getDefault().getHarmonicPvName(),
      false, true, new DoubleToShortEpicsCaValueConverter());
    harmonicMediator = new OneWayPropertyChangeMediator(harmonic, this,
      EpicsPv.PROPNAME_VALUE, BeamFacade17id.PROPNAME_HARMONIC);
    gap = new EpicsPv(Resources.getDefault().getGapPvName(),
      false, true);
    gapMediator = new OneWayPropertyChangeMediator(gap, this,
      EpicsPv.PROPNAME_VALUE, BeamFacade17id.PROPNAME_GAP);

    energyAxis = new TpmacEnergyAxis(
      Resources.getDefault().getEnergyAssemblyPvName(),
      Resources.getDefault().getEnergyAxisPvName(), 0.001);
    Map<String,String> energyAxisMediatorPropertyMap =
      new HashMap<String,String>();
    energyAxisMediatorPropertyMap.put(
      TpmacEnergyAxis.PROPNAME_POSITION, PROPNAME_ENERGY_POSITION);
    energyAxisMediatorPropertyMap.put(
      TpmacEnergyAxis.PROPNAME_WAVELENGTH_POSITION,
      PROPNAME_WAVELENGTH_POSITION);
    energyAxisMediatorPropertyMap.put(
      TpmacEnergyAxis.PROPNAME_MOVING, PROPNAME_ENERGY_MOVING);
    energyAxisMediatorPropertyMap.put(
      TpmacEnergyAxis.PROPNAME_DESIRED_POSITION,
      PROPNAME_ENERGY_DESIRED_POSITION);
    energyAxisMediatorPropertyMap.put(
      TpmacEnergyAxis.PROPNAME_WAVELENGTH_DESIRED_POSITION,
      PROPNAME_WAVELENGTH_DESIRED_POSITION);
    energyAxisMediator = new OneWayPropertyChangeMediator(energyAxis, this,
      energyAxisMediatorPropertyMap);
  }

  protected void initializeBeamDefiningApertureSection()
  throws InterruptedException, EpicsCaException {
    minibeam = new Minibeam(
      Resources.getDefault().getMinibeamBasePvName());
    Map<String,String> sourceToTargetPropMap = new HashMap<String,String>();
    sourceToTargetPropMap.put(PROPNAME_APERTURE,
      Minibeam.PROPNAME_APERTURE);
    sourceToTargetPropMap.put(PROPNAME_APERTURE_RBV,
      Minibeam.PROPNAME_APERTURE_RBV);
    sourceToTargetPropMap.put(PROPNAME_APERTURE_SIZE,
      Minibeam.PROPNAME_APERTURE_SIZE);
    sourceToTargetPropMap.put(PROPNAME_APERTURE_MOVING,
      Minibeam.PROPNAME_APERTURE_MOVING);
    Map<String,Map<Integer,Integer>> sourceToTargetIndexMapMap =
      new HashMap<String,Map<Integer,Integer>>();
    sourceToTargetIndexMapMap.put(PROPNAME_APERTURE, null);
    sourceToTargetIndexMapMap.put(PROPNAME_APERTURE_RBV, null);
    Map<Integer,Integer> indexMap = new HashMap<Integer,Integer>();
    for (int i = 0; i < Minibeam.NUM_APERTURES; i++) indexMap.put(i, i);
    sourceToTargetIndexMapMap.put(PROPNAME_APERTURE_SIZE, indexMap);
    sourceToTargetIndexMapMap.put(PROPNAME_APERTURE_MOVING, null);
    minibeamMediator = new OneWayPropertyChangeMediator(minibeam, this,
      sourceToTargetPropMap, sourceToTargetIndexMapMap);
  }

  protected void initializeIntensitySection()
  throws InterruptedException, EpicsCaException {
    mirrorBpmIntensity = new EpicsPv(
      Resources.getDefault().getMirrorBpmIntensityPvName(), false, true,
      new DoubleToLongEpicsCaValueConverter());
    mirrorBpmIntensityMediator = new OneWayPropertyChangeMediator(
      mirrorBpmIntensity, this, EpicsPv.PROPNAME_VALUE,
      BeamFacade17id.PROPNAME_MIRROR_BPM_INTENSITY);

    sampleBpmIntensity = new EpicsPv(
      Resources.getDefault().getSampleBpmIntensityPvName(), false, true,
      new DoubleToLongEpicsCaValueConverter());
    sampleBpmIntensityMediator = new OneWayPropertyChangeMediator(
      sampleBpmIntensity, this, EpicsPv.PROPNAME_VALUE,
      BeamFacade17id.PROPNAME_SAMPLE_BPM_INTENSITY);

    ionChamberIntensity = new EpicsPv(
      Resources.getDefault().getIonChamberIntensityPvName(), false, true,
      new DoubleToLongEpicsCaValueConverter());
    ionChamberIntensityMediator = new OneWayPropertyChangeMediator(
      ionChamberIntensity, this, EpicsPv.PROPNAME_VALUE,
      BeamFacade17id.PROPNAME_ION_CHAMBER_INTENSITY);

    beamStopIntensity = new EpicsPv(
      Resources.getDefault().getBeamStopIntensityPvName(), false, true,
      new DoubleToLongEpicsCaValueConverter());
    beamStopIntensityMediator = new OneWayPropertyChangeMediator(
      beamStopIntensity, this, EpicsPv.PROPNAME_VALUE,
      BeamFacade17id.PROPNAME_BEAM_STOP_INTENSITY);
  }

  protected void initializeAttenuationSection()
  throws InterruptedException, EpicsCaException {
    filterRack = new ImcaFilterRack(
      Resources.getDefault().getFilterRackBasePvName());
    filterRackMediator = new OneWayPropertyChangeMediator(filterRack, this,
      ImcaFilterRack.PROPNAME_ATTENUATION,
      PROPNAME_ATTENUATION_PERCENTAGE);
  }

  public short getHarmonic() throws InterruptedException, EpicsCaException {
    return (Short)harmonic.getValue();
  }

  public double getGap() throws InterruptedException, EpicsCaException {
    return (Double)gap.getValue();
  }

  public double getEnergyPosition() {
    return energyAxis.getPosition();
  }

  public double getWavelengthPosition() {
    return energyAxis.getWavelengthPosition();
  }

  public boolean isEnergyMoving()
  throws InterruptedException, EpicsCaException {
    return energyAxis.isMoving();
  }

  public double getEnergyDesiredPosition() {
    return energyAxis.getDesiredPosition();
  }

  public void setEnergyDesiredPosition(double newValue) {
    energyAxis.setDesiredPosition(newValue);
  }

  public double getWavelengthDesiredPosition() {
    return energyAxis.getWavelengthDesiredPosition();
  }

  public void setWavelengthDesiredPosition(double newValue) {
    energyAxis.setWavelengthDesiredPosition(newValue);
  }

  public void setEnergyMotionCommand(TpmacAssemblyMotionCommand newValue)
  throws InterruptedException, EpicsCaException {
    energyAxis.setMotionCommand(newValue);
  }

  public MinibeamAperture getAperture()
  throws InterruptedException, EpicsCaException {
    return minibeam.getAperture();
  }

  public void setAperture(MinibeamAperture newAperture)
  throws InterruptedException, EpicsCaException {
    minibeam.setAperture(newAperture);
  }

  public MinibeamApertureRbv getApertureRbv()
  throws InterruptedException, EpicsCaException {
    return minibeam.getApertureRbv();
  }

  public double getApertureSize(int index)
  throws InterruptedException, EpicsCaException {
    return minibeam.getApertureSize(index);
  }

  public boolean isApertureMoving() {
    return minibeam.isApertureMoving();
  }

  public void setApertureMotionCommand(MinibeamApertureMotionCommand command)
  throws InterruptedException, EpicsCaException {
    minibeam.setApertureMotionCommand(command);
  }

  public long getMirrorBpmIntensity()
  throws InterruptedException, EpicsCaException {
    return (Long)mirrorBpmIntensity.getValue();
  }

  public long getSampleBpmIntensity()
  throws InterruptedException, EpicsCaException {
    return (Long)sampleBpmIntensity.getValue();
  }

  public long getIonChamberIntensity()
  throws InterruptedException, EpicsCaException {
    return (Long)ionChamberIntensity.getValue();
  }

  public long getBeamStopIntensity()
  throws InterruptedException, EpicsCaException {
    return (Long)beamStopIntensity.getValue();
  }

  public double getAttenuationPercentage()
  throws InterruptedException, EpicsCaException {
    return filterRack.getAttenuation();
  }

  @Override
  public ImcaFilterRack getFilterRackModel() {
    return filterRack;
  }
}
