package imcacat.imcabc;

import imcacat.epicsca4j.ca.EpicsCaException;
import imcacat.epicsca4j.model.EpicsMotor;

import java.lang.reflect.InvocationTargetException;

import gov.aps.jca.TimeoutException;

/**
 * I automatically optimize Horizontal Focus and Vertical Focus.
 */
public class BeamFocusAutoOptimization17bm extends Optimization17bm {
  protected static final long FOCUS_MOVE_TIMEOUT = 60000;
  protected static final long MIRROR_BEND_MOVE_TIMEOUT = 60000;

  protected EpicsMotor focusMotor;
  protected EpicsMotor mirrorBendMotor;
  protected EpicsMotor energyMotor;

  protected double originalFocusMotorPosition;
  protected double originalMirrorBendMotorPosition;

  public BeamFocusAutoOptimization17bm() {
    super();
  }

  @Override
  protected void initialize()
  throws InterruptedException, EpicsCaException {
    super.initialize();
    initializeForHorizontalOptimization();
    initializeForVerticalOptimization();
  }

  protected void initializeForHorizontalOptimization()
  throws InterruptedException, EpicsCaException {
    focusMotor = new EpicsMotor(
      Resources.getDefault().getFocusMotorPvName(), "\u03BCm");
    originalFocusMotorPosition = focusMotor.getPosition();
    energyMotor = new EpicsMotor(
      Resources.getDefault().getEnergyMotorPvName(), "eV");
  }

  protected void initializeForVerticalOptimization()
  throws InterruptedException, EpicsCaException {
    mirrorBendMotor = new EpicsMotor(
      Resources.getDefault().getMirrorBendMotorPvName(), "\u03BCm");
    originalMirrorBendMotorPosition = mirrorBendMotor.getPosition();
  }

  @Override
  public void release() {
    if (focusMotor != null) focusMotor.release();
    if (mirrorBendMotor != null) mirrorBendMotor.release();
    if (energyMotor != null) energyMotor.release();
    super.release();
  }

  @Override
  protected void go()
  throws InterruptedException, TimeoutException, EpicsCaException {
    setStatus("Configuring beamline for Horizontal Focus optimization");
    disableAutobp();
    moveSlitsForHorizontalFocusOptimization();
    setProgress(10);

    setStatus("Scanning Horizontal Focus");
    scanHorizontalFocus();
    setProgress(30);

    setStatus("Moving Horizontal Focus to maximum");
    moveFocus(getXWithMaximumY(measuredXYSeries));
    setProgress(40);

    setStatus("Allowing beam position correction");
    allowBeamPositionCorrection();
    setProgress(50);

    setStatus("Configuring beamline for Vertical Focus optimization");
    moveSlitsForVerticalFocusOptimization();
    setProgress(55);

    setStatus("Scanning Vertical Focus");
    scanVerticalFocus();
    setProgress(70);

    setStatus("Moving Vertical Focus to maximum");
    moveMirrorBend(getXWithMaximumY(measuredXYSeries));
    setProgress(80);

    setStatus("Configuring beamline for use");
    moveMiddleAndDownstreamAndSmallSlitsToDefaultSize();
    enableAutobp();

    setProgressAndStatusToComplete();
  }

  protected void moveSlitsForVerticalFocusOptimization()
      throws InterruptedException, TimeoutException, EpicsCaException {
    moveMiddleSlits(DEFAULT_MIDDLE_SLITS_WIDTH, DEFAULT_MIDDLE_SLITS_HEIGHT);
    moveDownstreamSlits(1000.0, 300.0);
    moveSmallSlit(600.0);
  }

  protected void moveSlitsForHorizontalFocusOptimization()
      throws InterruptedException, TimeoutException, EpicsCaException {
    moveMiddleSlits(DEFAULT_MIDDLE_SLITS_WIDTH, DEFAULT_MIDDLE_SLITS_HEIGHT);
    moveDownstreamSlits(1000.0, 1000.0);
    moveSmallSlit(300.0);
  }

  protected void scanVerticalFocus()
  throws EpicsCaException, InterruptedException {
    performMirrorBendScan();
    if (!scan.completedSuccessfully())
      throw new RuntimeException("Vertical Focus scan did not complete " +
        "successfully");
    try {
      setMeasuredDataSeriesInEventDispatchThreadAndWait(getData(scan));
    } catch (InvocationTargetException e) {
      throw new RuntimeException("Failed to update measured data series " +
        "with complete data from scan", e);
    }
  }

  protected void scanHorizontalFocus()
  throws EpicsCaException, InterruptedException {
    performFocusScan();
    if (!scan.completedSuccessfully())
      throw new RuntimeException("Horizontal Focus scan did not complete " +
        "successfully");
    try {
      setMeasuredDataSeriesInEventDispatchThreadAndWait(getData(scan));
    } catch (InvocationTargetException e) {
      throw new RuntimeException("Failed to update measured data series " +
        "with complete data from scan", e);
    }
  }

  @Override
  public void cleanUp() {
    Exception caughtException = null;

    try {
      scan.stop();
    } catch (Exception e) {
      if (caughtException == null) caughtException = e;
    }

    if (focusMotor != null) {
      try {
        focusMotor.stop();
      } catch (Exception e) {
        if (caughtException == null) caughtException = e;
      }
    }

    if (mirrorBendMotor != null) {
      try {
        mirrorBendMotor.stop();
      } catch (Exception e) {
        if (caughtException == null) caughtException = e;
      }
    }

    try {
      moveMiddleAndDownstreamAndSmallSlitsToDefaultSize();
    } catch (Exception e) {
      if (caughtException == null) caughtException = e;
    }

    if (focusMotor != null) {
      try {
        focusMotor.setDesiredPosition(originalFocusMotorPosition);
        focusMotor.moveAndWait(FOCUS_MOVE_TIMEOUT);
      } catch (Exception e) {
        if (caughtException == null) caughtException = e;
      }
    }

    if (mirrorBendMotor != null) {
      try {
        mirrorBendMotor.setDesiredPosition(originalMirrorBendMotorPosition);
        mirrorBendMotor.moveAndWait(MIRROR_BEND_MOVE_TIMEOUT);
      } catch (Exception e) {
        if (caughtException == null) caughtException = e;
      }
    }

    try {
      enableAutobp();
    } catch (Exception e) {
      if (caughtException == null) caughtException = e;
    }

    try {
      super.cleanUp();
    } catch (Exception e) {
      if (caughtException == null) caughtException = e;
    }

    if (caughtException != null) throw new RuntimeException(caughtException);
  }

  protected void performFocusScan() throws EpicsCaException,
  InterruptedException {
    setXYDatasetTitle("Horizontal Focus optimization scan");
    setXDataLabel("Horizontal Focus (" + focusMotor.getUnitOfMeasure() + ")");
    setYDataLabel("Small Ion Chamber intensity (counts)");
    scanResetter.reset();
    scan.setBeforeScanPvLink(
      Resources.getDefault().
        getSmallIonChamberIntensityAcquisitionTimePvName());
    scan.setBeforeScanCommandData(1.0f);
    scan.setPositionerPvLink(0,
      Resources.getDefault().getDesiredFocusPositionPvName());
    scan.setPositionerReadbackPvLink(0,
      Resources.getDefault().getFocusPositionPvName());
    scan.setPositionerReadbackDelta(0, 5.0);
    scan.setPositionerCenteredLinearMode(0, getFocusPosition(), -5.0, 20);
    scan.setPositionerSettlingTime(1.0f);
    scan.setDetectorTriggerPvLink(0,
      Resources.getDefault().
        getSmallIonChamberIntensityAcquisitionStartPvName());
    scan.setDetectorTriggerCommandData(0, 1.0f);
    scan.setDetectorPvLink(0,
      Resources.getDefault().
        getSmallIonChamberIntensityAcquisitionSumPvName());
    openExposureShutter();
    scan.go();
    scan.waitToFinish();
    closeExposureShutter();
  }

  protected void performMirrorBendScan() throws EpicsCaException,
  InterruptedException {
    setXYDatasetTitle("Vertical Focus optimization scan");
    setXDataLabel("Vertical Focus (" + mirrorBendMotor.getUnitOfMeasure() +
      ")");
    setYDataLabel("Small Ion Chamber intensity (counts)");
    scanResetter.reset();
    scan.setBeforeScanPvLink(
      Resources.getDefault().
        getSmallIonChamberIntensityAcquisitionTimePvName());
    scan.setBeforeScanCommandData(1.0f);
    scan.setPositionerPvLink(0,
      Resources.getDefault().getDesiredMirrorBendPositionPvName());
    scan.setPositionerReadbackPvLink(0,
      Resources.getDefault().getMirrorBendPositionPvName());
    scan.setPositionerReadbackDelta(0, 10.0);
    scan.setPositionerCenteredLinearMode(0, getMirrorBendPosition(), 10.0,
      18);
    scan.setPositionerSettlingTime(1.0f);
    scan.setDetectorTriggerPvLink(0,
      Resources.getDefault().
        getSmallIonChamberIntensityAcquisitionStartPvName());
    scan.setDetectorTriggerCommandData(0, 1.0f);
    scan.setDetectorPvLink(0,
      Resources.getDefault().
        getSmallIonChamberIntensityAcquisitionSumPvName());
    openExposureShutter();
    scan.go();
    scan.waitToFinish();
    closeExposureShutter();
  }

  protected double getFocusPosition()
  throws InterruptedException, EpicsCaException {
    return focusMotor.getPosition();
  }

  protected void moveFocus(double desiredPosition)
  throws InterruptedException, TimeoutException, EpicsCaException {
    focusMotor.setDesiredPosition(desiredPosition);
    focusMotor.moveAndWait(FOCUS_MOVE_TIMEOUT);
  }

  protected double getMirrorBendPosition()
  throws InterruptedException, EpicsCaException {
    return mirrorBendMotor.getPosition();
  }

  protected void moveMirrorBend(double desiredPosition)
  throws InterruptedException, TimeoutException, EpicsCaException {
    mirrorBendMotor.setDesiredPosition(desiredPosition);
    mirrorBendMotor.moveAndWait(MIRROR_BEND_MOVE_TIMEOUT);
  }

  protected double getEnergyPosition()
  throws InterruptedException, EpicsCaException {
    return energyMotor.getPosition();
  }
}
