package imcacat.imcabc;

import imcacat.epicsca4j.ca.EpicsCaException;
import imcacat.epicsca4j.util.StringUtilities;
import imcacat.epicsca4j.util.UiUtilities;

import java.awt.BorderLayout;
import java.awt.Frame;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;

import javax.swing.JComponent;
import javax.swing.JDialog;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JSeparator;
import javax.swing.SwingConstants;
import javax.swing.SwingUtilities;
import javax.swing.border.EmptyBorder;

import com.jgoodies.forms.builder.PanelBuilder;
import com.jgoodies.forms.factories.Borders;
import com.jgoodies.forms.layout.CellConstraints;
import com.jgoodies.forms.layout.FormLayout;

/**
 * I am the Beam tab view.
 */
public abstract class BeamPanel<M extends BeamPresentationModel<? extends
BeamFacade,? extends FilterRackPresentationModel<?>>>
extends JPanel {
  private static final long serialVersionUID = 1;

  protected static final long ENERGY_MOTION_START_TIMEOUT = 3000;

  protected M model;

  protected JDialog filterRackDialog;

  protected boolean energyMotionStartCommanded = false;
  protected long energyMotionStartCommandedTime = -1;
  protected double energyMoveDesiredPosition;
  protected double energyMoveActualPosition;
  protected double energyMoveDeadband =
    Resources.getDefault().getEnergyDeadband();

  public BeamPanel(M model)
  throws InterruptedException, EpicsCaException {
    super();
    this.model = model;
    setLayout(new BorderLayout());
    add(buildMainPanel(), BorderLayout.CENTER);
  }

  protected JComponent buildMainPanel()
  throws InterruptedException, EpicsCaException {
    FormLayout layout = new FormLayout(
      "pref, pref, fill:pref:grow",
      "fill:pref:grow");
    JPanel panel = new JPanel(layout);
    CellConstraints cc = new CellConstraints();
    panel.add(buildLeftPanel(), cc.xy(1, 1));
    panel.add(buildLeftRightSeparator(), cc.xy(2,1));
    panel.add(buildRightPanel(), cc.xy(3, 1));
    return panel;
  }

  protected abstract JComponent buildLeftPanel();

  protected JComponent buildRightPanel()
  throws InterruptedException, EpicsCaException {
    FormLayout layout = new FormLayout(
      "fill:pref:grow",
      "pref, 4dlu, fill:pref:grow"
    );
    PanelBuilder builder = new PanelBuilder(layout);
    builder.border(Borders.DIALOG);
    CellConstraints cc = new CellConstraints();

    builder.addSeparator("Beam position monitor", cc.xy(1, 1));

    JPanel panel = new JPanel(new AspectRatioLayout());
    panel.add(new BpmChartPanel());
    builder.add(panel, cc.xy(1, 3));

    return builder.getPanel();
  }

  protected JComponent buildLeftRightSeparator() {
    JPanel panel = new JPanel(new BorderLayout());
    panel.add(new JSeparator(SwingConstants.VERTICAL), BorderLayout.CENTER);
    panel.setBorder(new EmptyBorder(10, 0, 10, 0));
    return panel;
  }

  protected void changeAttenuationActionPerformed() {
    if (filterRackDialog == null) {
      Frame rootFrame = UiUtilities.getRootFrame();
      filterRackDialog = createFilterRackDialog(rootFrame);
      filterRackDialog.pack();
      filterRackDialog.setLocationRelativeTo(rootFrame);
      filterRackDialog.setResizable(false);
    }
    filterRackDialog.setVisible(true);
  }

  protected abstract JDialog createFilterRackDialog(Frame root);

  protected void registerInterestInCommandedEnergyMove(TwoStateButtonAction
      moveStopAction, final String moveCommand) {
    moveStopAction.addActionListener(
      new ActionListener() {
        public void actionPerformed(ActionEvent event) {
          energyMotionCommanded(moveCommand.equals(event.getActionCommand()));
        }
      });

    model.getModel(BeamFacade.PROPNAME_ENERGY_MOVING).
      addValueChangeListener(new PropertyChangeListener() {
        public void propertyChange(PropertyChangeEvent event) {
          energyMovingChanged((Boolean)event.getNewValue());
        }
      });
  }

  protected void energyMotionCommanded(boolean start) {
    if (start) {
      energyMotionStartCommanded = true;
      energyMotionStartCommandedTime = System.nanoTime();
      energyMoveDesiredPosition = (Double)model.getModel(
        BeamFacade.PROPNAME_ENERGY_DESIRED_POSITION).getValue();
    } else {
      energyMotionStartCommanded = false;
    }
  }

  protected void energyMovingChanged(Boolean moving) {
    if (moving == null || moving.booleanValue()) {
      if ((System.nanoTime() - energyMotionStartCommandedTime) >
          millisecondsToNanoseconds(ENERGY_MOTION_START_TIMEOUT))
        energyMotionStartCommanded = false;
      return;
    }
    if (!energyMotionStartCommanded) return;

    energyMotionStartCommanded = false;
    energyMoveActualPosition = (Double)model.getModel(
      BeamFacade.PROPNAME_ENERGY_POSITION).getValue();

    if (Math.abs(energyMoveDesiredPosition - energyMoveActualPosition) <=
        energyMoveDeadband)
      return;

    SwingUtilities.invokeLater(new Runnable() {
      public void run() {
        JOptionPane.showMessageDialog(BeamPanel.this,
          "The energy move did not succeed.\n\n" +
          StringUtilities.wordWrap(
            "The desired energy was " +
            String.format("%.3f", energyMoveDesiredPosition) +
            " eV, but the actual energy is " +
            String.format("%.3f", energyMoveActualPosition) +
            " eV.  The difference exceeds the energy move positioning " +
            "tolerance.  Check the control system to ensure energy can be " +
            "moved successfully.", 78, "\n"), null,
          JOptionPane.WARNING_MESSAGE);
      }
    });
  }

  protected static long millisecondsToNanoseconds(long milliseconds) {
    return milliseconds * 1000000L;
  }
}
