package imcacat.imcabc;

import imcacat.epicsca4j.ca.EpicsCaException;
import imcacat.epicsca4j.model.EpicsMotor;

import java.lang.reflect.InvocationTargetException;

import gov.aps.jca.TimeoutException;

/**
 * I automatically optimize the BPM position.
 */
public class BpmAutoOptimization17bm extends Optimization17bm {
  protected static final long HORIZONTAL_BPM_MOVE_TIMEOUT = 60000;
  protected static final long VERTICAL_BPM_MOVE_TIMEOUT = 60000;

  protected EpicsMotor horizontalBpmMotor;
  protected EpicsMotor verticalBpmMotor;

  protected double originalHorizontalBpmMotorPosition;
  protected double originalVerticalBpmMotorPosition;
  protected double originalDownstreamHorizontalCenter;
  protected double originalDownstreamVerticalCenter;

  public BpmAutoOptimization17bm() {
    super();
  }

  @Override
  protected void initialize()
  throws InterruptedException, EpicsCaException {
    super.initialize();
    horizontalBpmMotor = new EpicsMotor(
      Resources.getDefault().getBpmHorizontalMotorPvName(), "\u03BCm");
    verticalBpmMotor = new EpicsMotor(
      Resources.getDefault().getBpmVerticalMotorPvName(), "\u03BCm");
    originalHorizontalBpmMotorPosition = horizontalBpmMotor.getPosition();
    originalVerticalBpmMotorPosition = verticalBpmMotor.getPosition();
    originalDownstreamHorizontalCenter = getDownstreamHorizontalSlitCenter();
    originalDownstreamVerticalCenter = getDownstreamVerticalSlitCenter();
  }

  @Override
  public void release() {
    if (horizontalBpmMotor != null) horizontalBpmMotor.release();
    if (verticalBpmMotor != null) verticalBpmMotor.release();
    super.release();
  }

  @Override
  protected void go()
  throws InterruptedException, TimeoutException, EpicsCaException {
    setStatus("Configuring beamline for horizontal optimization");
    disableAutobp();
    moveMiddleSlits(DEFAULT_MIDDLE_SLITS_WIDTH, DEFAULT_MIDDLE_SLITS_HEIGHT);
    moveDownstreamSlits(50.0, 1000.0);
    moveSmallSlit(800.0);
    setProgress(10);

    setStatus("Scanning Downstream Horizontal Slit");
    scanDownstreamHorizontalSlitCenter();
    setProgress(35);

    setStatus("Fitting horizontal scan data to Gaussian function");
    GaussianFunctionCurveFitter fitter;
    try {
      fitter = new GaussianFunctionCurveFitter(measuredXYSeries.toArray());
    } catch (Exception e) {
      throw new RuntimeException("Failed to fit measured data to Gaussian " +
        "function", e);
    }
    try {
      setFitDataSeriesInEventDispatchThreadAndWait(fitter);
    } catch (InvocationTargetException e) {
      throw new RuntimeException("Failed to add fit data to " +
        "XYSeriesCollection", e);
    }
    setProgress(40);

    setStatus("Moving BPM Horizontal to optimal position");
    moveHorizontalBpmRelative(
      -1.0 * (fitter.getC() - originalDownstreamHorizontalCenter));
    setProgress(45);

    setStatus("Moving DS Horizontal Slit center to original position");
    moveDownstreamHorizontalSlitCenter(originalDownstreamHorizontalCenter);
    setProgress(50);

    setStatus("Allowing beam position correction");
    allowBeamPositionCorrection();
    setProgress(55);

    setStatus("Configuring beamline for vertical optimization");
    moveDownstreamSlits(1000.0, 50.0);
    setProgress(60);

    setStatus("Scanning Downstream Vertical Slit");
    scanDownstreamVerticalSlitCenter();
    setProgress(80);

    setStatus("Fitting vertical scan data to Gaussian function");
    try {
      fitter = new GaussianFunctionCurveFitter(measuredXYSeries.toArray());
    } catch (Exception e) {
      throw new RuntimeException("Failed to fit measured data to Gaussian " +
        "function", e);
    }
    try {
      setFitDataSeriesInEventDispatchThreadAndWait(fitter);
    } catch (InvocationTargetException e) {
      throw new RuntimeException("Failed to add fit data to " +
        "XYSeriesCollection", e);
    }
    setProgress(85);

    setStatus("Moving BPM Vertical to optimal position");
    moveVerticalBpmRelative(
      -1.0 * (fitter.getC() - originalDownstreamVerticalCenter));
    setProgress(90);

    setStatus("Moving DS Vertical Slit center to original position");
    moveDownstreamVerticalSlitCenter(originalDownstreamVerticalCenter);
    setProgress(95);

    setStatus("Configuring beamline for use");
    moveMiddleAndDownstreamAndSmallSlitsToDefaultSize();
    enableAutobp();

    setProgressAndStatusToComplete();
  }

  protected void scanDownstreamVerticalSlitCenter()
  throws EpicsCaException, InterruptedException {
    performDownstreamVerticalSlitCenterScan();
    if (!scan.completedSuccessfully())
      throw new RuntimeException("Downstream Vertical Slit scan did not " +
        "complete successfully");
    try {
      setMeasuredDataSeriesInEventDispatchThreadAndWait(getData(scan));
    } catch (InvocationTargetException e) {
      throw new RuntimeException("Failed to update measured data series " +
        "with complete data from scan", e);
    }
  }

  protected void scanDownstreamHorizontalSlitCenter()
  throws EpicsCaException, InterruptedException {
    performDownstreamHorizontalSlitCenterScan();
    if (!scan.completedSuccessfully())
      throw new RuntimeException("Downstream Horizontal Slit scan did not " +
        "complete successfully");
    try {
      setMeasuredDataSeriesInEventDispatchThreadAndWait(getData(scan));
    } catch (InvocationTargetException e) {
      throw new RuntimeException("Failed to update measured data series " +
        "with complete data from scan", e);
    }
  }

  @Override
  public void cleanUp() {
    Exception caughtException = null;

    try {
      scan.stop();
    } catch (Exception e) {
      if (caughtException == null) caughtException = e;
    }

    try {
      horizontalBpmMotor.stop();
    } catch (Exception e) {
      if (caughtException == null) caughtException = e;
    }

    try {
      verticalBpmMotor.stop();
    } catch (Exception e) {
      if (caughtException == null) caughtException = e;
    }

    try {
      moveDownstreamHorizontalSlitCenter(originalDownstreamHorizontalCenter);
    } catch (Exception e) {
      if (caughtException == null) caughtException = e;
    }

    try {
      moveDownstreamVerticalSlitCenter(originalDownstreamVerticalCenter);
    } catch (Exception e) {
      if (caughtException == null) caughtException = e;
    }

    try {
      moveMiddleAndDownstreamAndSmallSlitsToDefaultSize();
    } catch (Exception e) {
      if (caughtException == null) caughtException = e;
    }

    try {
      horizontalBpmMotor.setDesiredPosition(
        originalHorizontalBpmMotorPosition);
      horizontalBpmMotor.moveAndWait(HORIZONTAL_BPM_MOVE_TIMEOUT);
    } catch (Exception e) {
      if (caughtException == null) caughtException = e;
    }

    try {
      verticalBpmMotor.setDesiredPosition(originalVerticalBpmMotorPosition);
      verticalBpmMotor.moveAndWait(VERTICAL_BPM_MOVE_TIMEOUT);
    } catch (Exception e) {
      if (caughtException == null) caughtException = e;
    }

    try {
      enableAutobp();
    } catch (Exception e) {
      if (caughtException == null) caughtException = e;
    }

    try {
      super.cleanUp();
    } catch (Exception e) {
      if (caughtException == null) caughtException = e;
    }

    if (caughtException != null) throw new RuntimeException(caughtException);
  }

  protected void performDownstreamHorizontalSlitCenterScan()
  throws EpicsCaException, InterruptedException {
    setXYDatasetTitle("Downstream Horizontal Slit center optimization scan");
    setXDataLabel("Downstream Horizontal Slit center (" +
      downstreamHorizontalSlit.getUnitOfMeasure() + ")");
    setYDataLabel("Small Ion Chamber intensity (counts)");
    scanResetter.reset();
    scan.setBeforeScanPvLink(
      Resources.getDefault().
        getSmallIonChamberIntensityAcquisitionTimePvName());
    scan.setBeforeScanCommandData(1.0f);
    scan.setPositionerPvLink(0,
      Resources.getDefault().
        getDesiredDownstreamHorizontalSlitCenterPvName());
    scan.setPositionerReadbackPvLink(0,
      Resources.getDefault().getDownstreamHorizontalSlitCenterPvName());
    scan.setPositionerReadbackDelta(0, 2.0);
    scan.setPositionerCenteredLinearMode(0,
      getDownstreamHorizontalSlitCenter(), 50.0, 21);
    scan.setPositionerSettlingTime(1.0f);
    scan.setDetectorTriggerPvLink(0,
      Resources.getDefault().
        getSmallIonChamberIntensityAcquisitionStartPvName());
    scan.setDetectorTriggerCommandData(0, 1.0f);
    scan.setDetectorPvLink(0,
      Resources.getDefault().
        getSmallIonChamberIntensityAcquisitionSumPvName());
    openExposureShutter();
    scan.go();
    scan.waitToFinish();
    closeExposureShutter();
  }

  protected void performDownstreamVerticalSlitCenterScan()
  throws EpicsCaException, InterruptedException {
    setXYDatasetTitle("Downstream Vertical Slit center optimization scan");
    setXDataLabel("Downstream Vertical Slit center (" +
      downstreamVerticalSlit.getUnitOfMeasure() + ")");
    setYDataLabel("Small Ion Chamber intensity (counts)");
    scanResetter.reset();
    scan.setBeforeScanPvLink(
      Resources.getDefault().
        getSmallIonChamberIntensityAcquisitionTimePvName());
    scan.setBeforeScanCommandData(1.0f);
    scan.setPositionerPvLink(0,
      Resources.getDefault().
        getDesiredDownstreamVerticalSlitCenterPvName());
    scan.setPositionerReadbackPvLink(0,
      Resources.getDefault().getDownstreamVerticalSlitCenterPvName());
    scan.setPositionerReadbackDelta(0, 2.0);
    scan.setPositionerCenteredLinearMode(0, getDownstreamVerticalSlitCenter(),
      50.0, 21);
    scan.setPositionerSettlingTime(1.0f);
    scan.setDetectorTriggerPvLink(0,
      Resources.getDefault().
        getSmallIonChamberIntensityAcquisitionStartPvName());
    scan.setDetectorTriggerCommandData(0, 1.0f);
    scan.setDetectorPvLink(0,
      Resources.getDefault().
        getSmallIonChamberIntensityAcquisitionSumPvName());
    openExposureShutter();
    scan.go();
    scan.waitToFinish();
    closeExposureShutter();
  }

  protected double getHorizontalBpmPosition()
  throws InterruptedException, EpicsCaException {
    return horizontalBpmMotor.getPosition();
  }

  protected void moveHorizontalBpm(double desiredPosition)
  throws InterruptedException, TimeoutException, EpicsCaException {
    horizontalBpmMotor.setDesiredPosition(desiredPosition);
    horizontalBpmMotor.moveAndWait(HORIZONTAL_BPM_MOVE_TIMEOUT);
  }

  protected void moveHorizontalBpmRelative(double relativeDesiredPosition)
  throws InterruptedException, TimeoutException, EpicsCaException {
    moveHorizontalBpm(getHorizontalBpmPosition() + relativeDesiredPosition);
  }

  protected double getVerticalBpmPosition()
  throws InterruptedException, EpicsCaException {
    return verticalBpmMotor.getPosition();
  }

  protected void moveVerticalBpm(double desiredPosition)
  throws InterruptedException, TimeoutException, EpicsCaException {
    verticalBpmMotor.setDesiredPosition(desiredPosition);
    verticalBpmMotor.moveAndWait(VERTICAL_BPM_MOVE_TIMEOUT);
  }

  protected void moveVerticalBpmRelative(double relativeDesiredPosition)
  throws InterruptedException, TimeoutException, EpicsCaException {
    moveVerticalBpm(getVerticalBpmPosition() + relativeDesiredPosition);
  }
}
