package imcacat.imcabc;

import imcacat.epicsca4j.ca.EpicsCaException;
import imcacat.epicsca4j.model.EpicsMotor;
import imcacat.epicsca4j.util.StringUtilities;
import imcacat.epicsca4j.util.UiUtilities;

import java.lang.reflect.InvocationTargetException;
import java.text.DecimalFormat;

import javax.swing.JOptionPane;

import gov.aps.jca.TimeoutException;

/**
 * I manually optimize Chi.
 */
public class ChiManualOptimization17bm extends Optimization17bm {
  protected static final long CHI_MOVE_TIMEOUT = 60000;

  protected EpicsMotor chiMotor;

  protected double originalChiMotorPosition;

  public ChiManualOptimization17bm() {
    super();
  }

  @Override
  protected void initialize()
  throws InterruptedException, EpicsCaException {
    super.initialize();
    chiMotor = new EpicsMotor(Resources.getDefault().getChiMotorPvName(),
      "\u00B5rad");
    originalChiMotorPosition = chiMotor.getPosition();
  }

  @Override
  public void release() {
    if (chiMotor != null) chiMotor.release();
    super.release();
  }

  @Override
  protected void go()
  throws InterruptedException, TimeoutException, EpicsCaException {
    setStatus("Configuring beamline for Chi optimization");
    disableAutobp();
    setProgress(10);

    setStatus("Scanning Chi");
    scanChi();
    setProgress(60);

    setStatus("Fitting Chi scan data to Gaussian function");
    GaussianFunctionCurveFitter fitter;
    try {
      fitter = new GaussianFunctionCurveFitter(measuredXYSeries.toArray());
    } catch (Exception e) {
      throw new RuntimeException("Failed to fit measured data to Gaussian " +
        "function", e);
    }
    try {
      setFitDataSeriesInEventDispatchThreadAndWait(fitter);
    } catch (InvocationTargetException e) {
      throw new RuntimeException("Failed to add fit data to " +
        "XYSeriesCollection", e);
    }
    setProgress(70);

    double measuredMax = getXWithMaximumY(measuredXYSeries);
    double fitMax = fitter.getC();

    setStatus("Waiting for measured or fit choice");
    DecimalFormat positionFormat = new DecimalFormat("0.00");
    int choice = JOptionPane.showOptionDialog(UiUtilities.getRootFrame(),
      StringUtilities.wordWrap("The Chi scan has finished. The measured " +
        "maximum is at " +
        positionFormat.format(measuredMax) + " " + getChiUnitOfMeasure() +
        ", and the fit maximum is at " +
        positionFormat.format(fitMax) + " " + getChiUnitOfMeasure() +
        ". To which maximum would you like to move Chi?",
        78, "\n"),
      "Move to measured or fit?", JOptionPane.YES_NO_CANCEL_OPTION,
      JOptionPane.QUESTION_MESSAGE, null, new String[] {"Measured", "Fit",
      "Cancel"}, "Fit");
    setProgress(75);

    if (choice == JOptionPane.YES_OPTION) {
      setStatus("Moving Chi to measured maximum");
      moveChi(measuredMax);
    } else if (choice == JOptionPane.NO_OPTION) {
      setStatus("Moving Chi to fit maximum");
      moveChi(fitMax);
    } else {
      setStatus("Canceling");
      moveChi(originalChiMotorPosition);
      enableAutobp();
      setStatus("Canceled");
      return;
    }
    setProgress(90);

    setStatus("Configuring beamline for use");
    enableAutobp();

    setProgressAndStatusToComplete();
  }

  @Override
  public void cleanUp() {
    Exception caughtException = null;

    try {
      scan.stop();
    } catch (Exception e) {
      if (caughtException == null) caughtException = e;
    }
    try {
      chiMotor.stop();
    } catch (Exception e) {
      if (caughtException == null) caughtException = e;
    }
    try {
      chiMotor.setDesiredPosition(originalChiMotorPosition);
      chiMotor.moveAndWait(CHI_MOVE_TIMEOUT);
    } catch (Exception e) {
      if (caughtException == null) caughtException = e;
    }
    try {
      enableAutobp();
    } catch (Exception e) {
      if (caughtException == null) caughtException = e;
    }

    try {
      super.cleanUp();
    } catch (Exception e) {
      if (caughtException == null) caughtException = e;
    }

    if (caughtException != null) throw new RuntimeException(caughtException);
  }

  protected void scanChi()
  throws EpicsCaException, InterruptedException {
    performChiScan();
    if (!scan.completedSuccessfully())
      throw new RuntimeException("Chi scan did not complete successfully");
    try {
      setMeasuredDataSeriesInEventDispatchThreadAndWait(getData(scan));
    } catch (InvocationTargetException e) {
      throw new RuntimeException("Failed to update measured data series " +
        "with complete data from scan", e);
    }
  }

  protected void performChiScan() throws EpicsCaException,
  InterruptedException {
    setXYDatasetTitle("Chi optimization scan");
    setXDataLabel("Chi (" + getChiUnitOfMeasure() + ")");
    setYDataLabel("BPM intensity (counts)");
    scanResetter.reset();
    scan.setBeforeScanPvLink(
      Resources.getDefault().getBpmIntensityAcquisitionTimePvName());
    scan.setBeforeScanCommandData(1.0f);
    scan.setPositionerPvLink(0,
      Resources.getDefault().getDesiredChiPositionPvName());
    scan.setPositionerReadbackPvLink(0,
      Resources.getDefault().getChiPositionPvName());
    scan.setPositionerReadbackDelta(0, 5.0);
    scan.setPositionerCenteredLinearMode(0, getChiPosition(), 100.0, 61);
    scan.setPositionerSettlingTime(1.0f);
    scan.setDetectorTriggerPvLink(0,
      Resources.getDefault().getBpmIntensityAcquisitionStartPvName());
    scan.setDetectorTriggerCommandData(0, 1.0f);
    scan.setDetectorPvLink(0,
      Resources.getDefault().getBpmIntensityAcquisitionSumPvName());
    scan.go();
    scan.waitToFinish();
  }

  protected double getChiPosition()
  throws InterruptedException, EpicsCaException {
    return chiMotor.getPosition();
  }

  protected String getChiUnitOfMeasure() {
    return chiMotor.getUnitOfMeasure();
  }

  protected void moveChi(double desiredPosition)
  throws InterruptedException, TimeoutException, EpicsCaException {
    chiMotor.setDesiredPosition(desiredPosition);
    chiMotor.moveAndWait(CHI_MOVE_TIMEOUT);
  }
}
