package imcacat.imcabc;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.LineNumberReader;
import java.nio.charset.Charset;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.List;
import java.util.regex.Pattern;

/**
 * I read a <tt>chooch</tt> 5.0.2 anomalous scattering factors output file
 * (normally a file  with an extension of <tt>.efs</tt>).  I expect the file
 * to contain data lines of the form:
 * <p>
 * <tt><i>&lt;energy&gt;</i> <i>&lt;f''&gt;</i> <i>&lt;f'&gt;</i></tt>
 * <p>
 * Example:
 * <p>
 * <tt>12638.4261  0.52  -6.43</tt>
 */
public class ChoochAnomalousScatteringFactorsReader extends LineNumberReader {
  protected static final String CHARACTER_SET_NAME = "US-ASCII";
  protected static final Pattern LINE_SPLIT_PATTERN = Pattern.compile(
    "[ \\t]+");

  protected List<Double> energyData = new ArrayList<Double>();
  protected List<Double> fPrimeData = new ArrayList<Double>();
  protected List<Double> fDoublePrimeData = new ArrayList<Double>();

  protected AnomalousScatteringFactors factors;

  public ChoochAnomalousScatteringFactorsReader(File file)
  throws FileNotFoundException, IOException {
    this(new FileInputStream(file));
  }

  public ChoochAnomalousScatteringFactorsReader(InputStream in)
  throws IOException {
    super(new InputStreamReader(in, Charset.forName(CHARACTER_SET_NAME)));
  }

  public AnomalousScatteringFactors parse()
  throws IOException, ParseException {
    if (factors == null) {
      parseInput();
      factors = new AnomalousScatteringFactors(energyData, fPrimeData,
        fDoublePrimeData);
    }
    return factors;
  }

  public List<Double> getEnergyData() {
    return energyData;
  }

  public List<Double> getFPrimeData() {
    return fPrimeData;
  }

  public List<Double> getFDoublePrimeData() {
    return fDoublePrimeData;
  }

  protected void parseInput() throws IOException, ParseException {
    parseData();
  }

  protected void parseData() throws IOException, ParseException {
    for (;;) {
      String line = readLine();
      if (line == null) return;
      parseDataLine(line);
    }
  }

  protected void parseDataLine(String line)
  throws IOException, ParseException {
    String parts[] = LINE_SPLIT_PATTERN.split(line.trim());
    if (parts.length != 3)
      throw new ParseException("Expected 3 values at line " +
        getLineNumber() + ": " + line, 0);
    try {
      energyData.add(new Double(parts[0]));
      fDoublePrimeData.add(new Double(parts[1]));
      fPrimeData.add(new Double(parts[2]));
    } catch (NumberFormatException e) {
      throw new ParseException("Expected \"float float float\" at line " +
        getLineNumber(), 0);
    }
  }
}
