package imcacat.imcabc;

import imcacat.epicsca4j.util.StringUtilities;
import imcacat.imcabc.util.Pair;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.nio.charset.Charset;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;
import java.util.regex.Pattern;

/**
 * I write raw fluorescence data for <tt>chooch</tt>.  What I write may be
 * used as the input file (usually with the extension <code>.raw</code>) for
 * <tt>chooch</tt>.
 */
public class ChoochInputFileWriter extends BufferedWriter {
  protected static final Pattern EDGE_SPLIT_PATTERN = Pattern.compile("_");
  protected static final String CHARACTER_SET_NAME = "US-ASCII";

  public ChoochInputFileWriter(File file) throws IOException {
    this(new FileOutputStream(file));
  }

  public ChoochInputFileWriter(OutputStream out) throws IOException {
    super(new OutputStreamWriter(out, Charset.forName(CHARACTER_SET_NAME)));
  }

  public void write(MadEdge edge, List<Double> energyData,
      List<Double> fluorescenceData)
  throws IOException {
    writeComment(edge);
    List<Double> energies = energyData;
    List<Double> fluorescences = fluorescenceData;
    if (!isMonotonicallyIncreasing(energies)) {
      Pair<List<Double>,List<Double>> data =
          convertToMonotonicallyIncreasingEnergy(energies, fluorescences);
      energies = data.getFirst();
      fluorescences = data.getSecond();
    }
    writeNumberOfDataPoints(energies.size());
    for (int i = 0; i < energies.size(); i++)
      write(energies.get(i), fluorescences.get(i));
  }

  protected void writeComment(MadEdge edge) throws IOException {
    write(String.format(
        "Fluorescence spectrum for %s %s %.1f eV\n",
        StringUtilities.toTitleCase(edge.getElement().toString()),
        EDGE_SPLIT_PATTERN.split(edge.toString())[1], edge.getEnergy()));
  }

  protected void writeNumberOfDataPoints(int size) throws IOException {
    write(size + "\n");
  }

  protected void write(Double energy, Double fluorescence)
  throws IOException {
    write(String.format((Locale)null, "%9.3f %15.9f\n", energy,
        fluorescence));
  }

  protected boolean isMonotonicallyIncreasing(List<Double> values) {
    Double previous = null;
    for (double each : values) {
      if (previous != null && each < previous) return false;
      previous = each;
    }
    return true;
  }

  protected Pair<List<Double>,List<Double>>
      convertToMonotonicallyIncreasingEnergy(List<Double> energies,
      List<Double> fluorescences) {
    ArrayList<Double> resultingEnergies = new ArrayList<Double>(
        energies.size());
    ArrayList<Double> resultingFluorescences = new ArrayList<Double>(
        fluorescences.size());
    Double previousEnergy = null;
    Iterator<Double> energiesIterator = energies.iterator();
    Iterator<Double> fluorescencesIterator = fluorescences.iterator();
    while (energiesIterator.hasNext()) {
      double eachEnergy = energiesIterator.next();
      double eachFluorescence = fluorescencesIterator.next();
      if (previousEnergy == null) {
        resultingEnergies.add(eachEnergy);
        resultingFluorescences.add(eachFluorescence);
        previousEnergy = eachEnergy;
      } else {
        if (eachEnergy >= previousEnergy) {
          resultingEnergies.add(eachEnergy);
          resultingFluorescences.add(eachFluorescence);
          previousEnergy = eachEnergy; 
        }
      }
    }
    return new Pair<List<Double>,List<Double>>(resultingEnergies,
        resultingFluorescences);
  }
}
