package imcacat.imcabc;

import imcacat.epicsca4j.util.StringUtilities;
import imcacat.imcabc.util.FileUtilities;

import java.io.File;
import java.io.IOException;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.List;
import java.util.regex.Pattern;

import org.apache.log4j.Logger;
import org.jdesktop.swingworker.SwingWorker;

/**
 * I calculate anomalous scattering factors (i.e. f' and f'' values) from
 * fluorescence scan data by invoking an external program called
 * <tt>chooch</tt>.
 * <p>
 * I read my data file, create a new data file for <tt>chooch</tt> (in the
 * format <tt>chooch</tt> expects), run <tt>chooch</tt>, and read the
 * resulting data file produced by <tt>chooch</tt> to get the calculated
 * anomalous scattering factors.
 *
 * @see ChoochWorkerResult
 * @see <a href="http://www.ccp4.ac.uk/ccp4bin/viewcvs/chooch/">Chooch</a>
 */
public class ChoochWorker extends SwingWorker<ChoochWorkerResult,Void> {
  public static final String PROPNAME_STATE = "state";
  public static final String PROPNAME_PROGRESS = "progress";

  protected static final Pattern EDGE_SPLIT_PATTERN = Pattern.compile("_");

  protected static final String CHOOCH_INPUT_FILE_EXTENSION = "raw";
  protected static final String CHOOCH_FACTORS_OUTPUT_FILE_EXTENSION = "efs";
  protected static final String CHOOCH_PLOT_OUTPUT_FILE_EXTENSION = "ps";

  private static final Logger LOG = Logger.getLogger(ChoochWorker.class);

  protected File dataFile;
  protected MadEdge edge;

  public ChoochWorker(File dataFile) {
    super();
    this.dataFile = dataFile;
  }

  @Override
  protected ChoochWorkerResult doInBackground() throws Exception {
    String[] choochCommand = new String[] {};
    Process choochProcess = null;
    try {
      loadDataFileAndCreateChoochInputFile();

      removeChoochPlotOutputFileIfPresent();

      choochCommand = createChoochCommand();
      choochProcess = Runtime.getRuntime().exec(choochCommand, null,
        getChoochWorkingDirectory());
      int exitValue = choochProcess.waitFor();

      if (exitValue == 0)
        return new ChoochWorkerResult(getCommandAsString(choochCommand),
          readAnomalousScatteringFactorsFromChoochOutputFile());
      else
        return new ChoochWorkerResult(getCommandAsString(choochCommand),
          exitValue);
    } catch (Exception e) {
      return new ChoochWorkerResult(getCommandAsString(choochCommand), e);
    } finally {
      if (choochProcess != null) choochProcess.destroy();
    }
  }

  protected String[] createChoochCommand() {
    String command[] = new String[12];
    command[0] = Resources.getDefault().getChoochCommand();
    command[1] = "-e";
    command[2] = getElementAsChoochString();
    command[3] = "-a";
    command[4] = getEdgeAsChoochString();
    command[5] = "-r";
    command[6] = getMonochromatorEnergyResolutionAsChoochString();
    command[7] = "-o";
    command[8] = getChoochFactorsOutputFileName();
    command[9] = "-p";
    command[10] = getChoochPlotOutputFileName();
    command[11] = getChoochInputFileName();
    return command;
  }

  protected String getCommandAsString(String[] command) {
    return StringUtilities.join(command, " ");
  }

  protected String getElementAsChoochString() {
    return StringUtilities.toTitleCase(edge.getElement().toString());
  }

  protected String getEdgeAsChoochString() {
    return StringUtilities.toTitleCase(EDGE_SPLIT_PATTERN.split(
        edge.toString())[1]);
  }

  protected String getMonochromatorEnergyResolutionAsChoochString() {
    return String.format("%e",
      Resources.getDefault().getChoochMonochromatorEnergyResolution());
  }

  /**
   * Loads the scan data file and creates the chooch input file.
   *
   * @throws ChoochWorkerLoadDataFileException if an IOException or
   *         ParseException occurred while loading the scan data file
   * @throws ChoochWorkerCreateChoochInputFileException if an IOException
   *         occurred while creating the chooch input file
   *
   * @return chooch input file
   */
  protected File loadDataFileAndCreateChoochInputFile()
  throws ChoochWorkerLoadDataFileException,
  ChoochWorkerCreateChoochInputFileException {
    MadFluorescenceScanDataReader reader = null;
    MadFluorescenceScanData data;
    try {
      reader = new MadFluorescenceScanDataReader(dataFile);
      data = reader.parse();
      reader.close(); reader = null;
    } catch (IOException e) {
      throw new ChoochWorkerLoadDataFileException(e);
    } catch (ParseException e) {
      throw new ChoochWorkerLoadDataFileException(e);
    } finally {
      if (reader != null) try { reader.close(); } catch (Exception e) {}
    }

    edge = data.getEdge();
    List<Double> scaledFluorescenceData = createScaledFluorescenceData(
      data.getIntensity(), data.getFluorescence());

    File inputFile = null;
    ChoochInputFileWriter writer = null;
    try {
      inputFile = getChoochInputFile();
      inputFile.createNewFile();
      writer = new ChoochInputFileWriter(inputFile);
      writer.write(edge, data.getEnergy(), scaledFluorescenceData);
      writer.close(); writer = null;
    } catch (IOException e) {
      throw new ChoochWorkerCreateChoochInputFileException(e);
    } finally {
      if (writer != null) try { writer.close(); } catch (Exception e) {}
    }
    return inputFile;
  }

  /*
   * WORKAROUND: chooch seems to try to append another plot to the plot output
   * file if it already exists. To get the desired behavior of just
   * overwriting an existing plot output file, try to remove it if it exists
   * so chooch will create a new one containing just one plot.
   */
  protected void removeChoochPlotOutputFileIfPresent() {
    try {
      File plotFile = getChoochPlotOutputFile();
      if (plotFile.exists()) plotFile.delete();
    } catch (Exception e) {
      LOG.error("Failed to get or delete chooch plot output file", e);
    }
  }

  /**
   * Reads anomalous scattering factors from the chooch output file.
   *
   * @throws ChoochWorkerReadChoochOutputFileException if an
   *         <code>IOException</code> or <code>ParseException</code> occurred
   *         while reading the chooch output file
   *
   * @return anomalous scattering factors
   */
  protected AnomalousScatteringFactors
  readAnomalousScatteringFactorsFromChoochOutputFile()
  throws ChoochWorkerReadChoochOutputFileException {
    ChoochAnomalousScatteringFactorsReader reader = null;
    try {
      File factorsFile = getChoochFactorsOutputFile();
      reader = new ChoochAnomalousScatteringFactorsReader(factorsFile);
      AnomalousScatteringFactors factors = reader.parse();
      reader.close(); reader = null;
      factors.setEdge(edge);
      return factors;
    } catch (IOException e) {
      throw new ChoochWorkerReadChoochOutputFileException(e);
    } catch (ParseException e) {
      throw new ChoochWorkerReadChoochOutputFileException(e);
    } finally {
      if (reader != null) try { reader.close(); } catch (Exception e) {}
    }
  }

  protected List<Double> createScaledFluorescenceData(
      List<Long> intensityData, List<Long> fluorescenceData) {
    List<Double> data = new ArrayList<Double>(intensityData.size());
    for (int i = 0; i < intensityData.size(); i++)
      data.add(fluorescenceData.get(i).doubleValue() /
        intensityData.get(i).doubleValue());
    return data;
  }

  protected File getChoochInputFile() throws IOException {
    return new File(dataFile.getParentFile(), getChoochInputFileName());
  }

  protected File getChoochWorkingDirectory() {
    return dataFile.getParentFile();
  }

  protected File getChoochFactorsOutputFile() throws IOException {
    return new File(dataFile.getParentFile(),
      getChoochFactorsOutputFileName());
  }

  protected String getChoochFactorsOutputFileName() {
    return FileUtilities.createFileNameWithExtension(dataFile.getName(),
      CHOOCH_FACTORS_OUTPUT_FILE_EXTENSION);
  }

  protected File getChoochPlotOutputFile() throws IOException {
    return new File(dataFile.getParentFile(), getChoochPlotOutputFileName());
  }

  protected String getChoochPlotOutputFileName() {
    return FileUtilities.createFileNameWithExtension(dataFile.getName(),
      CHOOCH_PLOT_OUTPUT_FILE_EXTENSION);
  }

  protected String getChoochInputFileName() {
    return FileUtilities.createFileNameWithExtension(dataFile.getName(),
      CHOOCH_INPUT_FILE_EXTENSION);
  }
}
