package imcacat.imcabc;

import org.jfree.chart.ChartMouseEvent;
import org.jfree.chart.ChartMouseListener;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.event.ChartProgressEvent;
import org.jfree.chart.event.ChartProgressListener;
import org.jfree.chart.plot.XYPlot;

/**
 * I detect when an <code>XYPlot</code> crosshair has changed and invoke my
 * <code>crosshairChanged(Double, Double)</code> method with the new crosshair
 * position. Subclasses should implement this method to do something useful.
 * <p>
 * I should be added as a <code>ChartProgressListener</code> to the
 * <code>JFreeChart</code> instance and as a <code>ChartMouseListener</code>
 * to the <code>ChartPanel</code> instance. The same instance of me should be
 * added to both, not two instances.
 * <p>
 * I am useful because when the mouse clicked event is delivered, the x and y
 * position available from the <code>XYPlot</code> is not the correct
 * position; it is the previous position. The correct position is not
 * available until the chart has finished drawing. I provide the functionality
 * to get the correct crosshair position.
 * <p>
 * <b>NOTE:</b> I expect the plot of the <code>JFreeChart</code> instance to
 * which I am added as a listener to be an <code>XYPlot</code> (i.e. if
 * <code>chart</code> is the <code>JFreeChart</code> instance to which I have
 * been added as a listener, the following cast is valid:
 * <code>(XYPlot)chart.getPlot()</code>).
 *
 * @see org.jfree.chart.plot.XYPlot
 * @see org.jfree.chart.JFreeChart
 * @see org.jfree.chart.ChartPanel
 */
public abstract class CrosshairListener implements ChartMouseListener,
    ChartProgressListener {
  protected boolean hasMouseBeenClicked = false;

  public void chartMouseClicked(ChartMouseEvent event) {
    hasMouseBeenClicked = true;
  }

  public void chartMouseMoved(ChartMouseEvent event) {
  }

  public void chartProgress(ChartProgressEvent event) {
    if (ChartProgressEvent.DRAWING_FINISHED == event.getType()) {
      if (hasMouseBeenClicked) fireCrosshairChangedEvent(event.getChart());
    }
  }

  protected void fireCrosshairChangedEvent(JFreeChart chart) {
    XYPlot plot = (XYPlot)chart.getPlot();
    crosshairChanged(plot.getDomainCrosshairValue(),
      plot.getRangeCrosshairValue());
    hasMouseBeenClicked = false;
  }

  abstract void crosshairChanged(Double domainValue, Double rangeValue);
}
