package imcacat.imcabc;

import imcacat.epicsca4j.bean.OneWayPropertyChangeMediator;
import imcacat.epicsca4j.ca.EpicsCaException;
import imcacat.epicsca4j.convert.ShortToEnumEpicsCaValueConverter;
import imcacat.epicsca4j.model.EpicsMotor;
import imcacat.epicsca4j.model.EpicsPv;
import imcacat.imcabc.util.PhysicsUtilities;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;

/**
 * I am an Energy motor. I support reporting my position in both eV and
 * Angstroms. I support setting my desired position in both eV and Angstroms.
 * I have an additional property, <code>moveMode</code>, that determines
 * the type of motion that will be performed when I am commanded to move.
 * <p>
 * I use {@link PhysicsUtilities#convertToEnergy} and
 * {@link PhysicsUtilities#convertToWavelength} to convert between Energy and
 * Wavelength.
 *
 * @see PhysicsUtilities
 */
public class EnergyEpicsMotor17bm extends EpicsMotor {
  private static final long serialVersionUID = 1;

  public static final String PROPNAME_MOVE_MODE = "moveMode";
  public static final String PROPNAME_WAVELENGTH_POSITION =
    "wavelengthPosition";
  public static final String PROPNAME_WAVELENGTH_DESIRED_POSITION =
    "wavelengthDesiredPosition";

  protected boolean isFineMotion = true;
  protected EpicsPv moveModeBoPv;
  protected EpicsPv moveModeMbbiPv;
  protected OneWayPropertyChangeMediator moveModeMbbiMediator;
  protected double wavelengthDesiredPosition;

  public EnergyEpicsMotor17bm(String recordName)
  throws InterruptedException, EpicsCaException {
    this(recordName, "eV");
  }

  public EnergyEpicsMotor17bm(String recordName, String unitOfMeasure)
  throws InterruptedException, EpicsCaException {
    super(recordName, unitOfMeasure);

    moveModeBoPv = new EpicsPv(
      Resources.getDefault().getEnergyMotorMoveModeBoPvName(), false,
      false, new ShortToEnumEpicsCaValueConverter(EnergyMotorMoveMode.class));

    moveModeMbbiPv = new EpicsPv(
      Resources.getDefault().getEnergyMotorMoveModeMbbiPvName(), false,
      true, new ShortToEnumEpicsCaValueConverter(EnergyMotorMoveMode.class));
    moveModeMbbiMediator = new OneWayPropertyChangeMediator(moveModeMbbiPv,
      this, EpicsPv.PROPNAME_VALUE, PROPNAME_MOVE_MODE);

    positionPv.addPropertyChangeListener(EpicsPv.PROPNAME_VALUE,
      new PropertyChangeListener() {
        public void propertyChange(PropertyChangeEvent e) {
          firePropertyChange(PROPNAME_WAVELENGTH_POSITION,
              PhysicsUtilities.convertToWavelength((Double)e.getOldValue()),
              PhysicsUtilities.convertToWavelength((Double)e.getNewValue()));
        }
      });
  }

  public Double getWavelengthPosition()
  throws InterruptedException, EpicsCaException {
    return PhysicsUtilities.convertToWavelength(getPosition());
  }

  public EnergyMotorMoveMode getMoveMode()
  throws InterruptedException, EpicsCaException {
    return (EnergyMotorMoveMode)moveModeMbbiPv.getValue();
  }

  /**
   * I set the move mode of the Energy pseudo motor.
   *
   * @param newValue the desired move mode
   *
   * @throws IllegalArgumentException if newValue is not FINE nor COARSE
   */
  public void setMoveMode(EnergyMotorMoveMode newValue)
  throws InterruptedException, EpicsCaException {
    if (newValue != EnergyMotorMoveMode.FINE &&
        newValue != EnergyMotorMoveMode.COARSE)
      throw new IllegalArgumentException("Move mode may only be set to " +
        "Fine or Coarse");
    moveModeBoPv.setValue(newValue);
  }

  @Override
  public void setDesiredPosition(Double newValue) {
    super.setDesiredPosition(newValue);
    basicSetWavelengthDesiredPosition(PhysicsUtilities.convertToWavelength(
        newValue));
  }

  /**
   * @return  desired position in Angstroms
   */
  public synchronized Double getWavelengthDesiredPosition() {
    return wavelengthDesiredPosition;
  }

  /**
   * @param newValue  desired position in Angstroms
   */
  public void setWavelengthDesiredPosition(Double newValue) {
    double oldValue = basicSetWavelengthDesiredPosition(newValue);
    if (oldValue != newValue)
      firePropertyChange(PROPNAME_WAVELENGTH_DESIRED_POSITION, oldValue,
        (double)newValue);
    super.setDesiredPosition(PhysicsUtilities.convertToEnergy(newValue));
  }

  protected synchronized double basicSetWavelengthDesiredPosition(
      double newValue) {
    double oldValue = wavelengthDesiredPosition;
    wavelengthDesiredPosition = newValue;
    return oldValue;
  }
}
