package imcacat.imcabc;

import imcacat.epicsca4j.bean.OneWayPropertyChangeMediator;
import imcacat.epicsca4j.bean.ExternallyFireablePropertyChangeBean;
import imcacat.epicsca4j.ca.EpicsCaException;
import imcacat.epicsca4j.convert.ShortToEnumEpicsCaValueConverter;
import imcacat.epicsca4j.model.EpicsPv;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.util.EnumMap;
import java.util.EnumSet;
import java.util.Map;

import org.jfree.util.Log;

import com.jgoodies.binding.beans.Model;

/**
 * I model a filter rack with 8 filters that may be inserted or retracted.  I
 * work with J. Lewis Muir's xia-pfcu-4 EPICS driver.
 */
public class FilterRack extends Model
implements ExternallyFireablePropertyChangeBean {
  private static final long serialVersionUID = 1;

  protected static final int NUM_FILTERS = 8;

  protected static final String BASE_FILTER_GET_NAME = "Filter";
  protected static final String BASE_FILTER_SET_NAME = "Filter";

  protected static final short FILTER_BO_STATE_RETRACT = 0;
  protected static final short FILTER_BO_STATE_INSERT = 1;

  protected static final String BASE_PROPNAME_FILTER_STATE = "filterState";

  public static final String PROPNAME_FILTER_STATES =
    BASE_PROPNAME_FILTER_STATE + "s";
  public static final String PROPNAME_FILTER_STATE0 =
    BASE_PROPNAME_FILTER_STATE + 0;
  public static final String PROPNAME_FILTER_STATE1 =
    BASE_PROPNAME_FILTER_STATE + 1;
  public static final String PROPNAME_FILTER_STATE2 =
    BASE_PROPNAME_FILTER_STATE + 2;
  public static final String PROPNAME_FILTER_STATE3 =
    BASE_PROPNAME_FILTER_STATE + 3;
  public static final String PROPNAME_FILTER_STATE4 =
    BASE_PROPNAME_FILTER_STATE + 4;
  public static final String PROPNAME_FILTER_STATE5 =
    BASE_PROPNAME_FILTER_STATE + 5;
  public static final String PROPNAME_FILTER_STATE6 =
    BASE_PROPNAME_FILTER_STATE + 6;
  public static final String PROPNAME_FILTER_STATE7 =
    BASE_PROPNAME_FILTER_STATE + 7;
  public static final String PROPNAME_INSERT = "insert";
  public static final String PROPNAME_RETRACT = "retract";

  protected String baseRecordName;

  protected FilterRackFilterState[] states;
  protected Map<FilterRackFilter,EpicsPv> statePvs;
  protected Map<FilterRackFilter,OneWayPropertyChangeMediator> stateMediators;
  protected Map<FilterRackFilter,EpicsPv> stateSetterPvs;

  public FilterRack(String baseRecordName)
  throws InterruptedException, EpicsCaException {
    this(baseRecordName, BASE_FILTER_SET_NAME);
  }

  public FilterRack(String baseRecordName, String baseFilterSetName)
      throws InterruptedException, EpicsCaException {
    super();
    this.baseRecordName = baseRecordName;

    statePvs = new EnumMap<FilterRackFilter,EpicsPv>(
      FilterRackFilter.class);
    stateMediators =
      new EnumMap<FilterRackFilter,OneWayPropertyChangeMediator>(
        FilterRackFilter.class);
    for (FilterRackFilter each :
        EnumSet.range(FilterRackFilter.ONE, FilterRackFilter.EIGHT)) {
      EpicsPv pv = new EpicsPv(
        createStateGetterPvName(baseRecordName, BASE_FILTER_GET_NAME, each),
        false, true, new ShortToEnumEpicsCaValueConverter(
          FilterRackFilterState.class));
      pv.addPropertyChangeListener(EpicsPv.PROPNAME_VALUE,
        new FilterStateListener(this, getIndex(each)));
      statePvs.put(each, pv);
      stateMediators.put(each, new OneWayPropertyChangeMediator(pv, this,
          EpicsPv.PROPNAME_VALUE, getStatePropertyName(each)));
    }

    stateSetterPvs = new EnumMap<FilterRackFilter,EpicsPv>(
      FilterRackFilter.class);
    for (FilterRackFilter each : FilterRackFilter.values()) {
      stateSetterPvs.put(each, new EpicsPv(
          createStateSetterPvName(baseRecordName, baseFilterSetName, each),
          false, false));
    }
  }

  public FilterRackFilterState[] getFilterStates()
  throws InterruptedException, EpicsCaException {
    if (states == null) {
      states = new FilterRackFilterState[NUM_FILTERS];
      for (int i = 0; i < NUM_FILTERS; i++) states[i] = getFilterStates(i);
    }
    return states;
  }

  protected void setFilterStates(int filterIndex,
      FilterRackFilterState newState)
  throws InterruptedException, EpicsCaException {
    FilterRackFilterState[] states = getFilterStates();
    states[filterIndex] = newState;
    firePropertyChange(PROPNAME_FILTER_STATES, null, states);
  }

  public FilterRackFilterState getFilterStates(int index)
  throws InterruptedException, EpicsCaException {
    return (FilterRackFilterState)statePvs.get(getFilter(index)).getValue();
  }

  public FilterRackFilterState getFilterState0()
  throws InterruptedException, EpicsCaException {
    return getFilterStates()[0];
  }
  public FilterRackFilterState getFilterState1()
  throws InterruptedException, EpicsCaException {
    return getFilterStates()[1];
  }
  public FilterRackFilterState getFilterState2()
  throws InterruptedException, EpicsCaException {
    return getFilterStates()[2];
  }
  public FilterRackFilterState getFilterState3()
  throws InterruptedException, EpicsCaException {
    return getFilterStates()[3];
  }
  public FilterRackFilterState getFilterState4()
  throws InterruptedException, EpicsCaException {
    return getFilterStates()[4];
  }
  public FilterRackFilterState getFilterState5()
  throws InterruptedException, EpicsCaException {
    return getFilterStates()[5];
  }
  public FilterRackFilterState getFilterState6()
  throws InterruptedException, EpicsCaException {
    return getFilterStates()[6];
  }
  public FilterRackFilterState getFilterState7()
  throws InterruptedException, EpicsCaException {
    return getFilterStates()[7];
  }

  public void setInsert(FilterRackFilter filter)
  throws InterruptedException, EpicsCaException {
    stateSetterPvs.get(filter).setValue(FILTER_BO_STATE_INSERT);
  }

  public void setRetract(FilterRackFilter filter)
  throws InterruptedException, EpicsCaException {
    stateSetterPvs.get(filter).setValue(FILTER_BO_STATE_RETRACT);
  }

  public void externallyFirePropertyChange(PropertyChangeEvent event) {
    firePropertyChange(event);
  }

  public static String getStatePropertyName(FilterRackFilter filter) {
    return getStatePropertyName(getIndex(filter));
  }

  public static String getStatePropertyName(int filter) {
    return BASE_PROPNAME_FILTER_STATE + filter;
  }

  protected void filterStateChanged(int filterIndex,
      FilterRackFilterState newState) {
    try {
      setFilterStates(filterIndex, newState);
    } catch (Exception e) {
      Log.error("Failed to set attenuation filter rack filter state", e);
      return;
    }
  }

  protected static String createStateSetterPvName(String baseRecordName,
      String baseFilterName, FilterRackFilter filter) {
    return createStatePvName(baseRecordName, baseFilterName, filter, "Bo");
  }

  protected static String createStateGetterPvName(String baseRecordName,
      String baseFilterName, FilterRackFilter filter) {
    return createStatePvName(baseRecordName, baseFilterName, filter, "Mbbi");
  }

  protected static String createStatePvName(String baseRecordName,
      String baseFilterName, FilterRackFilter filter, String suffix) {
    if (FilterRackFilter.ALL == filter)
      return baseRecordName + ":" + baseFilterName + "All" + suffix;
    if (FilterRackFilter.ONE == filter)
      return baseRecordName + ":" + baseFilterName + "1" + suffix;
    if (FilterRackFilter.TWO == filter)
      return baseRecordName + ":" + baseFilterName + "2" + suffix;
    if (FilterRackFilter.THREE == filter)
      return baseRecordName + ":" + baseFilterName + "3" + suffix;
    if (FilterRackFilter.FOUR == filter)
      return baseRecordName + ":" + baseFilterName + "4" + suffix;
    if (FilterRackFilter.FIVE == filter)
      return baseRecordName + ":" + baseFilterName + "5" + suffix;
    if (FilterRackFilter.SIX == filter)
      return baseRecordName + ":" + baseFilterName + "6" + suffix;
    if (FilterRackFilter.SEVEN == filter)
      return baseRecordName + ":" + baseFilterName + "7" + suffix;
    if (FilterRackFilter.EIGHT == filter)
      return baseRecordName + ":" + baseFilterName + "8" + suffix;
    throw new RuntimeException("Unknown value: " + filter);
  }

  /**
   * I get the <code>FilterRackFilter</code> for the given filter index.
   *
   * @param filterIndex  index, 0-7, of filter to get
   */
  public static FilterRackFilter getFilter(int filterIndex) {
    if (filterIndex == 0) return FilterRackFilter.ONE;
    if (filterIndex == 1) return FilterRackFilter.TWO;
    if (filterIndex == 2) return FilterRackFilter.THREE;
    if (filterIndex == 3) return FilterRackFilter.FOUR;
    if (filterIndex == 4) return FilterRackFilter.FIVE;
    if (filterIndex == 5) return FilterRackFilter.SIX;
    if (filterIndex == 6) return FilterRackFilter.SEVEN;
    if (filterIndex == 7) return FilterRackFilter.EIGHT;
    throw new IndexOutOfBoundsException("Must be between 0 and 7; instead " +
      "was " + filterIndex);
  }

  /**
   * I get the index of the specified <code>FilterRackFilter</code>.
   *
   * @return index, 0-7, of filter
   */
  public static int getIndex(FilterRackFilter filter) {
    if (FilterRackFilter.ONE   == filter) return 0;
    if (FilterRackFilter.TWO   == filter) return 1;
    if (FilterRackFilter.THREE == filter) return 2;
    if (FilterRackFilter.FOUR  == filter) return 3;
    if (FilterRackFilter.FIVE  == filter) return 4;
    if (FilterRackFilter.SIX   == filter) return 5;
    if (FilterRackFilter.SEVEN == filter) return 6;
    if (FilterRackFilter.EIGHT == filter) return 7;
    throw new RuntimeException("Unexpected value: " + filter);
  }

  private static class FilterStateListener implements PropertyChangeListener {
    private FilterRack rack;
    private int filterIndex;

    public FilterStateListener(FilterRack rack, int filterIndex) {
      this.rack = rack;
      this.filterIndex = filterIndex;
    }

    public void propertyChange(PropertyChangeEvent event) {
      rack.filterStateChanged(filterIndex,
        (FilterRackFilterState)event.getNewValue());
    }
  }
}
