package imcacat.imcabc;

import imcacat.epicsca4j.ca.EpicsCaException;
import imcacat.epicsca4j.model.EpicsMotor;
import imcacat.epicsca4j.util.StringUtilities;
import imcacat.epicsca4j.util.UiUtilities;

import java.lang.reflect.InvocationTargetException;
import java.text.DecimalFormat;

import javax.swing.JOptionPane;

import gov.aps.jca.TimeoutException;

/**
 * I manually optimize Focus.
 */
public class FocusManualOptimization17bm extends Optimization17bm {
  protected static final long FOCUS_MOVE_TIMEOUT = 60000;

  protected EpicsMotor focusMotor;

  protected double originalFocusMotorPosition;

  public FocusManualOptimization17bm() {
    super();
  }

  @Override
  protected void initialize()
  throws InterruptedException, EpicsCaException {
    super.initialize();
    focusMotor = new EpicsMotor(
      Resources.getDefault().getFocusMotorPvName(), "\u03BCm");
    originalFocusMotorPosition = focusMotor.getPosition();
  }

  @Override
  public void release() {
    if (focusMotor != null) focusMotor.release();
    super.release();
  }

  @Override
  protected void go()
  throws InterruptedException, TimeoutException, EpicsCaException {
    setStatus("Configuring beamline for Focus optimization");
    disableAutobp();
    setProgress(10);

    setStatus("Scanning Focus");
    scanFocus();
    setProgress(60);

    setStatus("Fitting Focus scan data to Gaussian function");
    GaussianFunctionCurveFitter fitter;
    try {
      fitter = new GaussianFunctionCurveFitter(measuredXYSeries.toArray());
    } catch (Exception e) {
      throw new RuntimeException("Failed to fit measured data to Gaussian " +
        "function", e);
    }
    try {
      setFitDataSeriesInEventDispatchThreadAndWait(fitter);
    } catch (InvocationTargetException e) {
      throw new RuntimeException("Failed to add fit data to " +
        "XYSeriesCollection", e);
    }
    setProgress(70);

    double measuredMax = getXWithMaximumY(measuredXYSeries);
    double fitMax = fitter.getC();

    setStatus("Waiting for measured or fit choice");
    DecimalFormat positionFormat = new DecimalFormat("0.00");
    int choice = JOptionPane.showOptionDialog(UiUtilities.getRootFrame(),
      StringUtilities.wordWrap("The Focus scan has finished. The measured " +
        "maximum is at " +
        positionFormat.format(measuredMax) + " " + getFocusUnitOfMeasure() +
        ", and the fit maximum is at " +
        positionFormat.format(fitMax) + " " + getFocusUnitOfMeasure() +
        ". To which maximum would you like to move Focus?",
        78, "\n"),
      "Move to measured or fit?", JOptionPane.YES_NO_CANCEL_OPTION,
      JOptionPane.QUESTION_MESSAGE, null, new String[] {"Measured", "Fit",
      "Cancel"}, "Fit");
    setProgress(75);

    if (choice == JOptionPane.YES_OPTION) {
      setStatus("Moving Focus to measured maximum");
      moveFocus(measuredMax);
    } else if (choice == JOptionPane.NO_OPTION) {
      setStatus("Moving Focus to fit maximum");
      moveFocus(fitMax);
    } else {
      setStatus("Canceling");
      moveFocus(originalFocusMotorPosition);
      enableAutobp();
      setStatus("Canceled");
      return;
    }
    setProgress(90);

    setStatus("Configuring beamline for use");
    enableAutobp();

    setProgressAndStatusToComplete();
  }

  @Override
  public void cleanUp() {
    Exception caughtException = null;

    try {
      scan.stop();
    } catch (Exception e) {
      if (caughtException == null) caughtException = e;
    }
    try {
      focusMotor.stop();
    } catch (Exception e) {
      if (caughtException == null) caughtException = e;
    }
    try {
      focusMotor.setDesiredPosition(originalFocusMotorPosition);
      focusMotor.moveAndWait(FOCUS_MOVE_TIMEOUT);
    } catch (Exception e) {
      if (caughtException == null) caughtException = e;
    }
    try {
      enableAutobp();
    } catch (Exception e) {
      if (caughtException == null) caughtException = e;
    }

    try {
      super.cleanUp();
    } catch (Exception e) {
      if (caughtException == null) caughtException = e;
    }

    if (caughtException != null) throw new RuntimeException(caughtException);
  }

  protected void scanFocus()
  throws EpicsCaException, InterruptedException {
    performFocusScan();
    if (!scan.completedSuccessfully())
      throw new RuntimeException("Focus scan did not complete successfully");
    try {
      setMeasuredDataSeriesInEventDispatchThreadAndWait(getData(scan));
    } catch (InvocationTargetException e) {
      throw new RuntimeException("Failed to update measured data series " +
        "with complete data from scan", e);
    }
  }

  protected void performFocusScan() throws EpicsCaException,
  InterruptedException {
    setXYDatasetTitle("Focus optimization scan");
    setXDataLabel("Focus (" + getFocusUnitOfMeasure() + ")");
    setYDataLabel("Small Ion Chamber intensity (counts)");
    scanResetter.reset();
    scan.setBeforeScanPvLink(
      Resources.getDefault().
        getSmallIonChamberIntensityAcquisitionTimePvName());
    scan.setBeforeScanCommandData(1.0f);
    scan.setPositionerPvLink(0,
      Resources.getDefault().getDesiredFocusPositionPvName());
    scan.setPositionerReadbackPvLink(0,
      Resources.getDefault().getFocusPositionPvName());
    scan.setPositionerReadbackDelta(0, 5.0);
    scan.setPositionerCenteredLinearMode(0, getFocusPosition(), -5.0, 20);
    scan.setPositionerSettlingTime(1.0f);
    scan.setDetectorTriggerPvLink(0,
      Resources.getDefault().
        getSmallIonChamberIntensityAcquisitionStartPvName());
    scan.setDetectorTriggerCommandData(0, 1.0f);
    scan.setDetectorPvLink(0,
      Resources.getDefault().
        getSmallIonChamberIntensityAcquisitionSumPvName());
    openExposureShutter();
    scan.go();
    scan.waitToFinish();
    closeExposureShutter();
  }

  protected double getFocusPosition()
  throws InterruptedException, EpicsCaException {
    return focusMotor.getPosition();
  }

  protected String getFocusUnitOfMeasure() {
    return focusMotor.getUnitOfMeasure();
  }

  protected void moveFocus(double desiredPosition)
  throws InterruptedException, TimeoutException, EpicsCaException {
    focusMotor.setDesiredPosition(desiredPosition);
    focusMotor.moveAndWait(FOCUS_MOVE_TIMEOUT);
  }
}
