package imcacat.imcabc;

import imcacat.imcabc.util.GaussianFitter;
import imcacat.imcabc.util.GaussianFunction;

import org.apache.commons.math.FunctionEvaluationException;
import org.apache.commons.math.optimization.general.
       LevenbergMarquardtOptimizer;

/**
 * I fit the specified data to a Gaussian function.  Specifically, the
 * function:
 * <p>
 * <tt>f(x) = a + b*exp(-(x - c)^2 / (2*d^2))</tt>
 * <p>
 * The computed fit parameters have the following meaning:
 * <ul>
 * <li><tt>a</tt> is a constant offset that shifts <tt>f(x)</tt> up or down
 * <li><tt>b</tt> is the height of the peak
 * <li><tt>c</tt> is the position of the center of the peak
 * <li><tt>d</tt> is related to FWHM by <tt>FWHM = 2*sqrt(2*ln(2))*d</tt>
 * </ul>
 * Notation key:
 * <ul>
 * <li><tt>x^n</tt>: <tt>x</tt> raised to the power of <tt>n</tt>
 * <li><tt>x*y</tt>: <tt>x</tt> multiplied by <tt>y</tt>
 * <li><tt>exp(x)</tt>: <i>e</i>^<tt>x</tt>
 * <li><tt>sqrt(x)</tt>: the square root of <tt>x</tt>
 * <li><tt>ln(x)</tt>: the natural logarithm of <tt>x</tt>
 * </ul>
 * References:
 * <ul>
 * <li><a href="http://en.wikipedia.org/wiki/Gaussian_function">Wikipedia
 *   Gaussian function entry</a>
 * </ul>
 */
public class GaussianFunctionCurveFitter implements XYFunction {
  protected double[][] dataPoints;
  protected GaussianFunction fitFunction;

  /**
   * Constructs an instance of me with the specified data points.
   *
   * @param dataPoints array containing two arrays of equal length where the
   *        first array contains the <tt>x</tt> point values and the second
   *        array contains the <tt>y</tt> point values
   */
  public GaussianFunctionCurveFitter(double[][] dataPoints)
  throws FunctionFittingException {
    super();
    setDataPoints(dataPoints);
  }

  public void setDataPoints(double[][] dataPoints)
  throws FunctionFittingException {
    if (dataPoints[0].length != dataPoints[1].length)
      throw new IllegalArgumentException("Size of X data points not same " +
        " as Y data points");
    this.dataPoints = dataPoints;
    performFit();
  }

  public double getY(double x) {
    try {
      return fitFunction.value(x);
    } catch (FunctionEvaluationException e) {
      throw new RuntimeException("Failed to evaluate function at x=" + x, e);
    }
  }

  public double getA() {
    return fitFunction.getA();
  }

  public double getB() {
    return fitFunction.getB();
  }

  public double getC() {
    return fitFunction.getC();
  }

  public double getD() {
    return fitFunction.getD();
  }

  protected void performFit() throws FunctionFittingException {
    GaussianFitter fitter = new GaussianFitter(
      new LevenbergMarquardtOptimizer());
    addDataPointsToGaussianFitter(fitter);
    try {
      fitFunction = fitter.fit();
    } catch (Exception e) {
      throw new FunctionFittingException(e);
    }
  }

  protected void addDataPointsToGaussianFitter(GaussianFitter fitter) {
    for (int i = 0; i < dataPoints[0].length; i++)
      fitter.addObservedPoint(dataPoints[0][i], dataPoints[1][i]);
  }
}
