package imcacat.imcabc;

import java.util.HashMap;
import java.util.Map;

import imcacat.epicsca4j.bean.OneWayPropertyChangeMediator;
import imcacat.epicsca4j.ca.EpicsCaException;
import imcacat.epicsca4j.convert.DoubleToIntegerEpicsCaValueConverter;
import imcacat.epicsca4j.convert.ShortToBooleanEpicsCaValueConverter;
import imcacat.epicsca4j.model.EpicsPv;

/**
 * I extend {@link FilterRack} to support attenuation percentage and total
 * thickness.  I work with J. Lewis Muir's imcaXiaPfcu4 EPICS module.
 */
public class ImcaFilterRack extends FilterRack {
  private static final long serialVersionUID = 1;

  private static final Map<String,ImcaFilterRack> INSTANCES =
    new HashMap<String,ImcaFilterRack>();

  protected static final String BASE_FILTER_SET_NAME = "ImcaFltr";

  protected static final String ATTENUATION_PV_NAME = ":Atten";
  protected static final String REQUESTED_ATTENUATION_PV_NAME = ":RqsAtten";
  protected static final String MAINTAIN_REQUESTED_ATTENUATION_PV_NAME =
    ":MntnRqsAtten";
  protected static final String FILTER_THICKNESSES_PV_NAME = ":FltrThick";
  protected static final String REQUESTED_THICKNESS_PV_NAME = ":RqsThickSum";
  protected static final String THICKNESS_PV_NAME = ":ThickSum";

  public static final String PROPNAME_ATTENUATION = "attenuation";
  public static final String PROPNAME_REQUESTED_ATTENUATION =
    "requestedAttenuation";
  public static final String PROPNAME_MAINTAIN_REQUESTED_ATTENUATION_ENABLED =
    "maintainRequestedAttenuationEnabled";
  public static final String PROPNAME_THICKNESS = "thickness";
  public static final String PROPNAME_REQUESTED_THICKNESS =
    "requestedThickness";

  protected EpicsPv attenuationPv;
  protected OneWayPropertyChangeMediator attenuationMediator;
  protected EpicsPv requestedAttenuationPv;
  protected OneWayPropertyChangeMediator requestedAttenuationMediator;
  protected EpicsPv maintainRequestedAttenuationPv;
  protected OneWayPropertyChangeMediator maintainRequestedAttenuationMediator;
  protected EpicsPv thicknessPv;
  protected OneWayPropertyChangeMediator thicknessMediator;
  protected EpicsPv requestedThicknessPv;
  protected OneWayPropertyChangeMediator requestedThicknessMediator;
  protected int[] filterThicknesses;
  protected EpicsPv[] filterThicknessPvs;
  protected int maxFilterThickness;

  public ImcaFilterRack(String baseRecordName)
  throws InterruptedException, EpicsCaException {
    super(baseRecordName, BASE_FILTER_SET_NAME);

    attenuationPv = new EpicsPv(baseRecordName + ATTENUATION_PV_NAME);
    attenuationMediator = new OneWayPropertyChangeMediator(attenuationPv,
      this, EpicsPv.PROPNAME_VALUE, ImcaFilterRack.PROPNAME_ATTENUATION);

    requestedAttenuationPv = new EpicsPv(
      baseRecordName + REQUESTED_ATTENUATION_PV_NAME);
    requestedAttenuationMediator = new OneWayPropertyChangeMediator(
      requestedAttenuationPv, this, EpicsPv.PROPNAME_VALUE,
      ImcaFilterRack.PROPNAME_REQUESTED_ATTENUATION);

    maintainRequestedAttenuationPv = new EpicsPv(
      baseRecordName + MAINTAIN_REQUESTED_ATTENUATION_PV_NAME, false, true,
      new ShortToBooleanEpicsCaValueConverter());
    maintainRequestedAttenuationMediator = new OneWayPropertyChangeMediator(
      maintainRequestedAttenuationPv, this, EpicsPv.PROPNAME_VALUE,
      ImcaFilterRack.PROPNAME_MAINTAIN_REQUESTED_ATTENUATION_ENABLED);

    thicknessPv = new EpicsPv(baseRecordName + THICKNESS_PV_NAME);
    thicknessMediator = new OneWayPropertyChangeMediator(thicknessPv,
      this, EpicsPv.PROPNAME_VALUE, ImcaFilterRack.PROPNAME_THICKNESS);

    requestedThicknessPv = new EpicsPv(
      baseRecordName + REQUESTED_THICKNESS_PV_NAME);
    requestedThicknessMediator = new OneWayPropertyChangeMediator(
      requestedThicknessPv, this, EpicsPv.PROPNAME_VALUE,
      ImcaFilterRack.PROPNAME_REQUESTED_THICKNESS);

    filterThicknesses = new int[NUM_FILTERS];
    filterThicknessPvs = new EpicsPv[NUM_FILTERS];
    maxFilterThickness = 0;
    for (int i = 0; i < NUM_FILTERS; i++) {
      filterThicknessPvs[i] = new EpicsPv(baseRecordName +
        FILTER_THICKNESSES_PV_NAME + (i + 1), false, false,
        new DoubleToIntegerEpicsCaValueConverter());
      filterThicknesses[i] = (Integer)filterThicknessPvs[i].getValue();
      maxFilterThickness += filterThicknesses[i];
    }
  }

  /**
   * Returns the percentage of flux that is being attenuated (blocked).
   *
   * @return attenuation percentage as a <code>double</code> between 0 and 100
   */
  public double getAttenuation()
  throws InterruptedException, EpicsCaException {
    return (Double)attenuationPv.getValue();
  }

  public double getRequestedAttenuation()
  throws InterruptedException, EpicsCaException {
    return (Double)requestedAttenuationPv.getValue();
  }

  public void setRequestedAttenuation(double attenuation)
  throws InterruptedException, EpicsCaException {
    requestedAttenuationPv.setValueAndWait(attenuation);
  }

  public boolean isMaintainRequestedAttenuationEnabled()
  throws InterruptedException, EpicsCaException {
    return (Boolean)maintainRequestedAttenuationPv.getValue();
  }

  public void setMaintainRequestedAttenuationEnabled(boolean enabled)
  throws InterruptedException, EpicsCaException {
    maintainRequestedAttenuationPv.setValueAndWait(enabled);
  }

  public double getThickness()
  throws InterruptedException, EpicsCaException {
    return (Double)thicknessPv.getValue();
  }

  public double getRequestedThickness()
  throws InterruptedException, EpicsCaException {
    return (Double)requestedThicknessPv.getValue();
  }

  public void setRequestedThickness(double thickness)
  throws InterruptedException, EpicsCaException {
    requestedThicknessPv.setValueAndWait(thickness);
  }

  public int[] getFilterThicknesses() {
    return filterThicknesses;
  }

  public int getMaxFilterThickness() {
    return maxFilterThickness;
  }

  public static synchronized ImcaFilterRack getInstance(String baseRecordName)
  throws InterruptedException, EpicsCaException {
    ImcaFilterRack result = INSTANCES.get(baseRecordName);
    if (result == null) {
      result = new ImcaFilterRack(baseRecordName);
      INSTANCES.put(baseRecordName, result);
    }
    return result;
  }
}
