package imcacat.imcabc;

import imcacat.epicsca4j.ca.EpicsCaException;
import imcacat.epicsca4j.util.StringUtilities;
import imcacat.imcabc.Resources;

import java.awt.BorderLayout;
import java.awt.Image;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.KeyEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;

import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.JComponent;
import javax.swing.JDialog;
import javax.swing.JFrame;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.KeyStroke;
import javax.swing.WindowConstants;

import net.roydesign.app.Application;

/**
 * I am the main frame (i.e. window) for this application.
 */
public abstract class ImcabcFrame extends JFrame {
  private static final long serialVersionUID = 1;

  public static final String FRAME_TITLE = Resources.getAppName();

  protected ImcabcPresentationModel model;

  protected JDialog aboutDialog;
  protected HelpFrame helpFrame;

  protected Action quitAction;
  protected Action helpAction;
  protected Action aboutAction;

  public ImcabcFrame(ImcabcPresentationModel model)
  throws InterruptedException, EpicsCaException {
    super();
    this.model = model;

    initializeActions();
    initializeSpecialMenuBar();
    initializeFrame();
  }

  protected void initializeActions() {
    quitAction = new AbstractAction() {
      private static final long serialVersionUID = 1;
      public void actionPerformed(ActionEvent e) {
        quitActionPerformed();
      }
    };
    if (Resources.isPlatformWindows()) {
      quitAction.putValue(Action.NAME, "Exit");
    } else {
      quitAction.putValue(Action.NAME, "Quit");
      quitAction.putValue(Action.ACCELERATOR_KEY,
        KeyStroke.getKeyStroke(KeyEvent.VK_Q,
          Toolkit.getDefaultToolkit().getMenuShortcutKeyMask()));
    }

    helpAction = new AbstractAction() {
      private static final long serialVersionUID = 1;
      public void actionPerformed(ActionEvent e) {
        helpActionPerformed();
      }
    };
    helpAction.putValue(Action.NAME,
      Resources.getAppName() + " Help");

    aboutAction = new AbstractAction() {
      private static final long serialVersionUID = 1;
      public void actionPerformed(ActionEvent e) {
        aboutActionPerformed();
      }
    };
    aboutAction.putValue(Action.NAME, "About " +
      Resources.getAppName());
  }

  protected JMenuBar buildMenuBar() {
    JMenuBar menuBar = new JMenuBar();

    JMenu fileMenu = new JMenu("File");
    fileMenu.setMnemonic(KeyEvent.VK_F);
    JMenuItem fileQuitMenuItem = new JMenuItem(quitAction);
    if (!Resources.isPlatformWindows())
      fileQuitMenuItem.setMnemonic(KeyEvent.VK_Q);
    fileMenu.add(fileQuitMenuItem);
    menuBar.add(fileMenu);

    JMenu helpMenu = new JMenu("Help");
    helpMenu.setMnemonic(KeyEvent.VK_H);
    JMenuItem helpHelpMenuItem = new JMenuItem(helpAction);
    helpHelpMenuItem.setMnemonic(KeyEvent.VK_H);
    helpMenu.add(helpHelpMenuItem);
    helpMenu.addSeparator();
    JMenuItem helpAboutMenuItem = new JMenuItem(aboutAction);
    helpAboutMenuItem.setMnemonic(KeyEvent.VK_A);
    helpMenu.add(helpAboutMenuItem);
    menuBar.add(helpMenu);

    return menuBar;
  }

  /* If running on Mac OS X, this makes the special Mac OS X Application
   * About and Quit menu items work as expected.
   */
  protected void initializeSpecialMenuBar() {
    Application mrjApp = Application.getInstance();
    mrjApp.getAboutJMenuItem().addActionListener(aboutAction);
    mrjApp.getQuitJMenuItem().addActionListener(quitAction);
  }

  protected void initializeFrame()
  throws InterruptedException, EpicsCaException {
    setTitle(getFrameTitle());
    setIconImage(getFrameIconImage());
    setJMenuBar(buildMenuBar());
    setContentPane(buildContentPane());

    setDefaultCloseOperation(WindowConstants.DO_NOTHING_ON_CLOSE);
    addWindowListener(new WindowAdapter() {
      public void windowClosing(WindowEvent e) {
        quitActionPerformed();
      }
    });
  }

  protected JComponent buildContentPane()
  throws InterruptedException, EpicsCaException {
    JPanel panel = new JPanel(new BorderLayout());
    panel.add(buildMainPanel(), BorderLayout.CENTER);
    return panel;
  }

  protected abstract JComponent buildMainPanel()
  throws InterruptedException, EpicsCaException;

  protected String getFrameTitle() {
    return Resources.getAppName() + " (" + getBeamlineDisplayName() +
      ")";
  }

  protected abstract Beamline getBeamline();

  protected String getBeamlineDisplayName() {
    return getBeamline().getDisplayName();
  }

  protected Image getFrameIconImage() {
    return Resources.getAppIcon().getImage();
  }

  protected void tellUserCannotExitBecauseOperationInProgress(
      Operation operation) {
    String operationName = Resources.getDefault().getDisplayString(
      operation);
    JOptionPane.showMessageDialog(this, StringUtilities.wordWrap(
      (StringUtilities.startsWithVowel(operationName) ? "An " : "A ") +
      operationName + " operation is still in progress. Please stop the " +
      "operation before quitting.", 78, "\n"), "Operation in progress",
      JOptionPane.WARNING_MESSAGE);
  }

  protected synchronized void helpActionPerformed() {
    if (helpFrame == null) {
      helpFrame = new HelpFrame(getBeamline());
      helpFrame.pack();
      helpFrame.setLocationRelativeTo(this);
    }
    helpFrame.setVisible(true);
  }

  protected void aboutActionPerformed() {
    if (aboutDialog == null) {
      aboutDialog = new AboutDialog(this, Resources.getAppName(),
        Resources.getAppVersion(), Resources.getLargeAppIcon());
      aboutDialog.setLocationRelativeTo(this);
    }
    aboutDialog.setVisible(true);
  }

  protected void quitActionPerformed() {
    Operation operation = (Operation)model.getModel(
      Imcabc.PROPNAME_OPERATION).getValue();
    if (Operation.IDLE.equals(operation)) basicExit(0);
    else tellUserCannotExitBecauseOperationInProgress(operation);
  }

  protected String getApplicationVersion() {
    return Resources.getAppVersion();
  }

  protected static void basicExit(int exitStatus) {
    ImcabcSystem.basicExit(exitStatus);
  }
}
