package imcacat.imcabc;

import java.io.IOException;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.Writer;

import imcacat.epicsca4j.event.QueuedEventDispatcher;

import javax.swing.SwingUtilities;
import javax.swing.UIManager;
import javax.swing.UnsupportedLookAndFeelException;

import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;

import com.jgoodies.looks.plastic.PlasticLookAndFeel;

/**
 * I am the core of the imcabc application.
 *
 * I handle starting the imcabc application as well as shutting it down.
 */
public class ImcabcSystem {
  private static final Logger LOG = Logger.getLogger(ImcabcSystem.class);

  public static void main(String[] args) {
    handleArguments(parseArguments(args));
    initializeEpicsca4jEventDispatcher();
    startGui();
  }

  protected static void handleArguments(CliArguments arguments) {
    if (arguments.help()) {
      printHelpAndExit();
    } else if (arguments.version()) {
      printVersionAndExit();
    } else if (arguments.hasConfFile()) {
      try {
        Resources.setDefault(new Resources(arguments.confFile()));
      } catch (IOException e) {
        String message = e.getLocalizedMessage();
        System.err.println(Resources.getAppCliName() +
          ": failed to read configuration file" +
          ((message == null) ? "" : ": " + message));
        basicExit(1);
      }
    }
  }

  protected static CliArguments parseArguments(String[] args) {
    try {
      return CliArgumentsParser.parse(args);
    } catch (CliArgumentException e) {
      String message = e.getMessage();
      if (message == null) message = "invalid command line argument(s)";
      System.err.println(Resources.getAppCliName() + ": error: " + message);
      System.err.println(Resources.getAppCliName() +
        ": invoke with -h option for a help message");
      basicExit(1);
      return null; /* NOTREACHED */
    }
  }

  protected static void printHelpAndExit() {
    printHelpAndExit(System.out, 0);
  }

  protected static void printHelpAndExit(OutputStream out, int exitStatus) {
    try {
      CliArgumentsParser.printHelp(out);
      out.flush();
    } catch (IOException e) {
      LOG.fatal("Fatal exception", e);
      basicExit(1);
    }
    basicExit(exitStatus);
  }

  protected static void printVersionAndExit() {
    printVersionAndExit(System.out);
  }

  protected static void printVersionAndExit(OutputStream out) {
    try {
      printVersion(out);
      out.flush();
    } catch (IOException e) {
      LOG.fatal("Fatal exception", e);
      basicExit(1);
    }
    basicExit(0);
  }

  protected static void printVersionAndExit(Writer out) {
    try {
      printVersion(out);
      out.flush();
    } catch (IOException e) {
      LOG.fatal("Fatal exception", e);
      basicExit(1);
    }
    basicExit(0);
  }

  protected static void printVersion(OutputStream out) throws IOException {
    Writer outWriter = new OutputStreamWriter(out);
    try {
      printVersion(outWriter);
    } finally {
      outWriter.close();
    }
  }

  protected static void printVersion(Writer out) throws IOException {
    out.write(Resources.getAppName() + " " + Resources.getAppVersion());
    out.write(Resources.getLineSeparator());
  }

  public static void basicExit(int exitStatus) {
    LogManager.shutdown();
    System.exit(exitStatus);
  }

  protected static void startGui() {
    initializeUiLookAndFeel();
    SwingUtilities.invokeLater(new Runnable() {
      public void run() {
        try {
          Beamline beamline = Resources.getDefault().getBeamline();
          ImcabcFrame frame;
          if (Beamline.BM17 == beamline) frame = new ImcabcFrame17bm();
          else if (Beamline.ID17 == beamline) frame = new ImcabcFrame17id();
          else throw new ProgrammerError("unexpected beamline: " + beamline);
          frame.pack();
          frame.setLocationRelativeTo(null);
          frame.setVisible(true);
        } catch (Exception e) {
          LOG.fatal("Fatal exception", e);
          basicExit(1);
        }
      }
    });
  }

  protected static void initializeUiLookAndFeel() {
    PlasticLookAndFeel.setSelectTextOnKeyboardFocusGained(true);
    try {
      UIManager.setLookAndFeel(new PlasticLookAndFeel());
    } catch (UnsupportedLookAndFeelException e) {
      LOG.fatal("Failed to set Look & Feel", e);
      throw new RuntimeException(e);
    }
  }

  protected static void initializeEpicsca4jEventDispatcher() {
    imcacat.epicsca4j.util.Resources.getDefaultProperties().setProperty(
      imcacat.epicsca4j.util.Resources.
        EPICSCA4J_JCA_CONTEXT_EVENT_DISPATCHER_KEY,
      QueuedEventDispatcher.class.getCanonicalName());
  }
}
