package imcacat.imcabc;

import imcacat.imcabc.util.ListUtilities;

import java.awt.geom.Point2D;
import java.util.List;

/**
 * I provide linear interpolation on a given set of data points.
 *
 * @see <a href="http://en.wikipedia.org/wiki/Interpolation">Wikipedia
 *      Interpolation entry</a>
 */
public class LinearInterpolator {
  public static final double[][] EMPTY_DATA_POINTS = new double[2][0];

  protected double[][] dataPoints;

  public LinearInterpolator() {
    this(EMPTY_DATA_POINTS);
  }

  /**
   * @param xValues  X data points; must be same size as <code>yValues</code>
   * @param yValues  Y data points; must be same size as <code>xValues</code>
   */
  public LinearInterpolator(List<? extends Number> xValues,
      List<? extends Number> yValues) {
    this(ListUtilities.toDouble2DArray(xValues, yValues));
  }

  /**
   * @param dataPoints  X and Y data points where dataPoints[0] contains the X
   *          values and dataPoints[1] contains the Y values. The number of X
   *          and Y values must be the same.
   */
  public LinearInterpolator(double[][] dataPoints) {
    super();
    setDataPoints(dataPoints);
  }

  public void setDataPoints(double[][] dataPoints) {
    if (dataPoints[0].length != dataPoints[1].length)
      throw new IllegalArgumentException("Size of X data points not same " +
        " as Y data points");
    this.dataPoints = dataPoints;
  }

  public double getX(double y) throws ValueOutOfRangeException {
    Point2D.Double[] points = getInterpolationPointsForY(y);
    Point2D.Double pointA = points[0];
    Point2D.Double pointB = points[1];
    if (pointA.getY() == y) return pointA.getX();
    if (pointB.getY() == y) return pointB.getX();
    return pointA.getX() +
      (((y - pointA.getY()) * (pointB.getX() - pointA.getX())) /
        (pointB.getY() - pointA.getY()));
  }

  public double getY(double x) throws ValueOutOfRangeException {
    Point2D.Double[] points = getInterpolationPointsForX(x);
    Point2D.Double pointA = points[0];
    Point2D.Double pointB = points[1];
    if (pointA.getX() == x) return pointA.getY();
    if (pointB.getX() == x) return pointB.getY();
    return pointA.getY() +
      (((x - pointA.getX()) * (pointB.getY() - pointA.getY())) /
        (pointB.getX() - pointA.getX()));
  }

  protected Point2D.Double[] getInterpolationPointsForX(double x)
  throws ValueOutOfRangeException {
    for (int i = 1; i < dataPoints[0].length; i++)
      if (isBetween(x, dataPoints[0][i-1], dataPoints[0][i]))
        return new Point2D.Double[] {
          new Point2D.Double(dataPoints[0][i-1], dataPoints[1][i-1]),
          new Point2D.Double(dataPoints[0][i], dataPoints[1][i])};
    throw new ValueOutOfRangeException("X value " + x + " not within X data");
  }

  protected Point2D.Double[] getInterpolationPointsForY(double y)
  throws ValueOutOfRangeException {
    for (int i = 1; i < dataPoints[1].length; i++)
      if (isBetween(y, dataPoints[1][i-1], dataPoints[1][i]))
        return new Point2D.Double[] {
          new Point2D.Double(dataPoints[0][i-1], dataPoints[1][i-1]),
          new Point2D.Double(dataPoints[0][i], dataPoints[1][i])};
    throw new ValueOutOfRangeException("Y value " + y + " not within Y data");
  }

  protected boolean isBetween(double value, double boundary1,
      double boundary2) {
    return (value >= boundary1 && value <= boundary2) ||
      (value >= boundary2 && value <= boundary1);
  }
}
