package imcacat.imcabc;

import imcacat.epicsca4j.util.StringUtilities;

import java.io.File;
import java.io.FileFilter;
import java.text.ParseException;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.regex.Pattern;

/**
 * I am a <code>FileFilter</code> for selecting MAD fluorescence scan data
 * files as produced by {@link MadFluorescenceScan}.
 * <p>
 * The syntax of a MAD fluorescence scan data file name is:
 * <pre>
 *   <i>&lt;element&gt;</i>_<i>&lt;edge&gt;</i>_scan_<i>&lt;number&gt;</i
 *   >.dat</pre>
 * where:
 * <ul>
 * <li><tt><i>&lt;element&gt;</i></tt> is the periodic table element symbol in
 *   title case
 * <li><tt><i>&lt;edge&gt;</i></tt> is the edge abbreviation in title case
 * <li><tt><i>&lt;number&gt;</i></tt> is the scan number between 001 and 999
 *   (inclusive)
 * </ul>
 * Examples:
 * <pre>
 * Se_K_scan_001.dat
 * Pb_L3_scan_005.dat</pre>
 */
public class MadFluorescenceScanDataFileFilter implements FileFilter {
  protected static final String EXTENSION = "dat";
  protected static final Pattern SPLIT_PATTERN = Pattern.compile("[_.]");
  protected static final Pattern DIGIT_PATTERN = Pattern.compile("[0-9]+");
  protected static final Pattern EDGE_SPLIT_PATTERN = Pattern.compile("_");
  protected static final Map<String,MadElement> STRING_TO_ELEMENT_MAP =
    new HashMap<String,MadElement>();
  protected static final Map<String,MadEdge> STRING_TO_EDGE_MAP =
    new HashMap<String,MadEdge>();

  static {
    for (MadElement each : MadElement.values())
      STRING_TO_ELEMENT_MAP.put(StringUtilities.toTitleCase(each.toString()),
        each);
    for (MadEdge each : MadEdge.values())
      STRING_TO_EDGE_MAP.put(each.getElement().toString() + "_" +
        StringUtilities.toTitleCase(EDGE_SPLIT_PATTERN.split(
          each.toString())[1]), each);
  }

  protected MadEdge edge;

  public MadFluorescenceScanDataFileFilter() {
    super();
    this.edge = null;
  }

  public MadFluorescenceScanDataFileFilter(MadEdge edge) {
    super();
    this.edge = edge;
  }

  public boolean accept(File aFile) {
    return parseFileName(aFile.getName(), edge) != null;
  }

  /**
   * @return List of two items: edge (MadEdge) and scan number (Integer) or
   *         null if there was a parse error
   */
  public static List<Object> parseFileName(String fileName) {
    return parseFileName(fileName, null);
  }

  /**
   * @return List of two items: edge (MadEdge) and scan number (Integer) or
   *         null if there was a parse error
   */
  public static List<Object> parseFileName(String fileName, MadEdge edge) {
    String parts[] = SPLIT_PATTERN.split(fileName);
    if (parts.length != 5) return null;
    if (!parts[2].equals("scan")) return null;
    if (!parts[4].equals(EXTENSION)) return null;
    MadElement parsedElement = parseElement(parts[0]);
    if (parsedElement == null) return null;
    MadEdge parsedEdge = parseEdge(parts[1], parsedElement);
    if (parsedEdge == null) return null;
    if (!parsedElement.equals(parsedEdge.getElement())) return null;
    if (edge != null && !parsedEdge.equals(edge)) return null;
    Integer scanNumber = parseScanNumber(parts[3]);
    if (scanNumber == null) return null;
    return Arrays.asList(new Object[] { parsedEdge, scanNumber });
  }

  protected static MadElement parseElement(String elementString) {
    return (MadElement)STRING_TO_ELEMENT_MAP.get(elementString);
  }

  protected static MadEdge parseEdge(String edgeString, MadElement element) {
    return (MadEdge)STRING_TO_EDGE_MAP.get(element.toString() + "_" +
      edgeString);
  }

  protected static Integer parseScanNumber(String scanNumberString) {
    if (scanNumberString.length() > 3) return null;
    if (!DIGIT_PATTERN.matcher(scanNumberString).matches()) return null;
    try {
      return Integer.valueOf(scanNumberString);
    } catch (NumberFormatException e) {
      return null;
    }
  }

  public static File constructNewFile(File directory, MadEdge edge,
      int scanNumber) {
    return new File(directory, constructNewFileName(edge, scanNumber));
  }

  public static File constructNewFileAfter(File directory, File file)
  throws ParseException {
    List<Object> parts = parseFileName(file.getName());
    if (parts == null)
      throw new ParseException("Failed to parse specified file: " +
        file.getName(), 0);
    return constructNewFile(directory, (MadEdge)parts.get(0),
      ((Integer)parts.get(1)) + 1);
  }

  public static String constructNewFileName(MadEdge edge, int scanNumber) {
    if (scanNumber > 999)
      throw new IllegalArgumentException("scanNumber must be <= 999");
    return String.format("%s_%s_scan_%03d.dat",
      getFileNameString(edge.getElement()),
      getFileNameString(edge),
      scanNumber);
  }

  protected static String getFileNameString(MadEdge edge) {
    return StringUtilities.toTitleCase(EDGE_SPLIT_PATTERN.split(
        edge.toString())[1]);
  }

  protected static String getFileNameString(MadElement element) {
    return StringUtilities.toTitleCase(element.toString());
  }
}
