package imcacat.imcabc;

import imcacat.epicsca4j.util.ObjectUtilities;
import imcacat.imcabc.util.CollectionUtilities;
import imcacat.imcabc.util.FileUtilities;
import imcacat.imcabc.util.ListUtilities;
import imcacat.imcabc.util.PhysicsUtilities;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.List;

import org.apache.log4j.Logger;
import org.jdesktop.swingworker.SwingWorker;

import com.jgoodies.binding.beans.Model;

/**
 * I process the data created by a {@link MadFluorescenceScan}. Given the
 * data file, I am able to:
 * <ul>
 * <li>calculate the anomalous scattering factors (via {@link ChoochWorker})
 * <li>provide the calculated Peak, Edge, Low, and High values for Energy,
 *   Wavelength, f', and f''
 * <li>save the calculated experiment parameters to a text file
 * </ul>
 */
public class MadFluorescenceScanDataProcessor extends Model {
  private static final long serialVersionUID = 1;

  private static final Logger LOG = Logger.getLogger(
    MadFluorescenceScanDataProcessor.class);

  protected static final String CHOOCH_PLOT_OUTPUT_FILE_EXTENSION = "ps";
  protected static final String EXPERIMENT_PARAMETERS_FILE_EXTENSION = "txt";

  public static final String PROPNAME_DATA_FILE = "dataFile";
  public static final String PROPNAME_CALCULATING = "calculating";
  public static final String PROPNAME_CALCULATION_DONE_REASON =
    "calculationDoneReason";

  public static final String PROPNAME_PEAK_ENERGY = "peakEnergy";
  public static final String PROPNAME_PEAK_WAVELENGTH = "peakWavelength";
  public static final String PROPNAME_PEAK_F_PRIME = "peakFPrime";
  public static final String PROPNAME_PEAK_F_DOUBLE_PRIME =
    "peakFDoublePrime";
  public static final String PROPNAME_EDGE_ENERGY = "edgeEnergy";
  public static final String PROPNAME_EDGE_WAVELENGTH = "edgeWavelength";
  public static final String PROPNAME_EDGE_F_PRIME = "edgeFPrime";
  public static final String PROPNAME_EDGE_F_DOUBLE_PRIME =
    "edgeFDoublePrime";
  public static final String PROPNAME_LOW_ENERGY = "lowEnergy";
  public static final String PROPNAME_LOW_WAVELENGTH = "lowWavelength";
  public static final String PROPNAME_LOW_F_PRIME = "lowFPrime";
  public static final String PROPNAME_LOW_F_DOUBLE_PRIME =
    "lowFDoublePrime";
  public static final String PROPNAME_HIGH_ENERGY = "highEnergy";
  public static final String PROPNAME_HIGH_WAVELENGTH = "highWavelength";
  public static final String PROPNAME_HIGH_F_PRIME = "highFPrime";
  public static final String PROPNAME_HIGH_F_DOUBLE_PRIME =
    "highFDoublePrime";

  protected File dataFile;
  protected MadEdge edge;

  protected ChoochWorker choochWorker;
  protected PropertyChangeListener choochWorkerPropertyChangeListener;
  protected boolean calculating;
  protected MadFluorescenceScanDataProcessorCalculationDoneReason
    calculationDoneReason;

  protected List<Double> factorsEnergyData;
  protected List<Double> factorsFPrimeData;
  protected List<Double> factorsFDoublePrimeData;

  protected LinearInterpolator energyAndFPrimeLinearInterpolator =
    new LinearInterpolator();
  protected LinearInterpolator energyAndFDoublePrimeLinearInterpolator =
    new LinearInterpolator();

  protected Double peakEnergy;
  protected Double peakWavelength;
  protected Double peakFPrime;
  protected Double peakFDoublePrime;
  protected Double edgeEnergy;
  protected Double edgeWavelength;
  protected Double edgeFPrime;
  protected Double edgeFDoublePrime;
  protected Double lowEnergy;
  protected Double lowWavelength;
  protected Double lowFPrime;
  protected Double lowFDoublePrime;
  protected Double highEnergy;
  protected Double highWavelength;
  protected Double highFPrime;
  protected Double highFDoublePrime;

  public MadFluorescenceScanDataProcessor() {
    super();

    choochWorkerPropertyChangeListener = new PropertyChangeListener() {
      public void propertyChange(PropertyChangeEvent event) {
        if (ChoochWorker.PROPNAME_STATE.equals(event.getPropertyName()))
          choochWorkerStateChanged(
            (SwingWorker.StateValue)event.getNewValue());
      }
    };

    calculationDoneReason =
      MadFluorescenceScanDataProcessorCalculationDoneReason.NONE;
  }

  public synchronized File getDataFile() {
    return dataFile;
  }

  public synchronized void setDataFile(File newValue) {
    File oldValue = dataFile;
    dataFile = newValue;
    if (!ObjectUtilities.areEqual(oldValue, newValue))
      firePropertyChange(PROPNAME_DATA_FILE, oldValue, newValue);
    clearAnomalousScatteringFactors();
  }

  /**
   * My data file must have been set before calling this method.
   */
  public synchronized void setCommand(
      MadFluorescenceScanDataProcessorCommand command)
  throws IOException, ParseException,
      MadFluorescenceScanDataProcessorException {
    if (command == MadFluorescenceScanDataProcessorCommand.CALCULATE)
      calculate();
    else if (command ==
      MadFluorescenceScanDataProcessorCommand.CANCEL_CALCULATE)
      cancelCalculate();
    else if (command ==
        MadFluorescenceScanDataProcessorCommand.VIEW_SPECTRA_PLOT)
      openSpectraPlotInExternalViewer();
  }

  /**
   * Calculates anomalous scattering factors.
   * <p>
   * My data file must have been set before calling this method.
   */
  public synchronized void calculate() {
    setChoochWorker(new ChoochWorker(dataFile));
    clearAnomalousScatteringFactors();
    setCalculationDoneReason(
      MadFluorescenceScanDataProcessorCalculationDoneReason.NONE);
    executeChoochWorker();
  }

  public synchronized void cancelCalculate() {
    if (choochWorker != null) choochWorker.cancel(true);
  }

  public synchronized void openSpectraPlotInExternalViewer()
  throws IOException {
    Runtime.getRuntime().exec(
      createOpenSpectraPlotInExternalViewerCommand(), null,
      getScanDataFileDirectory());
  }

  public synchronized void writeExperimentParameters(File destination)
  throws FileNotFoundException, IOException {
    if (!destination.exists()) destination.createNewFile();
    FileOutputStream stream = null;
    try {
      stream = new FileOutputStream(destination);
      writeExperimentParameters(stream);
      stream.close();
      stream = null;
    } finally {
      if (stream != null) {
        try {
          stream.close();
        } catch (Exception e) {
          /* IGNORE */
        }
      }
    }
  }

  public synchronized void writeExperimentParameters(OutputStream stream)
  throws IOException {
    MadExperimentParametersWriter writer = null;
    try {
      writer = new MadExperimentParametersWriter(stream);
      writer.write(this);
      writer.close();
      writer = null;
    } finally {
      if (writer != null) {
        try {
          writer.close();
        } catch (Exception e) {
          /* IGNORE */
        }
      }
    }
  }

  protected void setChoochWorker(ChoochWorker newChoochWorker) {
    if (choochWorker != null)
      choochWorker.removePropertyChangeListener(
        choochWorkerPropertyChangeListener);
    choochWorker = newChoochWorker;
    choochWorkerStateChanged(choochWorker.getState());
    choochWorker.addPropertyChangeListener(
      choochWorkerPropertyChangeListener);
  }

  protected void executeChoochWorker() {
    choochWorker.execute();
  }

  protected void choochWorkerStateChanged(SwingWorker.StateValue newValue) {
    if (SwingWorker.StateValue.DONE.equals(newValue)) {
      choochWorkerFinished();
      setCalculating(false);
    } else if (SwingWorker.StateValue.STARTED.equals(newValue)) {
      setCalculating(true);
    }
  }

  protected void choochWorkerFinished() {
    if (choochWorker.isCancelled()) {
      setCalculationDoneReason(
        MadFluorescenceScanDataProcessorCalculationDoneReason.CANCELED);
      return;
    }

    ChoochWorkerResult result;
    try {
      result = choochWorker.get();
    } catch (Exception e) {
      LOG.error("Failed to get ChoochWorkerResult", e);
      setCalculationDoneReason(
        MadFluorescenceScanDataProcessorCalculationDoneReason.EXCEPTION);
      return;
    }

    if (result.wasSuccessful()) {
      loadAnomalousScatteringFactors(result.getAnomalousScatteringFactors());
      setCalculationDoneReason(
        MadFluorescenceScanDataProcessorCalculationDoneReason.SUCCESS);
    } else if (result.hadNonZeroExitValue()) {
      StringBuilder buffer = new StringBuilder();
      buffer.append("chooch exited abnormally (");
      buffer.append("command=\"");
      buffer.append(result.getCommand());
      buffer.append("\" exitStatus=");
      buffer.append(result.getExitValue());
      buffer.append(")");
      LOG.warn(buffer.toString());
      setCalculationDoneReason(
        MadFluorescenceScanDataProcessorCalculationDoneReason.
          NON_ZERO_EXIT_VALUE);
    } else {
      LOG.error("ChoochWorker had an exception",
        result.getDoInBackgroundException());
      setCalculationDoneReason(lookUpCalculationDoneReason(
        result.getDoInBackgroundException()));
    }
  }

  protected MadFluorescenceScanDataProcessorCalculationDoneReason
  lookUpCalculationDoneReason(Exception e) {
    if (e instanceof ChoochWorkerLoadDataFileException) {
      Throwable cause = e.getCause();
      if (cause instanceof IOException)
        return MadFluorescenceScanDataProcessorCalculationDoneReason.
          LOAD_DATA_FILE_IO_EXCEPTION;
      else if (cause instanceof ParseException)
        return MadFluorescenceScanDataProcessorCalculationDoneReason.
          LOAD_DATA_FILE_PARSE_EXCEPTION;
      else
        throw new ProgrammerError(
          "unexpected ChoochWorkerLoadDataFileException cause");
    } else if (e instanceof ChoochWorkerCreateChoochInputFileException) {
      Throwable cause = e.getCause();
      if (cause instanceof IOException)
        return MadFluorescenceScanDataProcessorCalculationDoneReason.
          CREATE_CHOOCH_INPUT_FILE_IO_EXCEPTION;
      else
        throw new ProgrammerError(
          "unexpected ChoochWorkerCreateChoochInputFileException cause");
    } else if (e instanceof ChoochWorkerReadChoochOutputFileException) {
      Throwable cause = e.getCause();
      if (cause instanceof IOException)
        return MadFluorescenceScanDataProcessorCalculationDoneReason.
          READ_CHOOCH_OUTPUT_FILE_IO_EXCEPTION;
      else if (cause instanceof ParseException)
        return MadFluorescenceScanDataProcessorCalculationDoneReason.
          READ_CHOOCH_OUTPUT_FILE_PARSE_EXCEPTION;
      else
        throw new ProgrammerError(
          "unexpected ChoochWorkerReadChoochOutputFileException cause");
    } else {
      return MadFluorescenceScanDataProcessorCalculationDoneReason.EXCEPTION;
    }
  }

  protected void setCalculating(boolean newValue) {
    boolean oldValue = calculating;
    calculating = newValue;
    if (oldValue != newValue)
      firePropertyChange(PROPNAME_CALCULATING, oldValue, newValue);
  }

  public boolean isCalculating() {
    return calculating;
  }

  protected void setCalculationDoneReason(
      MadFluorescenceScanDataProcessorCalculationDoneReason newValue) {
    MadFluorescenceScanDataProcessorCalculationDoneReason oldValue =
      calculationDoneReason;
    calculationDoneReason = newValue;
    if (!ObjectUtilities.areEqual(oldValue, newValue))
      firePropertyChange(PROPNAME_CALCULATION_DONE_REASON, oldValue,
        newValue);
  }

  public MadFluorescenceScanDataProcessorCalculationDoneReason
  getCalculationDoneReason() {
    return calculationDoneReason;
  }

  protected String[] createOpenSpectraPlotInExternalViewerCommand() {
    List<String> command = new ArrayList<String>();
    command.addAll(Resources.getDefault().getDisplayPostScriptFileCommand());
    command.add(getChoochPlotOutputFileName());
    return command.toArray(new String[0]);
  }

  protected void loadAnomalousScatteringFactors(
      AnomalousScatteringFactors factors) {
    edge = factors.getEdge();

    energyAndFPrimeLinearInterpolator.setDataPoints(new double[][] {
        CollectionUtilities.toPrimitiveDoubleArray(factors.getEnergies()),
        CollectionUtilities.toPrimitiveDoubleArray(factors.getFPrimes())
      });
    energyAndFDoublePrimeLinearInterpolator.setDataPoints(new double[][] {
        CollectionUtilities.toPrimitiveDoubleArray(factors.getEnergies()),
        CollectionUtilities.toPrimitiveDoubleArray(factors.getFDoublePrimes())
      });

    setPeakEnergy(factors.getEnergies().get(ListUtilities.indexOfMax(
        factors.getFDoublePrimes())));
    setEdgeEnergy(factors.getEnergies().get(ListUtilities.indexOfMin(
        factors.getFPrimes())));
    updateLowEnergyFactors(getLowEnergy());
    updateHighEnergyFactors(getHighEnergy());
  }

  protected void clearAnomalousScatteringFactors() {
    energyAndFPrimeLinearInterpolator.setDataPoints(
      LinearInterpolator.EMPTY_DATA_POINTS);
    energyAndFDoublePrimeLinearInterpolator.setDataPoints(
      LinearInterpolator.EMPTY_DATA_POINTS);

    setPeakEnergy(null);
    setEdgeEnergy(null);
    updateLowEnergyFactors(getLowEnergy());
    updateHighEnergyFactors(getHighEnergy());
  }

  public Double getPeakEnergy() {
    return peakEnergy;
  }

  protected void setPeakEnergy(Double newValue) {
    basicSetPeakEnergy(newValue);
    basicSetPeakWavelength(PhysicsUtilities.convertToWavelength(newValue));
  }

  protected void basicSetPeakEnergy(Double newValue) {
    Double oldValue = peakEnergy;
    peakEnergy = newValue;
    if (!ObjectUtilities.areEqual(oldValue, newValue)) {
      firePropertyChange(PROPNAME_PEAK_ENERGY, oldValue, newValue);
      updatePeakEnergyFactors(newValue);
    }
  }

  public Double getPeakWavelength() {
    return peakWavelength;
  }

  protected void setPeakWavelength(Double newValue) {
    basicSetPeakWavelength(newValue);
    basicSetPeakEnergy(PhysicsUtilities.convertToEnergy(newValue));
  }

  protected void basicSetPeakWavelength(Double newValue) {
    Double oldValue = peakWavelength;
    peakWavelength = newValue;
    if (!ObjectUtilities.areEqual(oldValue, newValue))
      firePropertyChange(PROPNAME_PEAK_WAVELENGTH, oldValue, newValue);
  }

  public Double getEdgeEnergy() {
    return edgeEnergy;
  }

  protected void setEdgeEnergy(Double newValue) {
    basicSetEdgeEnergy(newValue);
    basicSetEdgeWavelength(PhysicsUtilities.convertToWavelength(newValue));
  }

  protected void basicSetEdgeEnergy(Double newValue) {
    Double oldValue = edgeEnergy;
    edgeEnergy = newValue;
    if (!ObjectUtilities.areEqual(oldValue, newValue)) {
      firePropertyChange(PROPNAME_EDGE_ENERGY, oldValue, newValue);
      updateEdgeEnergyFactors(newValue);
    }
  }

  public Double getEdgeWavelength() {
    return edgeWavelength;
  }

  protected void setEdgeWavelength(Double newValue) {
    basicSetEdgeWavelength(newValue);
    basicSetEdgeEnergy(PhysicsUtilities.convertToEnergy(newValue));
  }

  protected void basicSetEdgeWavelength(Double newValue) {
    Double oldValue = edgeWavelength;
    edgeWavelength = newValue;
    if (!ObjectUtilities.areEqual(oldValue, newValue))
      firePropertyChange(PROPNAME_EDGE_WAVELENGTH, oldValue, newValue);
  }

  public Double getLowEnergy() {
    return lowEnergy;
  }

  public void setLowEnergy(Double newValue) {
    basicSetLowEnergy(newValue);
    basicSetLowWavelength(PhysicsUtilities.convertToWavelength(newValue));
  }

  protected void basicSetLowEnergy(Double newValue) {
    Double oldValue = lowEnergy;
    lowEnergy = newValue;
    if (!ObjectUtilities.areEqual(oldValue, newValue)) {
      firePropertyChange(PROPNAME_LOW_ENERGY, oldValue, newValue);
      updateLowEnergyFactors(newValue);
    }
  }

  public Double getLowWavelength() {
    return lowWavelength;
  }

  public void setLowWavelength(Double newValue) {
    basicSetLowWavelength(newValue);
    basicSetLowEnergy(PhysicsUtilities.convertToEnergy(newValue));
  }

  protected void basicSetLowWavelength(Double newValue) {
    Double oldValue = lowWavelength;
    lowWavelength = newValue;
    if (!ObjectUtilities.areEqual(oldValue, newValue))
      firePropertyChange(PROPNAME_LOW_WAVELENGTH, oldValue, newValue);
  }

  public Double getHighEnergy() {
    return highEnergy;
  }

  public void setHighEnergy(Double newValue) {
    basicSetHighEnergy(newValue);
    basicSetHighWavelength(PhysicsUtilities.convertToWavelength(newValue));
  }

  protected void basicSetHighEnergy(Double newValue) {
    Double oldValue = highEnergy;
    highEnergy = newValue;
    if (!ObjectUtilities.areEqual(oldValue, newValue)) {
      firePropertyChange(PROPNAME_HIGH_ENERGY, oldValue, newValue);
      updateHighEnergyFactors(newValue);
    }
  }

  public Double getHighWavelength() {
    return highWavelength;
  }

  public void setHighWavelength(Double newValue) {
    basicSetHighWavelength(newValue);
    basicSetHighEnergy(PhysicsUtilities.convertToEnergy(newValue));
  }

  protected void basicSetHighWavelength(Double newValue) {
    Double oldValue = highWavelength;
    highWavelength = newValue;
    if (!ObjectUtilities.areEqual(oldValue, newValue))
      firePropertyChange(PROPNAME_HIGH_WAVELENGTH, oldValue, newValue);
  }

  public Double getEdgeFDoublePrime() {
    return edgeFDoublePrime;
  }

  protected void setEdgeFDoublePrime(Double newValue) {
    Double oldValue = edgeFDoublePrime;
    edgeFDoublePrime = newValue;
    if (!ObjectUtilities.areEqual(oldValue, newValue))
      firePropertyChange(PROPNAME_EDGE_F_DOUBLE_PRIME, oldValue, newValue);
  }

  public Double getEdgeFPrime() {
    return edgeFPrime;
  }

  protected void setEdgeFPrime(Double newValue) {
    Double oldValue = edgeFPrime;
    edgeFPrime = newValue;
    if (!ObjectUtilities.areEqual(oldValue, newValue))
      firePropertyChange(PROPNAME_EDGE_F_PRIME, oldValue, newValue);
  }

  public Double getHighFDoublePrime() {
    return highFDoublePrime;
  }

  protected void setHighFDoublePrime(Double newValue) {
    Double oldValue = highFDoublePrime;
    highFDoublePrime = newValue;
    if (!ObjectUtilities.areEqual(oldValue, newValue))
      firePropertyChange(PROPNAME_HIGH_F_DOUBLE_PRIME, oldValue, newValue);
  }

  public Double getHighFPrime() {
    return highFPrime;
  }

  protected void setHighFPrime(Double newValue) {
    Double oldValue = highFPrime;
    highFPrime = newValue;
    if (!ObjectUtilities.areEqual(oldValue, newValue))
      firePropertyChange(PROPNAME_HIGH_F_PRIME, oldValue, newValue);
  }

  public Double getLowFDoublePrime() {
    return lowFDoublePrime;
  }

  protected void setLowFDoublePrime(Double newValue) {
    Double oldValue = lowFDoublePrime;
    lowFDoublePrime = newValue;
    if (!ObjectUtilities.areEqual(oldValue, newValue))
      firePropertyChange(PROPNAME_LOW_F_DOUBLE_PRIME, oldValue, newValue);
  }

  public Double getLowFPrime() {
    return lowFPrime;
  }

  protected void setLowFPrime(Double newValue) {
    Double oldValue = lowFPrime;
    lowFPrime = newValue;
    if (!ObjectUtilities.areEqual(oldValue, newValue))
      firePropertyChange(PROPNAME_LOW_F_PRIME, oldValue, newValue);
  }

  public Double getPeakFDoublePrime() {
    return peakFDoublePrime;
  }

  protected void setPeakFDoublePrime(Double newValue) {
    Double oldValue = peakFDoublePrime;
    peakFDoublePrime = newValue;
    if (!ObjectUtilities.areEqual(oldValue, newValue))
      firePropertyChange(PROPNAME_PEAK_F_DOUBLE_PRIME, oldValue, newValue);
  }

  public Double getPeakFPrime() {
    return peakFPrime;
  }

  protected void setPeakFPrime(Double newValue) {
    Double oldValue = peakFPrime;
    peakFPrime = newValue;
    if (!ObjectUtilities.areEqual(oldValue, newValue))
      firePropertyChange(PROPNAME_PEAK_F_PRIME, oldValue, newValue);
  }

  protected void updatePeakEnergyFactors(Double energy) {
    setPeakFPrime(getFPrime(energy));
    setPeakFDoublePrime(getFDoublePrime(energy));
  }

  protected void updateEdgeEnergyFactors(Double energy) {
    setEdgeFPrime(getFPrime(energy));
    setEdgeFDoublePrime(getFDoublePrime(energy));
  }

  protected void updateLowEnergyFactors(Double energy) {
    setLowFPrime(getFPrime(energy));
    setLowFDoublePrime(getFDoublePrime(energy));
  }

  protected void updateHighEnergyFactors(Double energy) {
    setHighFPrime(getFPrime(energy));
    setHighFDoublePrime(getFDoublePrime(energy));
  }

  protected Double getFPrime(Double energy) {
    if (energy == null) return null;
    try {
      return energyAndFPrimeLinearInterpolator.getY(energy);
    } catch (ValueOutOfRangeException e) {
      return null;
    }
  }

  protected Double getFDoublePrime(Double energy) {
    if (energy == null) return null;
    try {
      return energyAndFDoublePrimeLinearInterpolator.getY(energy);
    } catch (ValueOutOfRangeException e) {
      return null;
    }
  }

  protected File getScanDataFileDirectory() {
    return dataFile.getParentFile();
  }

  protected String getChoochPlotOutputFileName() {
    return FileUtilities.createFileNameWithExtension(dataFile.getName(),
      CHOOCH_PLOT_OUTPUT_FILE_EXTENSION);
  }

  public synchronized File getExperimentParametersFile() {
    String fileName = getExperimentParametersFileName();
    if (fileName == null || dataFile == null) return null;
    return new File(dataFile.getParentFile(), fileName);
  }

  protected String getExperimentParametersFileName() {
    return dataFile == null
      ? null
      : FileUtilities.createFileNameWithExtension(dataFile.getName(),
          EXPERIMENT_PARAMETERS_FILE_EXTENSION);
  }

  public MadEdge getEdge() {
    return edge;
  }
}
