package imcacat.imcabc;

import imcacat.epicsca4j.util.StringUtilities;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.LineNumberReader;
import java.nio.charset.Charset;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.regex.Pattern;

/**
 * I read the data in the data text format produced by
 * {@link MadFluorescenceScanDataWriter}.
 *
 * @see MadFluorescenceScan
 * @see MadFluorescenceScanData
 */
public class MadFluorescenceScanDataReader extends LineNumberReader {
  protected static final Pattern LINE_SPLIT_PATTERN = Pattern.compile(
    "[ \\t]+");
  protected static final Pattern EDGE_SPLIT_PATTERN = Pattern.compile("_");
  protected static final Map<String,MadEdge> STRING_TO_EDGE_MAP =
    new HashMap<String,MadEdge>();

  static {
    for (MadEdge each : MadEdge.values())
      STRING_TO_EDGE_MAP.put(StringUtilities.toTitleCase(
          each.getElement().toString()) + " " +
        StringUtilities.toTitleCase(EDGE_SPLIT_PATTERN.split(
          each.toString())[1]), each);
  }

  protected MadEdge edge;
  protected List<Double> energyData = new ArrayList<Double>();
  protected List<Long> intensityData = new ArrayList<Long>();
  protected List<Long> fluorescenceData = new ArrayList<Long>();

  protected MadFluorescenceScanData data;

  public MadFluorescenceScanDataReader(File file)
  throws FileNotFoundException, IOException, ParseException {
    this(new FileInputStream(file));
  }

  public MadFluorescenceScanDataReader(InputStream in)
  throws IOException, ParseException {
    super(new InputStreamReader(in, Charset.forName(
        MadFluorescenceScanDataWriter.CHARACTER_SET_NAME)));
  }

  public MadFluorescenceScanData parse() throws IOException, ParseException {
    if (data == null) {
      parseInput();
      data = new MadFluorescenceScanData(edge, energyData, intensityData,
        fluorescenceData);
    }
    return data;
  }

  public MadEdge getEdge() {
    return edge;
  }

  public List<Double> getEnergyData() {
    return energyData;
  }

  public List<Long> getFluorescenceData() {
    return fluorescenceData;
  }

  public List<Long> getIntensityData() {
    return intensityData;
  }

  protected void parseInput() throws IOException, ParseException {
    parseEdge();
    parseData();
  }

  /**
   * I parse a line of the form "<tt>Edge: <i>ELEMENT</i> <i>EDGE</i>
   * <i>ENERGY</i> eV</tt>".
   * <p>
   * Example:
   * <p>
   * <tt>Edge: Se K 12658.0 eV</tt>
   */
  protected void parseEdge() throws IOException, ParseException {
    String line = readLine();
    if (line == null)
      throw new ParseException("Expected Edge header field at line " +
        getLineNumber(), 0);
    String parts[] = LINE_SPLIT_PATTERN.split(line.trim());
    if (parts.length < 3 || !parts[0].equals("Edge:"))
      throw new ParseException("Expected \"Edge: ELEMENT EDGE\" at line " +
        getLineNumber(), 0);
    edge = STRING_TO_EDGE_MAP.get(parts[1] + " " + parts[2]);
    if (edge == null)
      throw new ParseException("Enknown ELEMENT EDGE \"" + parts[1] + " " +
        parts[2] + "\" at line " + getLineNumber(), 0);
  }

  protected void parseData() throws IOException, ParseException {
    for (;;) {
      String line = readLine();
      if (line == null) return;
      parseDataLine(line);
    }
  }

  protected void parseDataLine(String line)
  throws IOException, ParseException {
    String parts[] = LINE_SPLIT_PATTERN.split(line.trim());
    if (parts.length != 3)
      throw new ParseException("Expected 3 values at line " +
        getLineNumber() + ": " + line, 0);
    try {
      energyData.add(new Double(parts[0]));
      intensityData.add(new Long(parts[1]));
      fluorescenceData.add(new Long(parts[2]));
    } catch (NumberFormatException e) {
      throw new ParseException("Expected \"float integer integer\" at line " +
        getLineNumber(), 0);
    }
  }
}
