package imcacat.imcabc;

import imcacat.epicsca4j.util.StringUtilities;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.nio.charset.Charset;
import java.util.List;
import java.util.Locale;
import java.util.regex.Pattern;

/**
 * I write the data from a {@link MadFluorescenceScan} to a text data format.
 * <p>
 * The text data format is as follows:
 * <p>
 * The first line is of the form:
 * <pre>
 * Edge: <i>&lt;element&gt;</i> <i>&lt;edge&gt;</i> <i>&lt;energy&gt;</i
 *   > eV</pre>
 * where:
 * <ul>
 * <li><tt><i>&lt;element&gt;</i></tt> is the periodic table element symbol in
 *   title case
 * <li><tt><i>&lt;edge&gt;</i></tt> is the edge abbreviation in title case
 * <li><tt><i>&lt;energy&gt;</i></tt> is the energy of the edge in eV
 *   (inclusive)
 * </ul>
 * The second and remaining lines are data points of the form:
 * <pre>
 * <i>&lt;energy&gt;</i> <i>&lt;intensity&gt;</i> <i>&lt;fluorescence&gt;</i
 *   ></pre>
 * where:
 * <ul>
 * <li><tt><i>&lt;energy&gt;</i></tt> is the Energy in eV
 * <li><tt><i>&lt;intensity&gt;</i></tt> is the Intensity monitor count sum
 * <li><tt><i>&lt;fluorescence&gt;</i></tt> is the Fluorescence detector
 *   region of interest count sum
 * </ul>
 * The columns are delimited by one or more space characters.
 * <p>
 * Example:
 * <pre>
 * Edge: Se K 12658.0 eV
 * 12637.995  4379    58
 * 12642.007  4381    59
 * 12645.993  4385    59
 * 12649.997  4393    66</pre>
 *
 * @see MadFluorescenceScanDataReader
 * @see MadFluorescenceScan
 */
public class MadFluorescenceScanDataWriter extends BufferedWriter {
  protected static final Pattern EDGE_SPLIT_PATTERN = Pattern.compile("_");
  protected static final String CHARACTER_SET_NAME = "US-ASCII";

  public MadFluorescenceScanDataWriter(File file) throws IOException {
    this(new FileOutputStream(file));
  }

  public MadFluorescenceScanDataWriter(OutputStream out) throws IOException {
    super(new OutputStreamWriter(out, Charset.forName(CHARACTER_SET_NAME)));
  }

  /**
   * @param energyData energy data in eV
   */
  public void write(MadEdge edge, List<Double> energyData,
      List<Float> intensityData, List<Float> fluorescenceData)
  throws IOException {
    write(edge);
    for (int i = 0; i < energyData.size(); i++)
      write(energyData.get(i), intensityData.get(i), fluorescenceData.get(i));
  }

  protected void write(MadEdge edge) throws IOException {
    write(String.format((Locale)null,
        "Edge: %s %s %.1f eV\n",
        StringUtilities.toTitleCase(edge.getElement().toString()),
        EDGE_SPLIT_PATTERN.split(edge.toString())[1], edge.getEnergy()));
  }

  protected void write(Double energy, Float intensity, Float fluorescence)
  throws IOException {
    write(String.format((Locale)null, "%9.3f %5d %5d\n", energy,
      intensity.intValue(), fluorescence.intValue()));
  }
}
