package imcacat.imcabc;

import gov.aps.jca.TimeoutException;
import imcacat.epicsca4j.ca.EpicsCaException;
import imcacat.epicsca4j.convert.ShortToBooleanEpicsCaValueConverter;
import imcacat.epicsca4j.convert.ShortToEnumEpicsCaValueConverter;
import imcacat.epicsca4j.model.EpicsMenuScan;
import imcacat.epicsca4j.model.EpicsPv;
import imcacat.epicsca4j.model.EpicsScan;
import imcacat.epicsca4j.model.EpicsScanDataPointEvent;
import imcacat.epicsca4j.model.EpicsScanDataPointListener;
import imcacat.epicsca4j.model.EpicsScanFactory;
import imcacat.epicsca4j.util.EpicsUtilities;
import imcacat.epicsca4j.util.StringUtilities;
import imcacat.epicsca4j.util.UiUtilities;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.File;
import java.io.IOException;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.TimeUnit;

import javax.swing.JOptionPane;

import org.apache.log4j.Logger;
import org.jdesktop.swingworker.SwingWorker;
import org.jfree.data.general.SeriesException;
import org.jfree.data.xy.XYDataset;
import org.jfree.data.xy.XYSeries;
import org.jfree.data.xy.XYSeriesCollection;

/**
 * I run a fluorescence scan for a MAD experiment.
 */
public abstract class MadFluorescenceScanRunner
extends SwingWorker<Void,Void> {
  private static final Logger LOG = Logger.getLogger(
    MadFluorescenceScanRunner.class);

  protected static final String MEASURED_SERIES_KEY = "Fluorescence";
  protected static final double[][] EMPTY_DATASET_SERIES =
    new double[][] {{}, {}};

  protected static final String RUNNING_FORMAT = "Running (step %d of %d)";

  protected static final long AUTOBP_REQUEST_PAUSE_TIMEOUT = 185000;

  protected static final short XFLASH_COOLING_MODE_THERMOSTAT = 1;
  protected static final String XFLASH_THERM_STATE_STABLE = "Stable";
  protected static final long XFLASH_COOL_DOWN_TIMEOUT = 45000;
  protected static final short XFLASH_CONFIG_COOL_DOWN_POLL_RATE =
    EpicsMenuScan.ONE_SECOND.asEpicsShort();
  protected static final short XFLASH_CONFIG_DURING_SCAN_POLL_RATE =
    EpicsMenuScan.PASSIVE.asEpicsShort();
  protected static final long XFLASH_CONFIG_DURING_SCAN_POLL_RATE_CHANGE_DLY =
    1000;

  protected static final long EXPERIMENT_STATE_CHANGE_TIMEOUT = 60000;
  protected static final long EXPERIMENT_STATE_ABORT_TIMEOUT = 10000;

  public static final String PROPNAME_RUNNING = "running";
  public static final String PROPNAME_STATUS = "status";
  public static final String PROPNAME_PROGRESS = "progress";
  public static final String PROPNAME_XY_DATASET = "XYDataset";
  public static final String PROPNAME_OUTPUT_FILE = "dataOutputFile";

  protected boolean isRunning;
  protected String status;

  protected MadEdge edge;
  protected MadScanType scanType;
  protected File outputDirectory;
  protected File outputFile;
  protected boolean shouldSaveDataToFile;
  protected int scanDataPointsCount;
  protected int numScanDataPoints;
  protected boolean okToSetRunningStatusAndProgress;
  protected EpicsScan scan;
  protected EpicsScanDataPointListener scanDataPointListener;
  protected EpicsScanResetter scanResetter;

  double energyAxisToEvScaleFactor;
  protected XYSeries measuredXYSeries;
  protected XYSeriesCollection xySeriesCollection;

  protected EpicsPv autobpRequestExternalPausePv;
  protected EpicsPv autobpReleaseExternalPausePv;
  protected EpicsPv energyAxisAutoStartPv;
  protected EpicsPv mcaRoiLowPv;
  protected EpicsPv mcaRoiHighPv;
  protected EpicsPv mcaAcquisitionTimePv;
  protected EpicsPv xFlashConfigCoolingModePv;
  protected short originalXFlashConfigCoolingMode;
  protected EpicsPv xFlashConfigThermStatePv;
  protected EpicsPv xFlashConfigPollRatePv;
  protected short originalXFlashConfigPollRate;
  protected ImcaFilterRack filterRack;
  protected boolean originalAutoAttenuatorEnabled;
  protected EpicsPv experimentStatePv;
  protected EpicsPv experimentStateRbvPv;
  protected EpicsPv experimentStateSetPv;
  protected EpicsPv experimentStateStatusPv;
  protected EpicsPv experimentStateAbortPv;

  public MadFluorescenceScanRunner()
  throws InterruptedException, EpicsCaException {
    this(MadEdge.SE_K);
  }

  public MadFluorescenceScanRunner(MadEdge edge)
  throws InterruptedException, EpicsCaException {
    this(edge, MadScanType.STANDARD);
  }

  public MadFluorescenceScanRunner(MadEdge edge, MadScanType scanType)
  throws InterruptedException, EpicsCaException {
    this(edge, scanType, Resources.getUserHomeDirectory());
  }

  public MadFluorescenceScanRunner(MadEdge edge, MadScanType scanType,
      File outputDirectory)
  throws InterruptedException, EpicsCaException {
    super();

    isRunning = false;
    status = "";

    this.edge = edge;
    this.scanType = scanType;
    this.outputDirectory = outputDirectory;

    shouldSaveDataToFile = true;
    scanDataPointsCount = 0;
    numScanDataPoints = 0;
    okToSetRunningStatusAndProgress = false;

    energyAxisToEvScaleFactor =
      Resources.getDefault().getEnergyAxisToEvScaleFactor();
    measuredXYSeries = new XYSeries(MEASURED_SERIES_KEY, true, false);
    xySeriesCollection = new XYSeriesCollection(measuredXYSeries);
  }

  protected void initialize() throws EpicsCaException, InterruptedException {
    scan = EpicsScanFactory.getScanInstance(
      Resources.getDefault().getMadScanPvName());

    scanDataPointListener = new EpicsScanDataPointListener() {
      public void dataReceived(EpicsScanDataPointEvent event) {
        receivedScanDataPoint(event);
      }
    };
    addDataPointListenerToScan();

    scanResetter = new EpicsScanResetter(
      Resources.getDefault().getMadScanResetPvName());

    autobpRequestExternalPausePv = new EpicsPv(
      Resources.getDefault().getAutobpRequestExternalPausePvName(), false,
      false);
    autobpReleaseExternalPausePv = new EpicsPv(
      Resources.getDefault().getAutobpReleaseExternalPausePvName(), false,
      false);

    if (Resources.getDefault().isEnergyAxisAutoStartSupported())
      energyAxisAutoStartPv = new EpicsPv(
        Resources.getDefault().getEnergyAxisAutoStartPvName(), false, false,
        new ShortToBooleanEpicsCaValueConverter((short)0, (short)1));

    mcaRoiLowPv = new EpicsPv(
      Resources.getDefault().getFluorescenceMcaRoiLowPvName(), false, false);
    mcaRoiHighPv = new EpicsPv(
      Resources.getDefault().getFluorescenceMcaRoiHighPvName(), false,
      false);
    mcaAcquisitionTimePv = new EpicsPv(
      Resources.getDefault().getFluorescenceMcaAcquisitionTimePvName(),
      false, false);

    if (Resources.getDefault().isXFlashConfigSupported()) {
      xFlashConfigCoolingModePv = new EpicsPv(
        Resources.getDefault().getXFlashConfigCoolingModePvName(), false,
        false);
      originalXFlashConfigCoolingMode =
        (Short)xFlashConfigCoolingModePv.getValue();
      xFlashConfigThermStatePv = new EpicsPv(
        Resources.getDefault().getXFlashConfigThermStatePvName(), false,
        true);
      xFlashConfigPollRatePv = new EpicsPv(
        Resources.getDefault().getXFlashConfigPollRatePvName(), false,
        false);
      originalXFlashConfigPollRate = (Short)xFlashConfigPollRatePv.getValue();
    }

    filterRack = ImcaFilterRack.getInstance(
      Resources.getDefault().getFilterRackBasePvName());
    originalAutoAttenuatorEnabled = false;

    if (Resources.getDefault().isExperimentStateSupported()) {
      experimentStatePv = new EpicsPv(
        Resources.getDefault().getExperimentStatePvName(), false, false,
        new ShortToEnumEpicsCaValueConverter(ExperimentState.class));
      experimentStateRbvPv = new EpicsPv(
        Resources.getDefault().getExperimentStateRbvPvName(), false, false,
        new ShortToEnumEpicsCaValueConverter(ExperimentState.class));
      experimentStateSetPv = new EpicsPv(
        Resources.getDefault().getExperimentStateSetPvName(), false, false);
      experimentStateStatusPv = new EpicsPv(
        Resources.getDefault().getExperimentStateStatusPvName(), false, true,
        new ShortToEnumEpicsCaValueConverter(ExperimentStateStatus.class));
      experimentStateAbortPv = new EpicsPv(
        Resources.getDefault().getExperimentStateAbortPvName(), false, false);
    }
  }

  public void release() {
    if (scan != null)
      removeDataPointListenerFromScan();
    if (scanResetter != null)
      scanResetter.release();
    if (autobpRequestExternalPausePv != null)
      autobpRequestExternalPausePv.release();
    if (autobpReleaseExternalPausePv != null)
      autobpReleaseExternalPausePv.release();
    if (energyAxisAutoStartPv != null)
      energyAxisAutoStartPv.release();
    if (mcaRoiLowPv != null)
      mcaRoiLowPv.release();
    if (mcaRoiHighPv != null)
      mcaRoiHighPv.release();
    if (mcaAcquisitionTimePv != null)
      mcaAcquisitionTimePv.release();
    if (xFlashConfigCoolingModePv != null)
      xFlashConfigCoolingModePv.release();
    if (xFlashConfigThermStatePv != null)
      xFlashConfigThermStatePv.release();
    if (xFlashConfigPollRatePv != null)
      xFlashConfigPollRatePv.release();
    if (experimentStatePv != null)
      experimentStatePv.release();
    if (experimentStateRbvPv != null)
      experimentStateRbvPv.release();
    if (experimentStateSetPv != null)
      experimentStateSetPv.release();
    if (experimentStateStatusPv != null)
      experimentStateStatusPv.release();
    if (experimentStateAbortPv != null)
      experimentStateAbortPv.release();
  }

  public Void doInBackground() throws Exception {
    setRunning(true);

    try {
      setStatus("Initializing");
      initialize();
    } catch (Exception e) {
      if (e instanceof InterruptedException) setStatus("");
      else setStatus("Initialization failed");
      LOG.error("Initialization failed", e);
      setRunning(false);
      return null;
    }

    try {
      go();
    } catch (InterruptedException e) {
      performCleanUp("Stopping", "", "Stop failed");
    } catch (Exception e) {
      LOG.error("Scan failed", e);
      performCleanUp("Failure; stopping", "", "Failed");
    } finally {
      try {
        release();
      } catch (Exception e) {
        LOG.error("Failed to release scan runner after run", e);
      }
    }

    setRunning(false);
    return null;
  }

  protected void performCleanUp(String beforeCleanUp, String cleanUpSuccess,
      String cleanUpFailure) {
    setStatus(beforeCleanUp);
    try {
      cleanUp();
      setStatus(cleanUpSuccess);
    } catch (Exception e2) {
      setStatus(cleanUpFailure + "; beamline in inconsistent state");
      LOG.error("Failed to perform scan clean up; beamline in inconsistent " +
        "state", e2);
    }
  }

  public void go()
  throws InterruptedException, EpicsCaException, TimeoutException,
         IOException, ExperimentStateNotAchievedException {
    if (shouldSaveDataToFile) {
      setStatus("Configuring data output file");
      configureDataOutputFile();
    }

    setStatus("Disabling auto beam positioner");
    disableAutobp();

    if (Resources.getDefault().isXFlashConfigSupported()) {
      setStatus("Starting detector cool down");
      if (originalXFlashConfigCoolingMode != XFLASH_COOLING_MODE_THERMOSTAT)
        xFlashConfigCoolingModePv.setValue(XFLASH_COOLING_MODE_THERMOSTAT);
    }

    if (isAutoAttenuatorEnabled()) {
      originalAutoAttenuatorEnabled = true;
      setStatus("Disabling auto attenuator");
      disableAutoAttenuator();
    }

    setStatus("Changing energy");
    List<Double> energyPositions = createEnergyPositions();
    int numPositions = energyPositions.size();
    if (numPositions > 0) moveEnergyAndWait(energyPositions.get(0));
    setNumScanDataPoints(numPositions);

    if (Resources.getDefault().isXFlashConfigSupported()) {
      setStatus("Waiting for detector to cool down");
      if (originalXFlashConfigPollRate != XFLASH_CONFIG_COOL_DOWN_POLL_RATE)
        xFlashConfigPollRatePv.setValue(XFLASH_CONFIG_COOL_DOWN_POLL_RATE);
      xFlashConfigThermStatePv.waitForValue(XFLASH_THERM_STATE_STABLE,
        XFLASH_COOL_DOWN_TIMEOUT);
      xFlashConfigPollRatePv.setValue(XFLASH_CONFIG_DURING_SCAN_POLL_RATE);
      Thread.sleep(XFLASH_CONFIG_DURING_SCAN_POLL_RATE_CHANGE_DLY);
    }

    if (Resources.getDefault().isExperimentStateSupported()) {
      setStatus("Setting expt. state to \u201CCollect Data\u201D");
      enterExperimentState(ExperimentState.COLLECT_DATA);
    }

    configureScan(energyPositions);

    setStatus(String.format(RUNNING_FORMAT, 0, numPositions));
    setOkToSetRunningStatusAndProgress(true);
    scan.go();
    scan.waitToFinish();
    setOkToSetRunningStatusAndProgress(false);
    if (scan.completedSuccessfully()) {
      setStatus(String.format(RUNNING_FORMAT, numPositions, numPositions));
      setProgress(100);
    } else {
      setStatus("Scan failed");
      if (!scan.completedSuccessfully())
        throw new RuntimeException("Scan did not complete successfully");
    }

    setStatus("Moving energy to theoretical edge");
    moveEnergyAndWait(edge.getEnergy() / energyAxisToEvScaleFactor);

    if (originalAutoAttenuatorEnabled) {
      setStatus("Enabling auto attenuator");
      enableAutoAttenuator();
    }

    if (Resources.getDefault().isExperimentStateSupported()) {
      setStatus("Setting expt. state to \u201CIdle\u201D");
      enterExperimentState(ExperimentState.IDLE);
    }

    if (Resources.getDefault().isXFlashConfigSupported()) {
      setStatus("Starting detector warm up");
      if (originalXFlashConfigCoolingMode != XFLASH_COOLING_MODE_THERMOSTAT) {
        xFlashConfigCoolingModePv.setValue(originalXFlashConfigCoolingMode);
        /* Make the cooling mode readback PV update to the new mode. */
        xFlashConfigPollRatePv.setValue(EpicsMenuScan.POINT_ONE_SECOND.asEpicsShort());
        Thread.sleep(500);
      }
      xFlashConfigPollRatePv.setValue(originalXFlashConfigPollRate);
    }

    setStatus("Enabling auto beam positioner");
    enableAutobp();

    setStatus("");
  }

  protected void disableAutobp()
  throws InterruptedException, TimeoutException, EpicsCaException {
    autobpRequestExternalPausePv.setValueAndWait(
      Resources.getDefault().getAutobpRequestExternalPauseValue(),
      AUTOBP_REQUEST_PAUSE_TIMEOUT);
  }

  protected void enableAutobp()
  throws InterruptedException, EpicsCaException {
    autobpReleaseExternalPausePv.setValue(
      Resources.getDefault().getAutobpReleaseExternalPauseValue());
  }

  protected boolean isAutoAttenuatorEnabled()
  throws InterruptedException, EpicsCaException {
    return filterRack.isMaintainRequestedAttenuationEnabled();
  }

  protected void disableAutoAttenuator()
  throws InterruptedException, EpicsCaException {
    filterRack.setMaintainRequestedAttenuationEnabled(false);
  }

  protected void enableAutoAttenuator()
  throws InterruptedException, EpicsCaException {
    filterRack.setMaintainRequestedAttenuationEnabled(true);
  }

  /**
   * Moves energy to the specified position, waiting for the move to complete.
   *
   * @param scaledDesiredPosition position to which energy should be moved;
   *        this position value is scaled (i.e. it is the desired energy
   *        position in eV divided by energyAxisToEvScaleFactor)
   */
  protected abstract void moveEnergyAndWait(double scaledDesiredPosition)
  throws InterruptedException, TimeoutException, EpicsCaException;

  /**
   * Stops an energy move.
   */
  protected abstract void stopEnergyMove()
  throws InterruptedException, EpicsCaException;

  protected void enterExperimentState(ExperimentState desiredState)
  throws InterruptedException, TimeoutException, EpicsCaException,
         ExperimentStateNotAchievedException {
    ExperimentStateStatus status =
      (ExperimentStateStatus)experimentStateStatusPv.getValue();
    if (ExperimentStateStatus.EXECUTING.equals(status)) {
      experimentStateAbortPv.setValue(EpicsUtilities.PROC_FIELD_PUT_VALUE);
      experimentStateStatusPv.waitForAnyOneValue(
        Arrays.asList(ExperimentStateStatus.DONE_OK,
          ExperimentStateStatus.DONE_ERROR),
        EXPERIMENT_STATE_ABORT_TIMEOUT);
    }

    experimentStatePv.setValueAndWait(desiredState);

    final List<ExperimentStateStatus> statuses = Collections.synchronizedList(
      new ArrayList<ExperimentStateStatus>());
    final CountDownLatch statusesLatch = new CountDownLatch(2);
    PropertyChangeListener statusListener = new PropertyChangeListener() {
      public void propertyChange(PropertyChangeEvent event) {
        statuses.add((ExperimentStateStatus)event.getNewValue());
        statusesLatch.countDown();
      }
    };

    try {
      experimentStateStatusPv.addPropertyChangeListener(
        EpicsPv.PROPNAME_VALUE, statusListener);
      experimentStateSetPv.setValue(EpicsUtilities.PROC_FIELD_PUT_VALUE);
      boolean ok = statusesLatch.await(EXPERIMENT_STATE_CHANGE_TIMEOUT,
        TimeUnit.MILLISECONDS);
      if (!ok) throw new ExperimentStateNotAchievedException();
    } finally {
      experimentStateStatusPv.removePropertyChangeListener(
        EpicsPv.PROPNAME_VALUE, statusListener);
    }

    List<ExperimentStateStatus> expectedStatuses = Arrays.asList(
      ExperimentStateStatus.EXECUTING, ExperimentStateStatus.DONE_OK);
    synchronized(statuses) {
      if (!expectedStatuses.equals(statuses) ||
          !desiredState.equals(experimentStateRbvPv.getValue()))
        throw new ExperimentStateNotAchievedException();
    }
  }

  protected void cleanUp() throws EpicsCaException, InterruptedException {
    Exception caughtException = null;

    removeDataPointListenerFromScan();

    try {
      scan.stop();
    } catch (Exception e) {
      if (caughtException == null) caughtException = e;
    }

    try {
      stopEnergyMove();
    } catch (Exception e) {
      if (caughtException == null) caughtException = e;
    }

    if (originalAutoAttenuatorEnabled) {
      try {
        enableAutoAttenuator();
      } catch (Exception e) {
        if (caughtException == null) caughtException = e;
      }
    }

    if (Resources.getDefault().isXFlashConfigSupported()) {
      try {
          xFlashConfigCoolingModePv.setValue(originalXFlashConfigCoolingMode);
      } catch (Exception e) {
        if (caughtException == null) caughtException = e;
      }

      try {
        /* Make the cooling mode readback PV update to the new mode. */
        xFlashConfigPollRatePv.setValue(EpicsMenuScan.POINT_ONE_SECOND.asEpicsShort());
        Thread.sleep(500);
      } catch (Exception e) {
        if (caughtException == null) caughtException = e;
      }

      try {
        xFlashConfigPollRatePv.setValue(originalXFlashConfigPollRate);
      } catch (Exception e) {
        if (caughtException == null) caughtException = e;
      }
    }

    try {
      enterExperimentState(ExperimentState.IDLE);
    } catch (Exception e) {
      if (caughtException == null) caughtException = e;
    }

    try {
      enableAutobp();
    } catch (Exception e) {
      if (caughtException == null) caughtException = e;
    }

    if (caughtException != null) throw new RuntimeException(caughtException);
  }

  public boolean isRunning() {
    return isRunning;
  }

  protected void setRunning(boolean newValue) {
    boolean oldValue = isRunning;
    isRunning = newValue;
    if (oldValue != newValue)
      firePropertyChange(PROPNAME_RUNNING, Boolean.valueOf(oldValue),
        Boolean.valueOf(newValue));
  }

  public String getStatus() {
    return status;
  }

  protected void setStatus(String newRawValue) {
    String newValue = newRawValue == null ? "" : newRawValue;
    String oldValue = status;
    status = newValue;
    if (!oldValue.equals(newValue))
      firePropertyChange(PROPNAME_STATUS, oldValue, newValue);
  }

  public XYDataset getXYDataset() {
    return xySeriesCollection;
  }

  protected void addDataPointListenerToScan() {
    scan.addDataPointListener(scanDataPointListener);
  }

  protected void removeDataPointListenerFromScan() {
    scan.removeDataPointListener(scanDataPointListener);
  }

  protected void configureDataOutputFile() throws IOException {
    setDataOutputFile(createNewDataOutputFile());
  }

  protected void setDataOutputFile(File newValue) {
    if (!newValue.isFile())
      throw new IllegalArgumentException("Output file \"" + newValue +
        "\" is not a file");
    if (!newValue.canWrite())
      throw new IllegalArgumentException("Output file \"" + newValue +
        "\" is not writable");

    File oldValue = outputFile;
    outputFile = newValue;
    if (oldValue != newValue)
      firePropertyChange(PROPNAME_OUTPUT_FILE, oldValue, newValue);
  }

  public File getDataOutputFile() {
    return outputFile;
  }

  protected File createNewDataOutputFile() throws IOException {
    MadFluorescenceScanDataFileFilter fileFilter =
      new MadFluorescenceScanDataFileFilter(edge);
    File files[] = outputDirectory.listFiles(fileFilter);
    MadFluorescenceScanDataFileComparator fileComparator =
      new MadFluorescenceScanDataFileComparator();
    Arrays.sort(files, fileComparator);
    File newFile;
    if (files.length == 0) {
      newFile = MadFluorescenceScanDataFileFilter.constructNewFile(
        outputDirectory, edge, 1);
    } else {
      try {
        newFile = MadFluorescenceScanDataFileFilter.constructNewFileAfter(
          outputDirectory, files[files.length - 1]);
      } catch (ParseException e) {
        throw new ProgrammerError("should have been able to parse file " +
          "name: " + files[files.length - 1].getName(), e);
      }
    }
    newFile.createNewFile();
    return newFile;
  }

  protected void configureScan(List<Double> energyPositions)
  throws InterruptedException, EpicsCaException {
    mcaRoiLowPv.setValueAndWait(edge.getLowChannel() + 1);
    mcaRoiHighPv.setValueAndWait(edge.getHighChannel() + 1);
    mcaAcquisitionTimePv.setValueAndWait(scanType.getAcquisitionTime());
    if (energyAxisAutoStartPv != null)
      energyAxisAutoStartPv.setValueAndWait(true);
    scanResetter.reset();
    scan.setBeforeScanPvLink(
      Resources.getDefault().getFluorescenceScanBeforeScanPvName());
    scan.setBeforeScanCommandData(
      Resources.getDefault().getFluorescenceScanBeforeScanValue());
    scan.setAfterScanPvLink(
      Resources.getDefault().getFluorescenceScanAfterScanPvName());
    scan.setAfterScanCommandData(
      Resources.getDefault().getFluorescenceScanAfterScanValue());
    scan.setAfterScanShouldWait(false);
    scan.setPositionerPvLink(0,
      Resources.getDefault().getDesiredEnergyPositionPvName());
    scan.setPositionerReadbackPvLink(0,
      Resources.getDefault().getEnergyPositionPvName());
    scan.setPositionerPositions(0, energyPositions);
    scan.setDetectorTriggerPvLink(0,
      Resources.getDefault().getFluorescenceAcquirePvName());
    scan.setDetectorTriggerCommandData(0, 1.0f);
    scan.setDetectorPvLink(0,
      Resources.getDefault().getFluorescenceIntensityPvName());
    scan.setDetectorPvLink(1,
      Resources.getDefault().getFluorescenceMcaRoiSumPvName());
  }

  protected List<Double> createEnergyPositions() {
    ArrayList<Double> positions = new ArrayList<Double>();
    double edgeEnergy = edge.getEnergy();
    for (double position = edgeEnergy + scanType.getFirstRegionBoundary();
        position <= edgeEnergy + scanType.getLastRegionBoundary();
        position += scanType.getStepSize(position, edgeEnergy))
      positions.add(position / energyAxisToEvScaleFactor);
    return positions;
  }

  protected void receivedScanDataPoint(EpicsScanDataPointEvent event) {
    if (event.isReset()) {
      setScanDataPointsCount(0);
      measuredXYSeries.clear();
    } else if (event.isEndOfData()) {
      setScanDataPointsCount(getNumScanDataPoints());
      removeDataPointListenerFromScan();
      setMeasuredDataSeries(
        scaleAll(scan.getPositionerReadbackPositions(0),
          energyAxisToEvScaleFactor),
        scan.getDetectorData(0), scan.getDetectorData(1));
      if (shouldSaveDataToFile) {
        try {
          saveDataToFile();
        } catch (IOException e) {
          LOG.error("Failed to save scan data file", e);
          JOptionPane.showMessageDialog(UiUtilities.getRootFrame(),
            "Could not save scan data file.\n\n" +
            StringUtilities.wordWrap("Check that the scan output location " +
              "is set correctly, that you have write permissions to the " +
              "location, and that the disk is not full.", 78, "\n"),
            ImcabcFrame.FRAME_TITLE, JOptionPane.ERROR_MESSAGE);
        }
      }
    } else if (event.isPoint()) {
      try {
        measuredXYSeries.add(
          event.getPositionerReadbackValues().get(0) *
            energyAxisToEvScaleFactor,
          new Float(event.getDetectorValues().get(1).floatValue() /
            event.getDetectorValues().get(0).floatValue()));
        incrementScanDataPointsCount();
      } catch (SeriesException e) {
        LOG.warn("Duplicate data point X value");
      }
    }
  }

  protected synchronized int getNumScanDataPoints() {
    return numScanDataPoints;
  }

  protected synchronized void setNumScanDataPoints(int numDataPoints) {
    numScanDataPoints = numDataPoints;
  }

  protected void incrementScanDataPointsCount() {
    setScanDataPointsCount(scanDataPointsCount + 1);
  }

  protected void setScanDataPointsCount(int count) {
    int numPoints = getNumScanDataPoints();
    scanDataPointsCount = Math.min(count, numPoints);
    setRunningStatusAndProgress(scanDataPointsCount, numPoints);
  }

  protected synchronized void setRunningStatusAndProgress(int count,
      int total) {
    if (!isOkToSetRunningStatusAndProgress()) return;
    setStatus(String.format(RUNNING_FORMAT, count, total));
    setProgress(total == 0 ?
      0 : Math.round(((float)count / (float)total) * 100.0f));
  }

  protected synchronized void setOkToSetRunningStatusAndProgress(boolean ok) {
    okToSetRunningStatusAndProgress = ok;
  }

  protected synchronized boolean isOkToSetRunningStatusAndProgress() {
    return okToSetRunningStatusAndProgress;
  }

  protected void saveDataToFile() throws IOException {
    MadFluorescenceScanDataWriter writer = new MadFluorescenceScanDataWriter(
      getDataOutputFile());
    writer.write(edge,
      scaleAll(scan.getPositionerReadbackPositions(0),
        energyAxisToEvScaleFactor),
      scan.getDetectorData(0), scan.getDetectorData(1));
    writer.close();
  }

  /*
   * energyData is in eV.
   */
  protected void setMeasuredDataSeries(List<Double> energyData,
      List<Float> intensityData, List<Float> fluorescenceData) {
    measuredXYSeries.clear();
    Iterator<Double> iX = energyData.iterator();
    Iterator<Float> iY0 = intensityData.iterator();
    Iterator<Float> iY1 = fluorescenceData.iterator();
    while (iX.hasNext() && iY0.hasNext() && iY1.hasNext()) {
      try {
        measuredXYSeries.add(iX.next(),
          new Float(iY1.next().floatValue() / iY0.next().floatValue()),
          false);
      } catch (SeriesException e) {
        LOG.error("Duplicate data point X value");
      }
    }
    measuredXYSeries.fireSeriesChanged();
  }

  /*
   * Returns data if scaleFactor == 1.0; otherwise returns a new List
   * containing all the values of data multiplied by scaleFactor.
   */
  protected List<Double> scaleAll(List<Double> data, double scaleFactor) {
    if (scaleFactor == 1.0) return data;
    List<Double> scaledData = new ArrayList<Double>(data.size());
    for (Double each : data) scaledData.add(each * scaleFactor);
    return scaledData;
  }
}
