package imcacat.imcabc;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;

import imcacat.epicsca4j.bean.ExternallyFireablePropertyChangeBean;
import imcacat.epicsca4j.bean.OneWayPropertyChangeMediator;
import imcacat.epicsca4j.ca.EpicsCaException;
import imcacat.epicsca4j.convert.ShortToEnumEpicsCaValueConverter;
import imcacat.epicsca4j.model.EpicsPv;
import imcacat.epicsca4j.util.EpicsUtilities;

import com.jgoodies.binding.beans.Model;

/**
 * I control a GM/CA-CAT mini-beam apparatus via J. Lewis Muir's minibeam
 * EPICS driver.
 */
public class Minibeam extends Model
implements ExternallyFireablePropertyChangeBean {
  private static final long serialVersionUID = 1;

  public static final String PROPNAME_APERTURE = "aperture";
  public static final String PROPNAME_APERTURE_RBV = "apertureRbv";
  public static final String PROPNAME_APERTURE_SIZE = "apertureSize";
  public static final String PROPNAME_APERTURE_MOVING = "apertureMoving";
  public static final String PROPNAME_APERTURE_MOTION_COMMAND =
    "apertureMotionCommand";

  public static final int NUM_APERTURES = 4;

  protected EpicsPv aperture;
  protected OneWayPropertyChangeMediator apertureMediator;

  protected EpicsPv apertureRbv;
  protected OneWayPropertyChangeMediator apertureRbvMediator;

  protected EpicsPv apertureMove;
  protected EpicsPv apertureAbort;

  protected EpicsPv[] apertureSizePvs;
  protected OneWayPropertyChangeMediator[] apertureSizeMediators;

  protected boolean apertureMoving;

  public Minibeam(String basePvName) {
    super();

    aperture = new EpicsPv(basePvName + "Aperture.VAL", false, true,
      new ShortToEnumEpicsCaValueConverter(MinibeamAperture.class));
    apertureMediator = new OneWayPropertyChangeMediator(aperture, this,
      EpicsPv.PROPNAME_VALUE, PROPNAME_APERTURE);

    apertureRbv = new EpicsPv(basePvName + "ApertureRbv.VAL", false, true,
      new ShortToEnumEpicsCaValueConverter(MinibeamApertureRbv.class));
    apertureRbvMediator = new OneWayPropertyChangeMediator(apertureRbv, this,
      EpicsPv.PROPNAME_VALUE, PROPNAME_APERTURE_RBV);

    apertureRbv.addPropertyChangeListener(EpicsPv.PROPNAME_VALUE,
      new PropertyChangeListener() {
        public void propertyChange(PropertyChangeEvent event) {
          apertureRbvChanged((MinibeamApertureRbv)event.getNewValue());
        }
    });

    apertureMove = new EpicsPv(basePvName + "ApertureMove.PROC", false,
      false);
    apertureAbort = new EpicsPv(basePvName + "ApertureAbort.PROC", false,
      false);

    apertureSizePvs = new EpicsPv[NUM_APERTURES];
    apertureSizeMediators = new OneWayPropertyChangeMediator[NUM_APERTURES];
    for (int i = 0; i < NUM_APERTURES; i++) {
      EpicsPv apertureSize = new EpicsPv(basePvName + "Ap" + (i + 1) +
        "Size.VAL");
      apertureSizePvs[i] = apertureSize;
      apertureSizeMediators[i] = new OneWayPropertyChangeMediator(
        apertureSize, this, EpicsPv.PROPNAME_VALUE,
        Minibeam.PROPNAME_APERTURE_SIZE, null, i);
    }
  }

  public MinibeamAperture getAperture()
  throws InterruptedException, EpicsCaException {
    return (MinibeamAperture)aperture.getValue();
  }

  public void setAperture(MinibeamAperture newAperture)
  throws InterruptedException, EpicsCaException {
    aperture.setValue(newAperture);
  }

  public MinibeamApertureRbv getApertureRbv()
  throws InterruptedException, EpicsCaException {
    return (MinibeamApertureRbv)apertureRbv.getValue();
  }

  public Double getApertureSize(int index)
  throws InterruptedException, EpicsCaException {
    return (Double)apertureSizePvs[index].getValue();
  }

  public boolean isApertureMoving() {
    return apertureMoving;
  }

  protected void setApertureMoving(boolean moving) {
    boolean oldApertureMoving = apertureMoving;
    apertureMoving = moving;
    if (oldApertureMoving == apertureMoving) return;
    firePropertyChange(PROPNAME_APERTURE_MOVING, oldApertureMoving,
      apertureMoving);
  }

  public void setApertureMotionCommand(MinibeamApertureMotionCommand command)
  throws InterruptedException, EpicsCaException {
    if (command == MinibeamApertureMotionCommand.MOVE) moveAperture();
    else if (command == MinibeamApertureMotionCommand.ABORT) abortAperture();
    else throw new RuntimeException("Unexpected command: " + command);
  }

  public void moveAperture() throws InterruptedException, EpicsCaException {
    apertureMove.setValue(EpicsUtilities.PROC_FIELD_PUT_VALUE);
  }

  public void abortAperture() throws InterruptedException, EpicsCaException {
    apertureAbort.setValue(EpicsUtilities.PROC_FIELD_PUT_VALUE);
  }

  protected void apertureRbvChanged(MinibeamApertureRbv newApertureRbv) {
    setApertureMoving(newApertureRbv == MinibeamApertureRbv.MOVING);
  }

  public void externallyFirePropertyChange(PropertyChangeEvent event) {
    firePropertyChange(event);
  }
}
