package imcacat.imcabc;

import imcacat.epicsca4j.ca.EpicsCaException;
import imcacat.epicsca4j.model.EpicsMotor;
import imcacat.epicsca4j.util.StringUtilities;
import imcacat.epicsca4j.util.UiUtilities;

import java.lang.reflect.InvocationTargetException;
import java.text.DecimalFormat;

import javax.swing.JOptionPane;

import gov.aps.jca.TimeoutException;

/**
 * I manually optimize the vertical focusing mirror Bend.
 */
public class MirrorBendManualOptimization17bm extends Optimization17bm {
  protected static final long BEND_MOVE_TIMEOUT = 60000;

  protected EpicsMotor bendMotor;

  protected double originalBendMotorPosition;

  public MirrorBendManualOptimization17bm() {
    super();
  }

  @Override
  protected void initialize()
  throws InterruptedException, EpicsCaException {
    super.initialize();
    bendMotor = new EpicsMotor(
      Resources.getDefault().getMirrorBendMotorPvName(), "\u03BCm");
    originalBendMotorPosition = bendMotor.getPosition();
  }

  @Override
  public void release() {
    if (bendMotor != null) bendMotor.release();
    super.release();
  }

  @Override
  protected void go()
  throws InterruptedException, TimeoutException, EpicsCaException {
    setStatus("Configuring beamline for Bend optimization");
    disableAutobp();
    setProgress(10);

    setStatus("Scanning Bend");
    scanBend();
    setProgress(60);

    setStatus("Fitting Bend scan data to Gaussian function");
    GaussianFunctionCurveFitter fitter;
    try {
      fitter = new GaussianFunctionCurveFitter(measuredXYSeries.toArray());
    } catch (Exception e) {
      throw new RuntimeException("Failed to fit measured data to Gaussian " +
        "function", e);
    }
    try {
      setFitDataSeriesInEventDispatchThreadAndWait(fitter);
    } catch (InvocationTargetException e) {
      throw new RuntimeException("Failed to add fit data to " +
        "XYSeriesCollection", e);
    }
    setProgress(70);

    double measuredMax = getXWithMaximumY(measuredXYSeries);
    double fitMax = fitter.getC();

    setStatus("Waiting for measured or fit choice");
    DecimalFormat positionFormat = new DecimalFormat("0.00");
    int choice = JOptionPane.showOptionDialog(UiUtilities.getRootFrame(),
      StringUtilities.wordWrap("The Bend scan has finished. The measured " +
        "maximum is at " +
        positionFormat.format(measuredMax) + " " + getBendUnitOfMeasure() +
        ", and the fit maximum is at " +
        positionFormat.format(fitMax) + " " + getBendUnitOfMeasure() +
        ". To which maximum would you like to move Bend?",
        78, "\n"),
      "Move to measured or fit?", JOptionPane.YES_NO_CANCEL_OPTION,
      JOptionPane.QUESTION_MESSAGE, null, new String[] {"Measured", "Fit",
      "Cancel"}, "Fit");
    setProgress(75);

    if (choice == JOptionPane.YES_OPTION) {
      setStatus("Moving Bend to measured maximum");
      moveBend(measuredMax);
    } else if (choice == JOptionPane.NO_OPTION) {
      setStatus("Moving Bend to fit maximum");
      moveBend(fitMax);
    } else {
      setStatus("Canceling");
      moveBend(originalBendMotorPosition);
      enableAutobp();
      setStatus("Canceled");
      return;
    }
    setProgress(90);

    setStatus("Configuring beamline for use");
    enableAutobp();

    setProgressAndStatusToComplete();
  }

  @Override
  public void cleanUp() {
    Exception caughtException = null;

    try {
      scan.stop();
    } catch (Exception e) {
      if (caughtException == null) caughtException = e;
    }
    try {
      bendMotor.stop();
    } catch (Exception e) {
      if (caughtException == null) caughtException = e;
    }
    try {
      bendMotor.setDesiredPosition(originalBendMotorPosition);
      bendMotor.moveAndWait(BEND_MOVE_TIMEOUT);
    } catch (Exception e) {
      if (caughtException == null) caughtException = e;
    }
    try {
      enableAutobp();
    } catch (Exception e) {
      if (caughtException == null) caughtException = e;
    }

    try {
      super.cleanUp();
    } catch (Exception e) {
      if (caughtException == null) caughtException = e;
    }

    if (caughtException != null) throw new RuntimeException(caughtException);
  }

  protected void scanBend()
  throws EpicsCaException, InterruptedException {
    performBendScan();
    if (!scan.completedSuccessfully())
      throw new RuntimeException("Bend scan did not complete successfully");
    try {
      setMeasuredDataSeriesInEventDispatchThreadAndWait(getData(scan));
    } catch (InvocationTargetException e) {
      throw new RuntimeException("Failed to update measured data series " +
        "with complete data from scan", e);
    }
  }

  protected void performBendScan() throws EpicsCaException,
  InterruptedException {
    setXYDatasetTitle("Bend optimization scan");
    setXDataLabel("Bend (" + getBendUnitOfMeasure() + ")");
    setYDataLabel("Small Ion Chamber intensity (counts)");
    scanResetter.reset();
    scan.setBeforeScanPvLink(
      Resources.getDefault().
        getSmallIonChamberIntensityAcquisitionTimePvName());
    scan.setBeforeScanCommandData(1.0f);
    scan.setPositionerPvLink(0,
      Resources.getDefault().getDesiredMirrorBendPositionPvName());
    scan.setPositionerReadbackPvLink(0,
      Resources.getDefault().getMirrorBendPositionPvName());
    scan.setPositionerReadbackDelta(0, 10.0);
    scan.setPositionerCenteredLinearMode(0, getBendPosition(), 10.0, 18);
    scan.setPositionerSettlingTime(1.0f);
    scan.setDetectorTriggerPvLink(0,
      Resources.getDefault().
        getSmallIonChamberIntensityAcquisitionStartPvName());
    scan.setDetectorTriggerCommandData(0, 1.0f);
    scan.setDetectorPvLink(0,
      Resources.getDefault().
        getSmallIonChamberIntensityAcquisitionSumPvName());
    openExposureShutter();
    scan.go();
    scan.waitToFinish();
    closeExposureShutter();
  }

  protected double getBendPosition()
  throws InterruptedException, EpicsCaException {
    return bendMotor.getPosition();
  }

  protected String getBendUnitOfMeasure() {
    return bendMotor.getUnitOfMeasure();
  }

  protected void moveBend(double desiredPosition)
  throws InterruptedException, TimeoutException, EpicsCaException {
    bendMotor.setDesiredPosition(desiredPosition);
    bendMotor.moveAndWait(BEND_MOVE_TIMEOUT);
  }
}
