package imcacat.imcabc;

import imcacat.epicsca4j.ca.EpicsCaException;
import imcacat.epicsca4j.model.EpicsMotor;
import imcacat.epicsca4j.util.StringUtilities;
import imcacat.epicsca4j.util.UiUtilities;

import java.lang.reflect.InvocationTargetException;
import java.text.DecimalFormat;

import javax.swing.JOptionPane;

import gov.aps.jca.TimeoutException;

/**
 * I manually optimize the vertical focusing mirror Vertical Jack.
 */
public class MirrorVerticalJackManualOptimization17bm
extends Optimization17bm {
  protected static final long VERTICAL_JACK_MOVE_TIMEOUT = 60000;

  protected EpicsMotor verticalJackMotor;

  protected double originalVerticalJackMotorPosition;

  public MirrorVerticalJackManualOptimization17bm() {
    super();
  }

  @Override
  protected void initialize()
  throws InterruptedException, EpicsCaException {
    super.initialize();
    verticalJackMotor = new EpicsMotor(
      Resources.getDefault().getMirrorVerticalJackMotorPvName(),
      "\u03BCm");
    originalVerticalJackMotorPosition = verticalJackMotor.getPosition();
  }

  @Override
  public void release() {
    if (verticalJackMotor != null) verticalJackMotor.release();
    super.release();
  }

  @Override
  protected void go()
  throws InterruptedException, TimeoutException, EpicsCaException {
    setStatus("Configuring beamline for Vertical Jack optimization");
    disableAutobp();
    setProgress(10);

    setStatus("Scanning Vertical Jack");
    scanVerticalJack();
    setProgress(60);

    setStatus("Fitting Vertical Jack scan data to Gaussian function");
    GaussianFunctionCurveFitter fitter;
    try {
      fitter = new GaussianFunctionCurveFitter(measuredXYSeries.toArray());
    } catch (Exception e) {
      throw new RuntimeException("Failed to fit measured data to Gaussian " +
        "function", e);
    }
    try {
      setFitDataSeriesInEventDispatchThreadAndWait(fitter);
    } catch (InvocationTargetException e) {
      throw new RuntimeException("Failed to add fit data to " +
        "XYSeriesCollection", e);
    }
    setProgress(70);

    double measuredMax = getXWithMaximumY(measuredXYSeries);
    double fitMax = fitter.getC();

    setStatus("Waiting for measured or fit choice");
    DecimalFormat positionFormat = new DecimalFormat("0.00");
    int choice = JOptionPane.showOptionDialog(UiUtilities.getRootFrame(),
      StringUtilities.wordWrap("The Vertical Jack scan has finished. The " +
        "measured maximum is at " +
        positionFormat.format(measuredMax) + " " +
        getVerticalJackUnitOfMeasure() +
        ", and the fit maximum is at " +
        positionFormat.format(fitMax) + " " + getVerticalJackUnitOfMeasure() +
        ". To which maximum would you like to move Vertical Jack?",
        78, "\n"),
      "Move to measured or fit?", JOptionPane.YES_NO_CANCEL_OPTION,
      JOptionPane.QUESTION_MESSAGE, null, new String[] {"Measured", "Fit",
      "Cancel"}, "Fit");
    setProgress(75);

    if (choice == JOptionPane.YES_OPTION) {
      setStatus("Moving Vertical Jack to measured maximum");
      moveVerticalJack(measuredMax);
    } else if (choice == JOptionPane.NO_OPTION) {
      setStatus("Moving Vertical Jack to fit maximum");
      moveVerticalJack(fitMax);
    } else {
      setStatus("Canceling");
      moveVerticalJack(originalVerticalJackMotorPosition);
      enableAutobp();
      setStatus("Canceled");
      return;
    }
    setProgress(90);

    setStatus("Configuring beamline for use");
    enableAutobp();

    setProgressAndStatusToComplete();
  }

  @Override
  public void cleanUp() {
    Exception caughtException = null;

    try {
      scan.stop();
    } catch (Exception e) {
      if (caughtException == null) caughtException = e;
    }
    try {
      verticalJackMotor.stop();
    } catch (Exception e) {
      if (caughtException == null) caughtException = e;
    }
    try {
      verticalJackMotor.setDesiredPosition(originalVerticalJackMotorPosition);
      verticalJackMotor.moveAndWait(VERTICAL_JACK_MOVE_TIMEOUT);
    } catch (Exception e) {
      if (caughtException == null) caughtException = e;
    }
    try {
      enableAutobp();
    } catch (Exception e) {
      if (caughtException == null) caughtException = e;
    }

    try {
      super.cleanUp();
    } catch (Exception e) {
      if (caughtException == null) caughtException = e;
    }

    if (caughtException != null) throw new RuntimeException(caughtException);
  }

  protected void scanVerticalJack()
  throws EpicsCaException, InterruptedException {
    performVerticalJackScan();
    if (!scan.completedSuccessfully())
      throw new RuntimeException("Vertical Jack scan did not complete " +
        "successfully");
    try {
      setMeasuredDataSeriesInEventDispatchThreadAndWait(getData(scan));
    } catch (InvocationTargetException e) {
      throw new RuntimeException("Failed to update measured data series " +
        "with complete data from scan", e);
    }
  }

  protected void performVerticalJackScan() throws EpicsCaException,
  InterruptedException {
    setXYDatasetTitle("Vertical Jack optimization scan");
    setXDataLabel("Vertical Jack (" + getVerticalJackUnitOfMeasure() + ")");
    setYDataLabel("BPM intensity (counts)");
    scanResetter.reset();
    scan.setBeforeScanPvLink(
      Resources.getDefault().getBpmIntensityAcquisitionTimePvName());
    scan.setBeforeScanCommandData(1.0f);
    scan.setPositionerPvLink(0,
      Resources.getDefault().
        getDesiredMirrorVerticalJackPositionPvName());
    scan.setPositionerReadbackPvLink(0,
      Resources.getDefault().getMirrorVerticalJackPositionPvName());
    scan.setPositionerReadbackDelta(0, 5.0);
    scan.setPositionerCenteredLinearMode(0, getVerticalJackPosition(), 100.0,
      31);
    scan.setPositionerSettlingTime(1.0f);
    scan.setDetectorTriggerPvLink(0,
      Resources.getDefault().getBpmIntensityAcquisitionStartPvName());
    scan.setDetectorTriggerCommandData(0, 1.0f);
    scan.setDetectorPvLink(0,
      Resources.getDefault().getBpmIntensityAcquisitionSumPvName());
    scan.go();
    scan.waitToFinish();
  }

  protected double getVerticalJackPosition()
  throws InterruptedException, EpicsCaException {
    return verticalJackMotor.getPosition();
  }

  protected String getVerticalJackUnitOfMeasure() {
    return verticalJackMotor.getUnitOfMeasure();
  }

  protected void moveVerticalJack(double desiredPosition)
  throws InterruptedException, TimeoutException, EpicsCaException {
    verticalJackMotor.setDesiredPosition(desiredPosition);
    verticalJackMotor.moveAndWait(VERTICAL_JACK_MOVE_TIMEOUT);
  }
}
