package imcacat.imcabc;

import imcacat.epicsca4j.ca.EpicsCaException;
import imcacat.epicsca4j.model.EpicsMotor;
import imcacat.epicsca4j.util.StringUtilities;
import imcacat.epicsca4j.util.UiUtilities;

import java.lang.reflect.InvocationTargetException;
import java.text.DecimalFormat;

import javax.swing.JOptionPane;

import gov.aps.jca.TimeoutException;

/**
 * I manually optimize Omega.
 */
public class OmegaManualOptimization17bm extends Optimization17bm {
  protected static final long OMEGA_MOVE_TIMEOUT = 60000;

  protected EpicsMotor omegaMotor;

  protected double originalOmegaMotorPosition;

  public OmegaManualOptimization17bm() {
    super();
  }

  @Override
  protected void initialize()
  throws InterruptedException, EpicsCaException {
    super.initialize();
    omegaMotor = new EpicsMotor(
      Resources.getDefault().getOmegaMotorPvName(), "\u00B5rad");
    originalOmegaMotorPosition = omegaMotor.getPosition();
  }

  @Override
  public void release() {
    if (omegaMotor != null) omegaMotor.release();
    super.release();
  }

  @Override
  protected void go()
  throws InterruptedException, TimeoutException, EpicsCaException {
    setStatus("Configuring beamline for Omega optimization");
    disableAutobp();
    setProgress(10);

    setStatus("Scanning Omega");
    scanOmega();
    setProgress(60);

    setStatus("Fitting Omega scan data to Gaussian function");
    GaussianFunctionCurveFitter fitter;
    try {
      fitter = new GaussianFunctionCurveFitter(measuredXYSeries.toArray());
    } catch (Exception e) {
      throw new RuntimeException("Failed to fit measured data to Gaussian " +
        "function", e);
    }
    try {
      setFitDataSeriesInEventDispatchThreadAndWait(fitter);
    } catch (InvocationTargetException e) {
      throw new RuntimeException("Failed to add fit data to " +
        "XYSeriesCollection", e);
    }
    setProgress(70);

    double measuredMax = getXWithMaximumY(measuredXYSeries);
    double fitMax = fitter.getC();

    setStatus("Waiting for measured or fit choice");
    DecimalFormat positionFormat = new DecimalFormat("0.00");
    int choice = JOptionPane.showOptionDialog(UiUtilities.getRootFrame(),
      StringUtilities.wordWrap("The Omega scan has finished. The measured " +
        "maximum is at " +
        positionFormat.format(measuredMax) + " " + getOmegaUnitOfMeasure() +
        ", and the fit maximum is at " +
        positionFormat.format(fitMax) + " " + getOmegaUnitOfMeasure() +
        ". To which maximum would you like to move Omega?",
        78, "\n"),
      "Move to measured or fit?", JOptionPane.YES_NO_CANCEL_OPTION,
      JOptionPane.QUESTION_MESSAGE, null, new String[] {"Measured", "Fit",
      "Cancel"}, "Fit");
    setProgress(75);

    if (choice == JOptionPane.YES_OPTION) {
      setStatus("Moving Omega to measured maximum");
      moveOmega(measuredMax);
    } else if (choice == JOptionPane.NO_OPTION) {
      setStatus("Moving Omega to fit maximum");
      moveOmega(fitMax);
    } else {
      setStatus("Canceling");
      moveOmega(originalOmegaMotorPosition);
      enableAutobp();
      setStatus("Canceled");
      return;
    }
    setProgress(90);

    setStatus("Configuring beamline for use");
    enableAutobp();

    setProgressAndStatusToComplete();
  }

  @Override
  public void cleanUp() {
    Exception caughtException = null;

    try {
      scan.stop();
    } catch (Exception e) {
      if (caughtException == null) caughtException = e;
    }
    try {
      omegaMotor.stop();
    } catch (Exception e) {
      if (caughtException == null) caughtException = e;
    }
    try {
      omegaMotor.setDesiredPosition(originalOmegaMotorPosition);
      omegaMotor.moveAndWait(OMEGA_MOVE_TIMEOUT);
    } catch (Exception e) {
      if (caughtException == null) caughtException = e;
    }
    try {
      enableAutobp();
    } catch (Exception e) {
      if (caughtException == null) caughtException = e;
    }

    try {
      super.cleanUp();
    } catch (Exception e) {
      if (caughtException == null) caughtException = e;
    }

    if (caughtException != null) throw new RuntimeException(caughtException);
  }

  protected void scanOmega()
  throws EpicsCaException, InterruptedException {
    performOmegaScan();
    if (!scan.completedSuccessfully())
      throw new RuntimeException("Omega scan did not complete successfully");
    try {
      setMeasuredDataSeriesInEventDispatchThreadAndWait(getData(scan));
    } catch (InvocationTargetException e) {
      throw new RuntimeException("Failed to update measured data series " +
        "with complete data from scan", e);
    }
  }

  protected void performOmegaScan() throws EpicsCaException,
  InterruptedException {
    setXYDatasetTitle("Omega optimization scan");
    setXDataLabel("Omega (" + getOmegaUnitOfMeasure() + ")");
    setYDataLabel("BPM intensity (counts)");
    scanResetter.reset();
    scan.setBeforeScanPvLink(
      Resources.getDefault().getBpmIntensityAcquisitionTimePvName());
    scan.setBeforeScanCommandData(1.0f);
    scan.setPositionerPvLink(0,
      Resources.getDefault().getDesiredOmegaPositionPvName());
    scan.setPositionerReadbackPvLink(0,
      Resources.getDefault().getOmegaPositionPvName());
    scan.setPositionerReadbackDelta(0, 2.0);
    scan.setPositionerCenteredLinearMode(0, getOmegaPosition(), 5.0, 20);
    scan.setPositionerSettlingTime(2.0f);
    scan.setDetectorTriggerPvLink(0,
      Resources.getDefault().getBpmIntensityAcquisitionStartPvName());
    scan.setDetectorTriggerCommandData(0, 1.0f);
    scan.setDetectorPvLink(0,
      Resources.getDefault().getBpmIntensityAcquisitionSumPvName());
    scan.go();
    scan.waitToFinish();
  }

  protected double getOmegaPosition()
  throws InterruptedException, EpicsCaException {
    return omegaMotor.getPosition();
  }

  protected String getOmegaUnitOfMeasure() {
    return omegaMotor.getUnitOfMeasure();
  }

  protected void moveOmega(double desiredPosition)
  throws InterruptedException, TimeoutException, EpicsCaException {
    omegaMotor.setDesiredPosition(desiredPosition);
    omegaMotor.moveAndWait(OMEGA_MOVE_TIMEOUT);
  }
}
