package imcacat.imcabc;

import imcacat.epicsca4j.ca.EpicsCaException;
import imcacat.epicsca4j.util.StringUtilities;

import java.awt.Frame;
import java.awt.event.ActionEvent;

import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.JOptionPane;

import org.apache.log4j.Logger;

/**
 * I open an {@link AutoOptimizationDetailsDialog} when I am performed.
 * <p>
 * Subclasses should implement my <code>createDialog</code> method to create
 * the dialog instance to be opened.
 * <p>
 * After the user dismisses the dialog, I release the view, presentation
 * model, and model by doing this (<code>dialog</code> is the object returned
 * by the <code>createDialog</code> method):
 * <pre>
 *   AutoOptimizationDetails model = dialog.getPresentationModel().getModel();
 *   dialog.getPresentationModel().release();
 *   model.release();
 * </pre>
 */
public class OpenAutoOptimizationDetailsDialogAction
extends AbstractAction {
  private static final long serialVersionUID = 1;

  private static final Logger LOG = Logger.getLogger(
    OpenAutoOptimizationDetailsDialogAction.class);

  protected String dialogTitle;
  protected Frame ownerFrame;

  public OpenAutoOptimizationDetailsDialogAction(Frame ownerFrame,
      String name, String dialogTitle) {
    super();
    this.ownerFrame = ownerFrame;
    this.dialogTitle = dialogTitle;
    putValue(Action.NAME, name);
  }

  public void actionPerformed(ActionEvent event) {
    /*
     * I run the dialog creation in a different Thread because it will fail
     * if the EPICS events are being dispatched on the Event Dispatch Thread.
     * This is because this method is usually called from the Event Dispatch
     * Thread and so EPICS CA connection and value events will not be received
     * when the model, containing EpicsPv instances or other
     * EPICS CA related instances, is created.
     */
    Runnable runnable = new Runnable() {
      public void run() {
        AutoOptimizationDetailsDialog dialog;
        try {
          dialog = createDialog();
        } catch (EpicsCaException e) {
          JOptionPane.showMessageDialog(ownerFrame,
            "Could not display details.\n\n" +
            StringUtilities.wordWrap("An error occurred while " +
              "communicating with the control system.  Please ensure that " +
              "the control system is running and that this computer is " +
              "connected to the control system network.", 78, "\n"),
            ImcabcFrame.FRAME_TITLE, JOptionPane.PLAIN_MESSAGE);
          return;
        } catch (InterruptedException e) {
          LOG.error("Unexpected interrupt", e);
          JOptionPane.showMessageDialog(ownerFrame,
            "Could not display details.\n\n" +
            StringUtilities.wordWrap("An unexpected interrupt error " +
              "occurred that prevented displaying the details view.  It is " +
              "safe to continue using this application.", 78, "\n"),
            ImcabcFrame.FRAME_TITLE, JOptionPane.ERROR_MESSAGE);
          return;
        }
        dialog.pack();
        dialog.setLocationRelativeTo(ownerFrame);
        dialog.setResizable(false);
        dialog.setVisible(true);

        AutoOptimizationDetails model =
          dialog.getPresentationModel().getBean();
        dialog.getPresentationModel().release();
        model.release();
      }
    };
    (new Thread(runnable)).start();
  }

  protected AutoOptimizationDetailsDialog createDialog()
  throws InterruptedException, EpicsCaException {
    return new AutoOptimizationDetailsDialog(ownerFrame, dialogTitle, true);
  }
}
