package imcacat.imcabc;

import gov.aps.jca.TimeoutException;
import imcacat.epicsca4j.ca.EpicsCaException;
import imcacat.epicsca4j.model.EpicsPv;
import imcacat.epicsca4j.model.EpicsScan;
import imcacat.epicsca4j.model.EpicsScanDataPointEvent;
import imcacat.epicsca4j.model.EpicsScanDataPointListener;
import imcacat.epicsca4j.util.ObjectUtilities;
import imcacat.imcabc.util.Pair;

import java.lang.reflect.InvocationTargetException;
import java.util.List;

import javax.swing.SwingUtilities;

import org.apache.log4j.Logger;
import org.jdesktop.swingworker.SwingWorker;
import org.jfree.data.general.SeriesException;
import org.jfree.data.xy.XYDataset;
import org.jfree.data.xy.XYSeries;
import org.jfree.data.xy.XYSeriesCollection;

/**
 * I am an operation that optimizes some property of a beamline.
 * <p>
 * Subclasses should override my <code>initialize</code>, <code>go</code>,
 * <code>cleanUp</code>, and <code>release</code> methods.
 * <p>
 * When my <code>doInBackground</code> method is called, I do the following:
 * <ol>
 * <li>call <code>initialize</code>
 * <li>stop if <code>initialize</code> threw an exception
 * <li>call <code>go</code>
 * <li>call <code>cleanUp</code> if <code>go</code> threw an exception
 * <li>call <code>release</code>
 * </ol>
 */
public abstract class Optimization extends SwingWorker<Void, Void> {
  private static final Logger LOG = Logger.getLogger(Optimization.class);

  protected static final long AUTOBP_REQUEST_PAUSE_TIMEOUT = 185000;

  public static final String MEASURED_SERIES_KEY = "Measured";
  public static final String FIT_SERIES_KEY = "Fit";

  public static final String PROPNAME_PROGRESS = "progress";
  public static final String PROPNAME_RUNNING = "running";
  public static final String PROPNAME_STATUS = "status";
  public static final String PROPNAME_XY_DATASET = "XYDataset";
  public static final String PROPNAME_XY_DATASET_TITLE = "XYDatasetTitle";
  public static final String PROPNAME_X_DATA_LABEL = "XDataLabel";
  public static final String PROPNAME_Y_DATA_LABEL = "YDataLabel";
  public static final String PROPNAME_X_POINT_OF_INTEREST =
    "XPointOfInterest";
  public static final String PROPNAME_Y_MEASURED_POINT_OF_INTEREST =
    "YMeasuredPointOfInterest";
  public static final String PROPNAME_Y_FIT_POINT_OF_INTEREST =
    "YFitPointOfInterest";

  protected XYSeries measuredXYSeries;
  protected XYSeriesCollection xySeriesCollection;

  protected LinearInterpolator measuredXYSeriesInterpolator;
  protected XYFunction fitFunction;

  protected boolean isRunning;
  protected String status;
  protected String xyDatasetTitle;
  protected String xDataLabel;
  protected String yDataLabel;

  protected Double xPointOfInterest;
  protected Double yMeasuredPointOfInterest;
  protected Double yFitPointOfInterest;

  protected EpicsPv autobpRequestExternalPausePv;
  protected EpicsPv autobpReleaseExternalPausePv;
  protected EpicsScan scan;
  protected EpicsScanDataPointListener scanDataPointListener;
  protected EpicsScanResetter scanResetter;

  public Optimization() {
    super();

    measuredXYSeries = new XYSeries(MEASURED_SERIES_KEY, true, false);
    xySeriesCollection = new XYSeriesCollection(measuredXYSeries);

    isRunning = false;
    status = "";
    xyDatasetTitle = "";
    xDataLabel = "";
    yDataLabel = "";
  }

  protected void initialize() throws InterruptedException, EpicsCaException {
    autobpRequestExternalPausePv = new EpicsPv(
      Resources.getDefault().getAutobpRequestExternalPausePvName(), false,
      false);
    autobpReleaseExternalPausePv = new EpicsPv(
      Resources.getDefault().getAutobpReleaseExternalPausePvName(), false,
      false);

    scan = getScanInstance();
    scanDataPointListener = new EpicsScanDataPointListener() {
      public void dataReceived(EpicsScanDataPointEvent event) {
        receivedScanDataPoint(event);
      }
    };

    addDataPointListenerToScan();

    scanResetter = getScanResetterInstance();
  }

  protected abstract EpicsScan getScanInstance()
  throws InterruptedException, EpicsCaException;

  protected abstract EpicsScanResetter getScanResetterInstance();

  public void release() {
    if (scan != null)
      removeDataPointListenerFromScan();
    if (scanResetter != null)
      scanResetter.release();
    if (autobpRequestExternalPausePv != null)
      autobpRequestExternalPausePv.release();
    if (autobpReleaseExternalPausePv != null)
      autobpReleaseExternalPausePv.release();
  }

  public Void doInBackground() throws Exception {
    setRunning(true);

    try {
      setStatus("Performing optimization initialization");
      initialize();
    } catch (Exception e) {
      setStatus("Optimization initialization failed");
      LOG.error("Optimization initialization failed", e);
      setRunning(false);
      return null;
    }

    try {
      go();
    } catch (InterruptedException e) {
      performCleanUp("Stopping", "Stopped", "Stop failed");
    } catch (Exception e) {
      LOG.error("Optimization failed", e);
      performCleanUp("Failed", "Failed", "Failed");
    } finally {
      try {
        release();
      } catch (Exception e) {
        LOG.error("Failed to release Optimization after run", e);
      }
    }

    setRunning(false);
    return null;
  }

  protected void performCleanUp(String beforeCleanUp, String cleanUpSuccess,
      String cleanUpFailure) {
    setStatus(beforeCleanUp + "; restoring good beamline state");
    try {
      cleanUp();
      setStatus(cleanUpSuccess + "; good beamline state restored");
    } catch (Exception e2) {
      setStatus(cleanUpFailure + "; beamline in inconsistent state");
      LOG.error("Failed to perform optimization clean up; beamline in an " +
        "inconsistent state", e2);
    }
  }

  protected abstract void go() throws InterruptedException,
      TimeoutException, EpicsCaException;

  /**
   * This method is invoked if I am interrupted while running. Subclasses may
   * override this method to perform any work that should be done to restore
   * the beamline to a good state.
   */
  protected void cleanUp() {
    removeDataPointListenerFromScan();
  }

  public boolean isRunning() {
    return isRunning;
  }

  public void setRunning(boolean newValue) {
    boolean oldValue = isRunning;
    isRunning = newValue;
    if (oldValue != newValue)
      firePropertyChange(PROPNAME_RUNNING, Boolean.valueOf(oldValue),
        Boolean.valueOf(newValue));
  }

  public Double getXPointOfInterest() {
    return xPointOfInterest;
  }

  public void setXPointOfInterest(Double newValue) {
    Double oldValue = xPointOfInterest;
    xPointOfInterest = newValue;
    if (!ObjectUtilities.areEqual(oldValue, newValue)) {
      firePropertyChange(PROPNAME_X_POINT_OF_INTEREST, oldValue, newValue);
      setYMeasuredPointOfInterest(computeYMeasuredPointOfInterest(newValue));
      setYFitPointOfInterest(computeYFitPointOfInterest(newValue));
    }
  }

  protected synchronized Double computeYMeasuredPointOfInterest(Double x) {
    if (x == null) return null;
    if (measuredXYSeriesInterpolator == null) return null;
    try {
      return measuredXYSeriesInterpolator.getY(x);
    } catch (Exception e) {
      return null;
    }
  }

  protected synchronized Double computeYFitPointOfInterest(Double x) {
    if (x == null) return null;
    if (fitFunction == null) return null;
    try {
      return fitFunction.getY(x);
    } catch (Exception e) {
      return null;
    }
  }

  public Double getYFitPointOfInterest() {
    return yFitPointOfInterest;
  }

  protected void setYFitPointOfInterest(Double newValue) {
    Double oldValue = yFitPointOfInterest;
    yFitPointOfInterest = newValue;
    if (!ObjectUtilities.areEqual(oldValue, newValue))
      firePropertyChange(PROPNAME_Y_FIT_POINT_OF_INTEREST, oldValue,
        newValue);
  }

  public Double getYMeasuredPointOfInterest() {
    return yMeasuredPointOfInterest;
  }

  protected void setYMeasuredPointOfInterest(Double newValue) {
    Double oldValue = yMeasuredPointOfInterest;
    yMeasuredPointOfInterest = newValue;
    if (!ObjectUtilities.areEqual(oldValue, newValue))
      firePropertyChange(PROPNAME_Y_MEASURED_POINT_OF_INTEREST, oldValue,
        newValue);
  }

  public String getStatus() {
    return status;
  }

  protected void setStatus(String newRawValue) {
    String newValue = newRawValue == null ? "" : newRawValue;
    String oldValue = status;
    status = newValue;
    if (!oldValue.equals(newValue))
      firePropertyChange(PROPNAME_STATUS, oldValue, newValue);
  }

  public String getXYDatasetTitle() {
    return xyDatasetTitle;
  }

  protected void setXYDatasetTitle(String newRawValue) {
    String newValue = newRawValue == null ? "" : newRawValue;
    String oldValue = xyDatasetTitle;
    xyDatasetTitle = newValue;
    if (!oldValue.equals(newValue))
      firePropertyChange(PROPNAME_XY_DATASET_TITLE, oldValue, newValue);
  }

  public String getXDataLabel() {
    return xDataLabel;
  }

  protected void setXDataLabel(String newRawValue) {
    String newValue = newRawValue == null ? "" : newRawValue;
    String oldValue = xDataLabel;
    xDataLabel = newValue;
    if (!oldValue.equals(newValue))
      firePropertyChange(PROPNAME_X_DATA_LABEL, oldValue, newValue);
  }

  public String getYDataLabel() {
    return yDataLabel;
  }

  protected void setYDataLabel(String newRawValue) {
    String newValue = newRawValue == null ? "" : newRawValue;
    String oldValue = yDataLabel;
    yDataLabel = newValue;
    if (!oldValue.equals(newValue))
      firePropertyChange(PROPNAME_Y_DATA_LABEL, oldValue, newValue);
  }

  public XYDataset getXYDataset() {
    return xySeriesCollection;
  }

  protected void disableAutobp()
  throws InterruptedException, TimeoutException, EpicsCaException {
    autobpRequestExternalPausePv.setValueAndWait(
      Resources.getDefault().getAutobpRequestExternalPauseValue(),
      AUTOBP_REQUEST_PAUSE_TIMEOUT);
  }

  protected void enableAutobp()
  throws InterruptedException, EpicsCaException {
    autobpReleaseExternalPausePv.setValue(
      Resources.getDefault().getAutobpReleaseExternalPauseValue());
  }

  protected double getXWithMaximumY(XYSeries series) {
    int seriesSize = series.getItemCount();
    int indexWithMaxY = 0;
    double maxY = series.getY(0).doubleValue();
    for (int i = 1; i < seriesSize; i++) {
      if (series.getY(i).doubleValue() > maxY) {
        indexWithMaxY = i;
        maxY = series.getY(i).doubleValue();
      }
    }
    return series.getX(indexWithMaxY).doubleValue();
  }

  /**
   * I invoke <code>setMeasuredDataSeriesInEventDispatchThreadAndWait(List,
   * List)</code>.
   *
   * @param data  the first object contains the X data and the second object
   *              contains the Y data
   *
   * @see #setMeasuredDataSeriesInEventDispatchThreadAndWait(List, List)
   * @see #getData(EpicsScan)
   */
  protected void setMeasuredDataSeriesInEventDispatchThreadAndWait(
    Pair<List<? extends Number>,List<? extends Number>> data)
  throws InterruptedException, InvocationTargetException {
    setMeasuredDataSeriesInEventDispatchThreadAndWait(data.getFirst(),
      data.getSecond());
  }

  /**
   * I create and set a measured XY series interpolator (based on the
   * specified data) used for the measured point of interest computation and
   * set (in the Event Dispatch Thread) the specified data in
   * <code>measuredXYSeries</code>.
   */
  protected void setMeasuredDataSeriesInEventDispatchThreadAndWait(
    List<? extends Number> xValues, List<? extends Number> yValues)
  throws InterruptedException, InvocationTargetException {
    setMeasuredXYSeriesInterpolator(new LinearInterpolator(xValues, yValues));
    SwingUtilities.invokeAndWait(new XYSeriesDataSetter(measuredXYSeries,
      xValues, yValues));
  }

  protected synchronized LinearInterpolator getMeasuredXYSeriesInterpolator()
  {
    return measuredXYSeriesInterpolator;
  }

  protected synchronized void setMeasuredXYSeriesInterpolator(
      LinearInterpolator interpolator) {
    measuredXYSeriesInterpolator = interpolator;
  }

  /**
   * I set the specified <code>dataFitFunction</code> as the fit function used
   * for the fit point of interest calculation and create a fit data series by
   * sampling the fit function and set the created fit data series (in the
   * Event Dispatch Thread) in <code>xySeriesCollection</code> as the fit data
   * for graphing.
   * <p>
   * I sample the fit function on a range determined from the first and last
   * X values in <code>measuredXYSeries</code>. This means that
   * <code>measuredXYSeries</code> should be set with up to date data before I
   * am called.
   */
  protected void setFitDataSeriesInEventDispatchThreadAndWait(
    XYFunction dataFitFunction)
  throws InterruptedException, InvocationTargetException {
    setFitFunction(dataFitFunction);
    SwingUtilities.invokeAndWait(new XYSeriesSetter(xySeriesCollection,
      createFitXYSeries(new XYFunctionSampler(dataFitFunction,
        measuredXYSeries.getX(0).doubleValue(), measuredXYSeries.getX(
          measuredXYSeries.getItemCount() - 1).doubleValue(), 400))));
  }

  protected synchronized XYFunction getFitFunction() {
    return fitFunction;
  }

  protected synchronized void setFitFunction(XYFunction function) {
    fitFunction = function;
  }

  protected void addDataPointListenerToScan() {
    scan.addDataPointListener(scanDataPointListener);
  }

  protected void removeDataPointListenerFromScan() {
    scan.removeDataPointListener(scanDataPointListener);
  }

  protected void receivedScanDataPoint(EpicsScanDataPointEvent event) {
    if (event.isReset()) {
      measuredXYSeries.clear();
      int index = xySeriesCollection.indexOf(FIT_SERIES_KEY);
      if (index != -1) xySeriesCollection.removeSeries(index);
      setMeasuredXYSeriesInterpolator(null);
      setFitFunction(null);
      setXPointOfInterest(null);
    } else if (event.isPoint()) {
      try {
        Pair<? extends Number, ? extends Number> point = getDataPoint(event);
        measuredXYSeries.add(point.getFirst(), point.getSecond());
      } catch (SeriesException e) {
        LOG.warn("Duplicate optimization data point X value");
      }
    }
  }

  /**
   * I get a data point from the specified event.
   * <p>
   * I use the position readback value at index 0 for the X value and the
   * detector value at index 0 for the Y value. Subclasses may override me to
   * change this behavior.
   *
   * @return a Pair instance with the X value as the first object and the Y
   *         value as the second object
   */
  protected Pair<? extends Number, ? extends Number> getDataPoint(
      EpicsScanDataPointEvent event) {
    return new Pair<Double,Float>(event.getPositionerReadbackValues().get(0),
      event.getDetectorValues().get(0));
  }

  /**
   * I get data from the specified scan. I expect the scan to have finished
   * successfully and to have data available via the scan object's
   * <code>getPositionerReadbackPositions</code> and
   * <code>getDetectorData</code> methods. I use the positioner readback
   * positions at index 0 for the X data and the detector data at index 0 for
   * the Y data. Subclasses may override me to change this behavior.
   */
  protected Pair<List<? extends Number>, List<? extends Number>> getData(
      EpicsScan scan) {
    return new Pair<List<? extends Number>,List<? extends Number>>(
        scan.getPositionerReadbackPositions(0), scan.getDetectorData(0));
  }

  protected void allowBeamPositionCorrection()
  throws EpicsCaException, TimeoutException, InterruptedException {
    enableAutobp();
    Thread.sleep(20 * 1000);
    disableAutobp();
  }

  protected XYSeries createFitXYSeries(XYFunctionSampler sampler) {
    XYSeries series = new XYSeries(FIT_SERIES_KEY, false, true);
    for (int i = 0; i < sampler.getNumberOfSamples(); i++)
      series.add(sampler.getX(i), sampler.getY(i), false);
    return series;
  }

  protected void setProgressAndStatusToComplete() {
    setProgress(100);
    setStatus("Completed successfully");
  }
}
