package imcacat.imcabc;

import imcacat.epicsca4j.ca.EpicsCaException;
import imcacat.epicsca4j.model.EpicsMotor;
import imcacat.epicsca4j.model.EpicsPv;
import imcacat.epicsca4j.model.EpicsScan;
import imcacat.epicsca4j.model.EpicsScanFactory;
import imcacat.epicsca4j.model.EpicsSlit;

import gov.aps.jca.TimeoutException;

/**
 * I am an <code>Optimization</code> for the 17-BM beamline.
 */
public abstract class Optimization17bm extends Optimization {
  protected static final double DEFAULT_MIDDLE_SLITS_WIDTH = 2000.0;
  protected static final double DEFAULT_MIDDLE_SLITS_HEIGHT = 2000.0;
  protected static final double DEFAULT_DOWNSTREAM_SLITS_WIDTH = 500.0;
  protected static final double DEFAULT_DOWNSTREAM_SLITS_HEIGHT = 200.0;
  protected static final double DEFAULT_SMALL_SLIT_WIDTH = 200.0;

  protected static final long MIDDLE_SLIT_MOVE_TIMEOUT = 20000;
  protected static final long DOWNSTREAM_SLIT_MOVE_TIMEOUT = 20000;
  protected static final long SMALL_SLIT_MOVE_TIMEOUT = 20000;

  protected static final short EXPOSURE_SHUTTER_OPEN = 1;
  protected static final short EXPOSURE_SHUTTER_CLOSE = 0;

  protected EpicsPv exposureShutterBoPv;
  protected boolean triedToOpenExposureShutter;

  protected EpicsSlit middleHorizontalSlit;
  protected EpicsSlit middleVerticalSlit;

  protected EpicsSlit downstreamHorizontalSlit;
  protected EpicsSlit downstreamVerticalSlit;

  protected EpicsMotor smallHorizontalSlitSize;

  public Optimization17bm() {
    super();

    triedToOpenExposureShutter = false;
  }

  @Override
  protected void initialize() throws InterruptedException, EpicsCaException {
    super.initialize();

    exposureShutterBoPv = new EpicsPv(
      Resources.getDefault().getExposureShutterBoPvName(), false, false);

    middleHorizontalSlit = new EpicsSlit(
      Resources.getDefault().getMiddleHorizontalSlitPvName(), "\u03BCm");
    middleVerticalSlit = new EpicsSlit(
      Resources.getDefault().getMiddleVerticalSlitPvName(), "\u03BCm");

    downstreamHorizontalSlit = new EpicsSlit(
      Resources.getDefault().getDownstreamHorizontalSlitPvName(),
      "\u03BCm");
    downstreamVerticalSlit = new EpicsSlit(
      Resources.getDefault().getDownstreamVerticalSlitPvName(),
      "\u03BCm");

    smallHorizontalSlitSize = new EpicsMotor(
      Resources.getDefault().getSmallHorizontalSlitSizeMotorPvName(),
      "\u03BCm");
  }

  protected EpicsScan getScanInstance()
  throws InterruptedException, EpicsCaException {
    return EpicsScanFactory.getScanInstance(
      Resources.getDefault().getScan1PvName());
  }

  protected EpicsScanResetter getScanResetterInstance() {
    return new EpicsScanResetter(
      Resources.getDefault().getScan1ResetPvName());
  }

  @Override
  public void release() {
    if (exposureShutterBoPv != null)
      exposureShutterBoPv.release();
    if (middleHorizontalSlit != null)
      middleHorizontalSlit.release();
    if (middleVerticalSlit != null)
      middleVerticalSlit.release();
    if (downstreamHorizontalSlit != null)
      downstreamHorizontalSlit.release();
    if (downstreamVerticalSlit != null)
      downstreamVerticalSlit.release();
    if (smallHorizontalSlitSize != null)
      smallHorizontalSlitSize.release();

    super.release();
  }

  @Override
  protected void cleanUp() {
    Exception caughtException = null;

    if (triedToOpenExposureShutter) {
      try {
        closeExposureShutter();
      } catch (Exception e) {
        if (caughtException == null) caughtException = e;
      }
    }

    try {
      super.cleanUp();
    } catch (Exception e) {
      if (caughtException == null) caughtException = e;
    }

    if (caughtException != null) throw new RuntimeException(caughtException);
  }

  protected double getMiddleHorizontalSlitCenter()
  throws InterruptedException, EpicsCaException {
    return middleHorizontalSlit.getCenter();
  }

  protected void moveMiddleHorizontalSlitCenter(double center)
  throws InterruptedException, TimeoutException, EpicsCaException {
    middleHorizontalSlit.setDesiredCenter(center);
    middleHorizontalSlit.moveCenterAndWait(MIDDLE_SLIT_MOVE_TIMEOUT);
  }

  protected double getMiddleVerticalSlitCenter()
  throws InterruptedException, EpicsCaException {
    return middleVerticalSlit.getCenter();
  }

  protected void moveMiddleVerticalSlitCenter(double center)
  throws InterruptedException, TimeoutException, EpicsCaException {
    middleVerticalSlit.setDesiredCenter(center);
    middleVerticalSlit.moveCenterAndWait(MIDDLE_SLIT_MOVE_TIMEOUT);
  }

  protected void moveMiddleSlits(double width, double height)
  throws InterruptedException, TimeoutException, EpicsCaException {
    middleHorizontalSlit.setDesiredSize(width);
    middleHorizontalSlit.moveSizeAndWait(MIDDLE_SLIT_MOVE_TIMEOUT);
    middleVerticalSlit.setDesiredSize(height);
    middleVerticalSlit.moveSizeAndWait(MIDDLE_SLIT_MOVE_TIMEOUT);
  }

  protected double getDownstreamHorizontalSlitCenter()
  throws InterruptedException, EpicsCaException {
    return downstreamHorizontalSlit.getCenter();
  }

  protected void moveDownstreamHorizontalSlitCenter(double center)
  throws InterruptedException, TimeoutException, EpicsCaException {
    downstreamHorizontalSlit.setDesiredCenter(center);
    downstreamHorizontalSlit.moveCenterAndWait(DOWNSTREAM_SLIT_MOVE_TIMEOUT);
  }

  protected double getDownstreamVerticalSlitCenter()
  throws InterruptedException, EpicsCaException {
    return downstreamVerticalSlit.getCenter();
  }

  protected void moveDownstreamVerticalSlitCenter(double center)
  throws InterruptedException, TimeoutException, EpicsCaException {
    downstreamVerticalSlit.setDesiredCenter(center);
    downstreamVerticalSlit.moveCenterAndWait(DOWNSTREAM_SLIT_MOVE_TIMEOUT);
  }

  protected void moveDownstreamSlits(double width, double height)
  throws InterruptedException, TimeoutException, EpicsCaException {
    downstreamHorizontalSlit.setDesiredSize(width);
    downstreamHorizontalSlit.moveSizeAndWait(DOWNSTREAM_SLIT_MOVE_TIMEOUT);
    downstreamVerticalSlit.setDesiredSize(height);
    downstreamVerticalSlit.moveSizeAndWait(DOWNSTREAM_SLIT_MOVE_TIMEOUT);
  }

  protected void moveSmallSlit(double width)
  throws InterruptedException, TimeoutException, EpicsCaException {
    smallHorizontalSlitSize.setDesiredPosition(width);
    smallHorizontalSlitSize.moveAndWait(SMALL_SLIT_MOVE_TIMEOUT);
  }

  protected void moveMiddleAndDownstreamAndSmallSlitsToDefaultSize()
  throws InterruptedException, TimeoutException, EpicsCaException {
    moveMiddleSlits(DEFAULT_MIDDLE_SLITS_WIDTH,
      DEFAULT_MIDDLE_SLITS_HEIGHT);
    moveDownstreamSlits(DEFAULT_DOWNSTREAM_SLITS_WIDTH,
      DEFAULT_DOWNSTREAM_SLITS_HEIGHT);
    moveSmallSlit(DEFAULT_SMALL_SLIT_WIDTH);
  }

  protected void openExposureShutter()
  throws InterruptedException, EpicsCaException {
    triedToOpenExposureShutter = true;
    exposureShutterBoPv.setValue(EXPOSURE_SHUTTER_OPEN);
  }

  protected void closeExposureShutter()
  throws InterruptedException, EpicsCaException {
    exposureShutterBoPv.setValue(EXPOSURE_SHUTTER_CLOSE);
  }
}
