package imcacat.imcabc;

import imcacat.epicsca4j.util.StringUtilities;
import imcacat.imcabc.util.Version;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InterruptedIOException;
import java.net.URL;
import java.util.Arrays;
import java.util.Collections;
import java.util.EnumMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.regex.Pattern;

import javax.swing.ImageIcon;

import org.apache.log4j.Logger;

/**
 * I provide resources (e.g. icons, PV names, properties) for this
 * application.
 */
public class Resources {
  private static final Logger LOG = Logger.getLogger(Resources.class);

  private static final String IMCABC_PROPERTIES_PATH = "/imcabc.properties";

  private static final Pattern NEWLINE_PAT = Pattern.compile("\\n");

  private static Resources DEFAULT = null;

  protected Properties properties;

  protected Map<MadElement,String> madElementToDisplayStringMap;
  protected Map<MadEdge,String> madEdgeToDisplayStringMap;
  protected Map<MadScanType,String> madScanTypeToDisplayStringMap;
  protected Map<Operation,String> operationToDisplayStringMap;

  public Resources() throws IOException {
      this(null);
  }

  public Resources(File conf) throws IOException {
    super();
    initialize();
    load(conf);
  }

  public Resources(InputStream conf, String confPath) throws IOException {
    super();
    initialize();
    load(conf, confPath);
  }

  private void initialize() {
    properties = new Properties();

    madElementToDisplayStringMap = new EnumMap<MadElement,String>(
        MadElement.class);
    for (MadElement each : MadElement.values())
      madElementToDisplayStringMap.put(each,
        StringUtilities.toTitleCase(each.toString()));

    madEdgeToDisplayStringMap = new EnumMap<MadEdge,String>(MadEdge.class);
    Pattern elementEdgeNameSeparatorPattern = Pattern.compile("_");
    for (MadEdge each : MadEdge.values()) {
      madEdgeToDisplayStringMap.put(each, String.format("%s (%.1f eV)",
          elementEdgeNameSeparatorPattern.split(each.toString())[1],
          each.getEnergy()));
    }

    madScanTypeToDisplayStringMap = new EnumMap<MadScanType,String>(
        MadScanType.class);
      for (MadScanType each : MadScanType.values())
        madScanTypeToDisplayStringMap.put(each,
          StringUtilities.toTitleCase(each.toString()));

    operationToDisplayStringMap = new EnumMap<Operation,String>(
        Operation.class);
    operationToDisplayStringMap.put(Operation.IDLE, "Idle");
    operationToDisplayStringMap.put(Operation.FLUORESCENCE_SCAN,
      "Fluorescence Scan");
    operationToDisplayStringMap.put(Operation.OPTIMIZATION, "Optimization");
    for (Operation each : Operation.values())
      if (!operationToDisplayStringMap.containsKey(each))
        throw new ProgrammerError("missing display string mapping for " +
          each + " Operation");
  }

  private void load(File conf) throws IOException {
    FileInputStream stream = null;
    try {
      String confPath = null;
      if (conf != null) {
        stream = new FileInputStream(conf);
        confPath = conf.getAbsolutePath();
      }
      load(stream, confPath);
      if (stream != null) {
        stream.close();
        stream = null;
      }
    } finally {
      closeSilently(stream);
    }
  }

  private void load(InputStream conf, String confPath) throws IOException {
    properties.clear();

    if (conf != null) {
      try {
        properties.load(conf);
      } catch (IOException e) {
        LOG.fatal("Failed to load alternate configuration" +
          ((confPath == null) ? "" : " file \"" + confPath + "\""), e);
        throw new RuntimeException(
          "Failed to load alternate configuration" +
            ((confPath == null) ? "" : " file \"" + confPath + "\""), e);
      }
    } else {
      InputStream stream = null;
      try {
        stream = Resources.class.getResourceAsStream(IMCABC_PROPERTIES_PATH);
        properties.load(stream);
        stream.close();
        stream = null;
      } catch (IOException e) {
        LOG.fatal("Failed to load default properties resource \"" +
            IMCABC_PROPERTIES_PATH + "\"", e);
        throw new RuntimeException("Failed to load default properties " +
          "resource \"" + IMCABC_PROPERTIES_PATH + "\"", e);
      } finally {
        closeSilently(stream);
      }

      File hostConfFile = new File(
        properties.getProperty("hostConfFilePath"));
      if (hostConfFile.exists()) {
        InputStream hostConfStream = null;
        try {
          hostConfStream = new FileInputStream(hostConfFile);
          properties.load(hostConfStream);
          hostConfStream.close();
          hostConfStream = null;
        } catch (Exception e) {
          LOG.fatal("Failed to load host configuration file \"" +
            hostConfFile.getAbsolutePath() + "\"", e);
          throw new RuntimeException(
            "Failed to load host configuration file \"" +
            hostConfFile.getAbsolutePath() + "\"", e);
        } finally {
          closeSilently(hostConfStream);
        }
      }

      File userConfFile = new File(getUserHomeDirectory(),
        properties.getProperty("userConfFilePath"));
      if (userConfFile.exists()) {
        InputStream userConfStream = null;
        try {
          userConfStream = new FileInputStream(userConfFile);
          properties.load(userConfStream);
          userConfStream.close();
          userConfStream = null;
        } catch (Exception e) {
          LOG.fatal("Failed to load user configuration file \"" +
              userConfFile.getAbsolutePath() + "\"", e);
          throw new RuntimeException(
            "Failed to load user configuration file \"" +
            userConfFile.getAbsolutePath() + "\"", e);
        } finally {
          closeSilently(userConfStream);
        }
      }
    }
  }

  public Beamline getBeamline() {
    return Beamline.getInstanceWithId(properties.getProperty("beamline"));
  }

  public String getRingCurrentPvName() {
    return properties.getProperty("ringCurrentPvName");
  }

  public String getBShutterBiPvName() {
    return properties.getProperty("bShutterBiPvName");
  }

  public String getBShutterBoPvName() {
    return properties.getProperty("bShutterBoPvName");
  }

  public String getExposureShutterBoPvName() {
    return properties.getProperty("exposureShutterBoPvName");
  }

  public String getFilterRackBasePvName() {
    return properties.getProperty("filterRackBasePvName");
  }

  public String getBpmXPositionPvName() {
    return properties.getProperty("bpmXPositionPvName");
  }

  public String getBpmYPositionPvName() {
    return properties.getProperty("bpmYPositionPvName");
  }

  public List<String> getDisplayPostScriptFileCommand() {
    String key = null;
    if (isPlatformMacOsX())
      key = "displayPostScriptFileCommandOnMacOsX";
    else if (isPlatformWindows())
      key = "displayPostScriptFileCommandOnWindows";
    else
      key = "displayPostScriptFileCommandOnOther";
    return Collections.unmodifiableList(
        Arrays.asList(NEWLINE_PAT.split(properties.getProperty(key))));
  }

  public String getChoochCommand() {
    return properties.getProperty("chooch");
  }

  public double getChoochMonochromatorEnergyResolution() {
    return Double.parseDouble(
      properties.getProperty("choochMonochromatorEnergyResolution"));
  }

  public String getDisplayString(MadElement element) {
    return madElementToDisplayStringMap.get(element);
  }

  public Map<MadElement,String> getMadElementToDisplayStringMap() {
    return Collections.unmodifiableMap(madElementToDisplayStringMap);
  }

  public String getDisplayString(MadEdge edge) {
    return madEdgeToDisplayStringMap.get(edge);
  }

  public Map<MadEdge,String> getMadEdgeToDisplayStringMap() {
    return Collections.unmodifiableMap(madEdgeToDisplayStringMap);
  }

  public String getDisplayString(MadScanType edge) {
    return madScanTypeToDisplayStringMap.get(edge);
  }

  public Map<MadScanType,String> getMadScanTypeToDisplayStringMap() {
    return Collections.unmodifiableMap(madScanTypeToDisplayStringMap);
  }

  public String getDisplayString(Operation operation) {
    return operationToDisplayStringMap.get(operation);
  }

  public String getMadScanPvName() {
    return properties.getProperty("madScanPvName");
  }

  public String getMadScanResetPvName() {
    return properties.getProperty("madScanResetPvName");
  }

  public String getFluorescenceMcaRoiLowPvName() {
    return properties.getProperty("fluorescenceMcaRoiLowPvName");
  }

  public String getFluorescenceMcaRoiHighPvName() {
    return properties.getProperty("fluorescenceMcaRoiHighPvName");
  }

  public String getFluorescenceMcaAcquisitionTimePvName() {
    return properties.getProperty("fluorescenceMcaAcquisitionTimePvName");
  }

  public String getAutobpRequestExternalPausePvName() {
    return properties.getProperty("autobpRequestExternalPausePvName");
  }

  public short getAutobpRequestExternalPauseValue() {
    return Short.parseShort(properties.getProperty(
      "autobpRequestExternalPauseValue"));
  }

  public String getAutobpReleaseExternalPausePvName() {
    return properties.getProperty("autobpReleaseExternalPausePvName");
  }

  public short getAutobpReleaseExternalPauseValue() {
    return Short.parseShort(properties.getProperty(
      "autobpReleaseExternalPauseValue"));
  }

  public String getFluorescenceAcquirePvName() {
    return properties.getProperty("fluorescenceAcquirePvName");
  }

  public String getFluorescenceIntensityPvName() {
    return properties.getProperty("fluorescenceIntensityPvName");
  }

  public String getFluorescenceMcaRoiSumPvName() {
    return properties.getProperty("fluorescenceMcaRoiSumPvName");
  }

  public String getFluorescenceScanBeforeScanPvName() {
    return properties.getProperty("fluorescenceScanBeforeScanPvName");
  }

  public float getFluorescenceScanBeforeScanValue() {
    return Float.parseFloat(
      properties.getProperty("fluorescenceScanBeforeScanValue"));
  }

  public String getFluorescenceScanAfterScanPvName() {
    return properties.getProperty("fluorescenceScanAfterScanPvName");
  }

  public float getFluorescenceScanAfterScanValue() {
    return Float.parseFloat(
      properties.getProperty("fluorescenceScanAfterScanValue"));
  }

  public double getEnergyAxisToEvScaleFactor() {
    return Double.parseDouble(
      properties.getProperty("energyAxisToEvScaleFactor"));
  }

  public boolean isEnergyAxisAutoStartSupported() {
    return getEnergyAxisAutoStartPvName() != null;
  }

  public String getEnergyAxisAutoStartPvName() {
    return properties.getProperty("energyAxisAutoStartPvName");
  }

  /**
   * Returns energy low limit in eV.
   */
  public double getEnergyLowLimit() {
    return Double.parseDouble(properties.getProperty("energyLowLimit"));
  }

  /**
   * Returns energy high limit in eV.
   */
  public double getEnergyHighLimit() {
    return Double.parseDouble(properties.getProperty("energyHighLimit"));
  }

  /**
   * Returns energy deadband in eV.
   */
  public double getEnergyDeadband() {
    return Double.parseDouble(properties.getProperty("energyDeadband"));
  }

  public boolean isXFlashConfigSupported() {
    return getXFlashConfigCoolingModePvName() != null &&
      getXFlashConfigThermStatePvName() != null &&
      getXFlashConfigPollRatePvName() != null;
  }

  public String getXFlashConfigCoolingModePvName() {
    return properties.getProperty("xFlashConfigCoolingModePvName");
  }

  public String getXFlashConfigThermStatePvName() {
    return properties.getProperty("xFlashConfigThermStatePvName");
  }

  public String getXFlashConfigPollRatePvName() {
    return properties.getProperty("xFlashConfigPollRatePvName");
  }

  public boolean isExperimentStateSupported() {
    return getExperimentStatePvName() != null &&
      getExperimentStateRbvPvName() != null &&
      getExperimentStateSetPvName() != null &&
      getExperimentStateStatusPvName() != null &&
      getExperimentStateAbortPvName() != null;
  }

  public String getExperimentStatePvName() {
    return properties.getProperty("experimentStatePvName");
  }

  public String getExperimentStateRbvPvName() {
    return properties.getProperty("experimentStateRbvPvName");
  }

  public String getExperimentStateSetPvName() {
    return properties.getProperty("experimentStateSetPvName");
  }

  public String getExperimentStateStatusPvName() {
    return properties.getProperty("experimentStateStatusPvName");
  }

  public String getExperimentStateAbortPvName() {
    return properties.getProperty("experimentStateAbortPvName");
  }

  protected String getDefaultPropertiesResourceName() {
    return "/imcabc.properties";
  }

  public String getEnergyMotorPvName() {
    return properties.getProperty("energyMotorPvName");
  }

  public String getEnergyMotorMoveModeBoPvName() {
    return properties.getProperty("energyMotorMoveModeBoPvName");
  }

  public String getEnergyMotorMoveModeMbbiPvName() {
    return properties.getProperty("energyMotorMoveModeMbbiPvName");
  }

  public String getDesiredEnergyPositionPvName() {
    switch (getBeamline()) {
      case ID17: return getEnergyAxisPvName() + "RqsPos";
      case BM17: return getEnergyMotorPvName() + ".VAL";
      default: throw new ProgrammerError("Unexpected beamline");
    }
  }

  public String getEnergyPositionPvName() {
    switch (getBeamline()) {
      case ID17: return getEnergyAxisPvName() + "ActPos";
      case BM17: return getEnergyMotorPvName() + ".RBV";
      default: throw new ProgrammerError("Unexpected beamline");
    }
  }

  public String getScansBasePvName() {
    return properties.getProperty("scansBasePvName");
  }

  public String getScan1PvName() {
    return getScansBasePvName() + "1";
  }

  public String getScansResetBasePvName() {
    return properties.getProperty("scansResetBasePvName");
  }

  public String getScan1ResetPvName() {
    return getScansResetBasePvName() + "1";
  }

  public String getFocusMotorPvName() {
    return properties.getProperty("focusMotorPvName");
  }

  public String getDesiredFocusPositionPvName() {
    return getFocusMotorPvName() + ".VAL";
  }

  public String getFocusPositionPvName() {
    return getFocusMotorPvName() + ".RBV";
  }

  public String getMirrorVerticalJackMotorPvName() {
    return properties.getProperty("mirrorVerticalJackMotorPvName");
  }

  public String getDesiredMirrorVerticalJackPositionPvName() {
    return getMirrorVerticalJackMotorPvName() + ".VAL";
  }

  public String getMirrorVerticalJackPositionPvName() {
    return getMirrorVerticalJackMotorPvName() + ".RBV";
  }

  public String getMirrorBendMotorPvName() {
    return properties.getProperty("mirrorBendMotorPvName");
  }

  public String getDesiredMirrorBendPositionPvName() {
    return getMirrorBendMotorPvName() + ".VAL";
  }

  public String getMirrorBendPositionPvName() {
    return getMirrorBendMotorPvName() + ".RBV";
  }

  public String getOmegaMotorPvName() {
    return properties.getProperty("omegaMotorPvName");
  }

  public String getDesiredOmegaPositionPvName() {
    return getOmegaMotorPvName() + ".VAL";
  }

  public String getOmegaPositionPvName() {
    return getOmegaMotorPvName() + ".RBV";
  }

  public String getChiMotorPvName() {
    return properties.getProperty("chiMotorPvName");
  }

  public String getDesiredChiPositionPvName() {
    return getChiMotorPvName() + ".VAL";
  }

  public String getChiPositionPvName() {
    return getChiMotorPvName() + ".RBV";
  }

  public String getBpmHorizontalMotorPvName() {
    return properties.getProperty("bpmHorizontalMotorPvName");
  }

  public String getDesiredBpmHorizontalPositionPvName() {
    return getBpmHorizontalMotorPvName() + ".VAL";
  }

  public String getBpmHorizontalPositionPvName() {
    return getBpmHorizontalMotorPvName() + ".RBV";
  }

  public String getBpmVerticalMotorPvName() {
    return properties.getProperty("bpmVerticalMotorPvName");
  }

  public String getDesiredBpmVerticalPositionPvName() {
    return getBpmVerticalMotorPvName() + ".VAL";
  }

  public String getBpmVerticalPositionPvName() {
    return getBpmVerticalMotorPvName() + ".RBV";
  }

  public String getSmallHorizontalSlitSizeMotorPvName() {
    return properties.getProperty("smallHorizontalSlitSizeMotorPvName");
  }

  public String getMiddleVerticalSlitPvName() {
    return properties.getProperty("middleVerticalSlitPvName");
  }

  public String getDesiredMiddleVerticalSlitSizePvName() {
    return getMiddleVerticalSlitPvName() + "size.VAL";
  }

  public String getMiddleVerticalSlitSizePvName() {
    return getMiddleVerticalSlitPvName() + "t2.C";
  }

  public String getDesiredMiddleVerticalSlitCenterPvName() {
    return getMiddleVerticalSlitPvName() + "center.VAL";
  }

  public String getMiddleVerticalSlitCenterPvName() {
    return getMiddleVerticalSlitPvName() + "t2.D";
  }

  public String getMiddleHorizontalSlitPvName() {
    return properties.getProperty("middleHorizontalSlitPvName");
  }

  public String getDesiredMiddleHorizontalSlitSizePvName() {
    return getMiddleHorizontalSlitPvName() + "size.VAL";
  }

  public String getMiddleHorizontalSlitSizePvName() {
    return getMiddleHorizontalSlitPvName() + "t2.C";
  }

  public String getDesiredMiddleHorizontalSlitCenterPvName() {
    return getMiddleHorizontalSlitPvName() + "center.VAL";
  }

  public String getMiddleHorizontalSlitCenterPvName() {
    return getMiddleHorizontalSlitPvName() + "t2.D";
  }

  public String getDownstreamVerticalSlitPvName() {
    return properties.getProperty("downstreamVerticalSlitPvName");
  }

  public String getDesiredDownstreamVerticalSlitSizePvName() {
    return getDownstreamVerticalSlitPvName() + "size.VAL";
  }

  public String getDownstreamVerticalSlitSizePvName() {
    return getDownstreamVerticalSlitPvName() + "t2.C";
  }

  public String getDesiredDownstreamVerticalSlitCenterPvName() {
    return getDownstreamVerticalSlitPvName() + "center.VAL";
  }

  public String getDownstreamVerticalSlitCenterPvName() {
    return getDownstreamVerticalSlitPvName() + "t2.D";
  }

  public String getDownstreamHorizontalSlitPvName() {
    return properties.getProperty("downstreamHorizontalSlitPvName");
  }

  public String getDesiredDownstreamHorizontalSlitSizePvName() {
    return getDownstreamHorizontalSlitPvName() + "size.VAL";
  }

  public String getDownstreamHorizontalSlitSizePvName() {
    return getDownstreamHorizontalSlitPvName() + "t2.C";
  }

  public String getDesiredDownstreamHorizontalSlitCenterPvName() {
    return getDownstreamHorizontalSlitPvName() + "center.VAL";
  }

  public String getDownstreamHorizontalSlitCenterPvName() {
    return getDownstreamHorizontalSlitPvName() + "t2.D";
  }

  public String getBpmIntensitySumPvName() {
    return properties.getProperty("bpmIntensitySumPvName");
  }

  public String getBpmIntensityAcquisitionTimePvName() {
    return properties.getProperty("bpmIntensityAcquisitionTimePvName");
  }

  public String getBpmIntensityAcquisitionStartPvName() {
    return properties.getProperty("bpmIntensityAcquisitionStartPvName");
  }

  public String getBpmIntensityAcquisitionSumPvName() {
    return properties.getProperty("bpmIntensityAcquisitionSumPvName");
  }

  public String getSlitAlignmentIntensitySumPvName() {
    return properties.getProperty("slitAlignmentIntensitySumPvName");
  }

  public String getSlitAlignmentIntensityAcquisitionTimePvName() {
    return properties.getProperty(
      "slitAlignmentIntensityAcquisitionTimePvName");
  }

  public String getSlitAlignmentIntensityAcquisitionStartPvName() {
    return properties.getProperty(
      "slitAlignmentIntensityAcquisitionStartPvName");
  }

  public String getSlitAlignmentIntensityAcquisitionSumPvName() {
    return properties.getProperty(
      "slitAlignmentIntensityAcquisitionSumPvName");
  }

  public String getSmallIonChamberIntensitySumPvName() {
    return properties.getProperty("smallIonChamberIntensitySumPvName");
  }

  public String getSmallIonChamberIntensityAcquisitionTimePvName() {
    return properties.getProperty(
      "smallIonChamberIntensityAcquisitionTimePvName");
  }

  public String getSmallIonChamberIntensityAcquisitionStartPvName() {
    return properties.getProperty(
      "smallIonChamberIntensityAcquisitionStartPvName");
  }

  public String getSmallIonChamberIntensityAcquisitionSumPvName() {
    return properties.getProperty(
      "smallIonChamberIntensityAcquisitionSumPvName");
  }

  public String getHarmonicPvName() {
    return properties.getProperty("harmonicPvName");
  }

  public String getGapPvName() {
    return properties.getProperty("gapPvName");
  }

  public String getMinibeamBasePvName() {
    return properties.getProperty("minibeamBasePvName");
  }

  public String getEnergyAssemblyPvName() {
    return properties.getProperty("energyAssemblyPvName");
  }

  public String getEnergyAxisPvName() {
    return properties.getProperty("energyAxisPvName");
  }

  public String getMinibeamAperturePvName() {
    return getMinibeamBasePvName() + "Aperture.VAL";
  }

  public String getMinibeamApertureRbvPvName() {
    return getMinibeamBasePvName() + "ApertureRbv.VAL";
  }

  public String getMinibeamApertureMovePvName() {
    return getMinibeamBasePvName() + "ApertureMove.PROC";
  }

  public String getMinibeamApertureAbortPvName() {
    return getMinibeamBasePvName() + "ApertureAbort.PROC";
  }

  public String getMinibeamApertureSizePvName(int index) {
    return getMinibeamBasePvName() + "Ap" + (index + 1) + "Size.VAL";
  }

  public String getBpm2IntensityPvName() {
    return properties.getProperty("bpm2IntensityPvName");
  }

  public String getBpm2IntensityAcquisitionTimePvName() {
    return properties.getProperty("bpm2IntensityAcquisitionTimePvName");
  }

  public String getBpm2IntensityAcquisitionControlPvName() {
    return properties.getProperty("bpm2IntensityAcquisitionControlPvName");
  }

  public String getBpm2IntensityAcquisitionSumPvName() {
    return properties.getProperty("bpm2IntensityAcquisitionSumPvName");
  }

  public String getBpm3IntensityPvName() {
    return properties.getProperty("bpm3IntensityPvName");
  }

  public String getBpm3IntensityAcquisitionTimePvName() {
    return properties.getProperty("bpm3IntensityAcquisitionTimePvName");
  }

  public String getBpm3IntensityAcquisitionControlPvName() {
    return properties.getProperty("bpm3IntensityAcquisitionControlPvName");
  }

  public String getBpm3IntensityAcquisitionSumPvName() {
    return properties.getProperty("bpm3IntensityAcquisitionSumPvName");
  }

  public String getBpm3AutoRangePvName() {
    return properties.getProperty("bpm3AutoRangePvName");
  }

  public String getBpm3AutoRangeRqsPvName() {
    return properties.getProperty("bpm3AutoRangeRqsPvName");
  }

  public String getBpm3RangePvName() {
    return properties.getProperty("bpm3RangePvName");
  }

  public String getBpm3RangeRqsPvName() {
    return properties.getProperty("bpm3RangeRqsPvName");
  }

  public String getBpm3StageHorizMotorPvName() {
    return properties.getProperty("bpm3StageHorizMotorPvName");
  }

  public String getBpm3StageVertMotorPvName() {
    return properties.getProperty("bpm3StageVertMotorPvName");
  }

  public double getSecondCrystalRollOptimizationBpm3StageHorizOffset() {
    return Double.parseDouble(
      properties.getProperty(
        "secondCrystalRollOptimizationBpm3StageHorizOffset"));
  }

  public double getSecondCrystalRollOptimizationBpm3StageVertOffset() {
    return Double.parseDouble(
      properties.getProperty(
        "secondCrystalRollOptimizationBpm3StageVertOffset"));
  }

  public String getMirrorBpmIntensityPvName() {
    return properties.getProperty("mirrorBpmIntensityPvName");
  }

  public String getSampleBpmIntensityPvName() {
    return properties.getProperty("sampleBpmIntensityPvName");
  }

  public String getIonChamberIntensityPvName() {
    return properties.getProperty("ionChamberIntensityPvName");
  }

  public String getBeamStopIntensityPvName() {
    return properties.getProperty("beamStopIntensityPvName");
  }

  public String getAScansBasePvName() {
    return properties.getProperty("aScansBasePvName");
  }

  public String getAScan1PvName() {
    return getAScansBasePvName() + "1";
  }

  public String getAScansResetBasePvName() {
    return properties.getProperty("aScansResetBasePvName");
  }

  public String getAScan1ResetPvName() {
    return getAScansResetBasePvName() + "1";
  }

  public String getBScansBasePvName() {
    return properties.getProperty("bScansBasePvName");
  }

  public String getBScan1PvName() {
    return getBScansBasePvName() + "1";
  }

  public String getBScansResetBasePvName() {
    return properties.getProperty("bScansResetBasePvName");
  }

  public String getBScan1ResetPvName() {
    return getBScansResetBasePvName() + "1";
  }

  public String getSecondCrystalTuningAssemblyPvName() {
    return properties.getProperty("secondCrystalTuningAssemblyPvName");
  }

  public String getSecondCrystalPitchDrivePvName() {
    return properties.getProperty("secondCrystalPitchDrivePvName");
  }

  public String getSecondCrystalPitchPositionPvName() {
    return getSecondCrystalPitchDrivePvName() + "ActPos";
  }

  public String getDesiredSecondCrystalPitchPositionPvName() {
    return getSecondCrystalPitchDrivePvName() + "RqsPos";
  }

  public String getSecondCrystalFinePitchPositionPvName() {
    return properties.getProperty("secondCrystalFinePitchPiezoActPosPvName");
  }

  public String getDesiredSecondCrystalFinePitchPositionPvName() {
    return properties.getProperty("secondCrystalFinePitchPiezoRqsPosPvName");
  }

  public String getSecondCrystalRollDrivePvName() {
    return properties.getProperty("secondCrystalRollDrivePvName");
  }

  public String getSecondCrystalRollPositionPvName() {
    return getSecondCrystalRollDrivePvName() + "ActPos";
  }

  public String getDesiredSecondCrystalRollPositionPvName() {
    return getSecondCrystalRollDrivePvName() + "RqsPos";
  }

  public double getSecondCrystalPitchOptimizationFinePitchPosition() {
    return Double.parseDouble(properties.getProperty(
      "secondCrystalPitchOptimizationFinePitchPosition"));
  }

  public String getMonoAutobpAverageIntensityResetPvName() {
    return properties.getProperty("monoAutobpAverageIntensityResetPvName");
  }

  public String getAttenuationPvName() {
    return properties.getProperty("attenuationPvName");
  }

  public String getAttenuationRqsPvName() {
    return properties.getProperty("attenuationRqsPvName");
  }

  public static ImageIcon getAppIcon() {
    String resourcePath = "resources/app-icon-16.png";
    URL url = Resources.class.getResource("/" + resourcePath);
    return url == null ? new ImageIcon() : new ImageIcon(url);
  }

  public static ImageIcon getLargeAppIcon() {
    String resourcePath = "resources/app-icon-48.png";
    URL url = Resources.class.getResource("/" + resourcePath);
    return (url == null) ? new ImageIcon() : new ImageIcon(url);
  }

  public static String getAppName() {
    return "Imcabc";
  }

  public static String getAppCliName() {
    return "imcabc";
  }

  public static String getAppVersion() {
    return Version.ID;
  }

  public static boolean isPlatformMacOsX() {
    return System.getProperty("os.name").toLowerCase().startsWith("mac os x");
  }

  public static boolean isPlatformWindows() {
    return System.getProperty("os.name").toLowerCase().startsWith("windows");
  }

  public static File getUserHomeDirectory() {
    return new File(System.getProperty("user.home"));
  }

  public static String getLineSeparator() {
    return System.getProperty("line.separator");
  }

  public static synchronized Resources getDefault() {
    if (DEFAULT == null) {
      try {
        DEFAULT = new Resources();
      } catch (IOException e) {
        throw new RuntimeException("Failed to create default instance", e);
      }
    }
    return DEFAULT;
  }

  public static synchronized void setDefault(Resources newDefault) {
    DEFAULT = newDefault;
  }

  private static void closeSilently(InputStream stream) {
    if (stream == null) return;
    try {
      stream.close();
    } catch (InterruptedIOException e) {
      Thread.currentThread().interrupt();
    } catch (Exception e) {
      /* IGNORE */
    }
  }
}
