package imcacat.imcabc;

import imcacat.epicsca4j.ca.EpicsCaException;
import imcacat.epicsca4j.model.EpicsPv;
import imcacat.epicsca4j.model.TpmacAssemblyAxDrMo;
import imcacat.epicsca4j.util.StringUtilities;
import imcacat.epicsca4j.util.UiUtilities;

import java.lang.reflect.InvocationTargetException;
import java.text.DecimalFormat;

import javax.swing.JOptionPane;

import gov.aps.jca.TimeoutException;

/**
 * I manually optimize the DCM's Second Crystal Pitch.
 */
public class SecondCrystalPitchManualOptimization17id
extends DcmOptimization17id {
  protected TpmacAssemblyAxDrMo secondCrystalPitchDrive;
  protected EpicsPv secondCrystalFinePitchDesiredPositionPv;
  protected EpicsPv secondCrystalFinePitchPositionPv;

  protected double originalSecondCrystalPitchDrivePosition;
  protected double originalSecondCrystalFinePitchPosition;

  @Override
  protected void initialize()
  throws InterruptedException, EpicsCaException {
    super.initialize();
    secondCrystalPitchDrive = new TpmacAssemblyAxDrMo(
      Resources.getDefault().getSecondCrystalTuningAssemblyPvName(),
      Resources.getDefault().getSecondCrystalPitchDrivePvName(), 1.0,
      "mrad");
    originalSecondCrystalPitchDrivePosition =
      secondCrystalPitchDrive.getPosition();

    secondCrystalFinePitchDesiredPositionPv = new EpicsPv(
      Resources.getDefault().
        getDesiredSecondCrystalFinePitchPositionPvName(),
      false, false);
    secondCrystalFinePitchPositionPv = new EpicsPv(
      Resources.getDefault().getSecondCrystalFinePitchPositionPvName(),
      false, false);
    originalSecondCrystalFinePitchPosition =
      (Double)secondCrystalFinePitchPositionPv.getValue();
  }

  @Override
  public void release() {
    if (secondCrystalPitchDrive != null) secondCrystalPitchDrive.release();
    if (secondCrystalFinePitchDesiredPositionPv != null)
      secondCrystalFinePitchDesiredPositionPv.release();
    if (secondCrystalFinePitchPositionPv != null)
      secondCrystalFinePitchPositionPv.release();
    super.release();
  }

  @Override
  protected void go()
  throws InterruptedException, TimeoutException, EpicsCaException {
    setStatus("Configuring beamline for optimization");
    disableAutobp();
    secondCrystalFinePitchDesiredPositionPv.setValueAndWait(
      Resources.getDefault().
        getSecondCrystalPitchOptimizationFinePitchPosition());
    setProgress(10);

    setStatus("Scanning Second Crystal Pitch");
    scanSecondCrystalPitch();
    setProgress(60);

    setStatus("Fitting scan data to Gaussian function");
    GaussianFunctionCurveFitter fitter = null;
    try {
      fitter = new GaussianFunctionCurveFitter(measuredXYSeries.toArray());
    } catch (Exception e) {}
    if (fitter != null) {
      try {
        setFitDataSeriesInEventDispatchThreadAndWait(fitter);
      } catch (InvocationTargetException e) {
        throw new RuntimeException("Failed to add fit data to " +
          "XYSeriesCollection", e);
      }
    }
    setProgress(70);

    double measuredMax = getXWithMaximumY(measuredXYSeries);
    double fitMax = 0.0;

    boolean shouldMoveToMeasuredMax = false;
    boolean shouldMoveToFitMax = false;

    DecimalFormat positionFormat = new DecimalFormat("0.000");
    if (fitter == null) {
      setStatus("Waiting for measured choice");
      int choice = JOptionPane.showOptionDialog(UiUtilities.getRootFrame(),
        StringUtilities.wordWrap("The Second Crystal Pitch scan has " +
          "finished. However, the measured data could not be fit to a " +
          "Gaussian function. Would you like to move Second Crystal Pitch " +
          "to the measured maximum at " +
          positionFormat.format(measuredMax) + " " +
          getSecondCrystalPitchUnitOfMeasure() +
          "?",
          78, "\n"),
        "Move to measured maximum?", JOptionPane.YES_NO_OPTION,
        JOptionPane.QUESTION_MESSAGE, null, new String[] {"Yes", "Cancel"},
        "Yes");
      if (choice == JOptionPane.YES_OPTION) shouldMoveToMeasuredMax = true;
    } else {
      fitMax = fitter.getC();
      setStatus("Waiting for measured or fit choice");
      int choice = JOptionPane.showOptionDialog(UiUtilities.getRootFrame(),
        StringUtilities.wordWrap("The Second Crystal Pitch scan has " +
          "finished. The measured maximum is at " +
          positionFormat.format(measuredMax) + " " +
          getSecondCrystalPitchUnitOfMeasure() +
          ", and the fit maximum is at " +
          positionFormat.format(fitMax) + " " +
          getSecondCrystalPitchUnitOfMeasure() +
          ". To which maximum would you like to move Second Crystal Pitch?",
          78, "\n"),
        "Move to measured or fit?", JOptionPane.YES_NO_CANCEL_OPTION,
        JOptionPane.QUESTION_MESSAGE, null, new String[] {"Measured", "Fit",
        "Cancel"}, "Fit");
      if (choice == JOptionPane.YES_OPTION) shouldMoveToMeasuredMax = true;
      else if (choice == JOptionPane.NO_OPTION) shouldMoveToFitMax = true;
    }
    setProgress(75);
    if (shouldMoveToMeasuredMax) {
      setStatus("Moving to measured maximum");
      moveSecondCrystalPitch(measuredMax);
    } else if (shouldMoveToFitMax) {
      setStatus("Moving to fit maximum");
      moveSecondCrystalPitch(fitMax);
    } else {
      setStatus("Canceling");
      moveSecondCrystalPitch(originalSecondCrystalPitchDrivePosition);
      secondCrystalFinePitchDesiredPositionPv.setValueAndWait(
        originalSecondCrystalFinePitchPosition);
      resetMonoAutobpAverageIntensity();
      enableAutobp();
      setStatus("Canceled");
      return;
    }
    setProgress(90);

    setStatus("Configuring beamline for use");
    resetMonoAutobpAverageIntensity();
    enableAutobp();

    setProgressAndStatusToComplete();
  }

  @Override
  public void cleanUp() {
    Exception caughtException = null;

    try {
      scan.stop();
    } catch (Exception e) {
      if (caughtException == null) caughtException = e;
    }
    try {
      secondCrystalPitchDrive.stopAndWait();
    } catch (Exception e) {
      if (caughtException == null) caughtException = e;
    }
    try {
      secondCrystalPitchDrive.setDesiredPosition(
        originalSecondCrystalPitchDrivePosition);
      secondCrystalPitchDrive.moveAndWait();
    } catch (Exception e) {
      if (caughtException == null) caughtException = e;
    }
    try {
      secondCrystalFinePitchDesiredPositionPv.setValueAndWait(
        originalSecondCrystalFinePitchPosition);
    } catch (Exception e) {
      if (caughtException == null) caughtException = e;
    }
    try {
      resetMonoAutobpAverageIntensity();
    } catch (Exception e) {
      if (caughtException == null) caughtException = e;
    }
    try {
      enableAutobp();
    } catch (Exception e) {
      if (caughtException == null) caughtException = e;
    }

    try {
      super.cleanUp();
    } catch (Exception e) {
      if (caughtException == null) caughtException = e;
    }

    if (caughtException != null) throw new RuntimeException(caughtException);
  }

  protected void scanSecondCrystalPitch()
  throws EpicsCaException, InterruptedException {
    performSecondCrystalPitchScan();
    if (!scan.completedSuccessfully())
      throw new RuntimeException(
        "Second Crystal Pitch scan did not complete successfully");
    try {
      setMeasuredDataSeriesInEventDispatchThreadAndWait(getData(scan));
    } catch (InvocationTargetException e) {
      throw new RuntimeException("Failed to update measured data series " +
        "with complete data from scan", e);
    }
  }

  protected void performSecondCrystalPitchScan() throws EpicsCaException,
  InterruptedException {
    setXYDatasetTitle("Second Crystal Pitch optimization scan");
    setXDataLabel("Second Crystal Pitch (" +
      getSecondCrystalPitchUnitOfMeasure() + ")");
    setYDataLabel("Mirror BPM intensity (counts)");
    secondCrystalPitchDrive.setAutoStartEnabled(true);
    scanResetter.reset();
    scan.setBeforeScanPvLink(
      Resources.getDefault().getBpm2IntensityAcquisitionTimePvName());
    scan.setBeforeScanCommandData(1.0f);
    scan.setPositionerPvLink(0,
      Resources.getDefault().
        getDesiredSecondCrystalPitchPositionPvName());
    scan.setPositionerReadbackPvLink(0,
      Resources.getDefault().getSecondCrystalPitchPositionPvName());
    scan.setPositionerReadbackDelta(0, 0.004);
    scan.setPositionerCenteredLinearMode(0, getSecondCrystalPitchPosition(),
      0.002, 30);
    scan.setPositionerSettlingTime(0.2f);
    scan.setDetectorTriggerPvLink(0,
      Resources.getDefault().getBpm2IntensityAcquisitionControlPvName());
    scan.setDetectorTriggerCommandData(0, 1.0f);
    scan.setDetectorPvLink(0,
      Resources.getDefault().getBpm2IntensityAcquisitionSumPvName());
    scan.go();
    scan.waitToFinish();
  }

  protected double getSecondCrystalPitchPosition()
  throws InterruptedException, EpicsCaException {
    return secondCrystalPitchDrive.getPosition();
  }

  protected String getSecondCrystalPitchUnitOfMeasure() {
    return secondCrystalPitchDrive.getUnitOfMeasure();
  }

  protected void moveSecondCrystalPitch(double desiredPosition)
  throws InterruptedException, TimeoutException, EpicsCaException {
    secondCrystalPitchDrive.setDesiredPosition(desiredPosition);
    secondCrystalPitchDrive.moveAndWait();
  }
}
