package imcacat.imcabc;

import imcacat.epicsca4j.ca.EpicsCaException;
import imcacat.epicsca4j.convert.ShortToBooleanEpicsCaValueConverter;
import imcacat.epicsca4j.model.EpicsMotor;
import imcacat.epicsca4j.model.EpicsPv;
import imcacat.epicsca4j.model.TpmacAssemblyAxDrMo;
import imcacat.epicsca4j.util.StringUtilities;
import imcacat.epicsca4j.util.UiUtilities;

import java.lang.reflect.InvocationTargetException;
import java.text.DecimalFormat;

import javax.swing.JOptionPane;

import gov.aps.jca.TimeoutException;

/**
 * I manually optimize the DCM's Second Crystal Roll.
 */
public class SecondCrystalRollManualOptimization17id
extends DcmOptimization17id {
  protected static final int SECOND_XTAL_ROLL_BPM_RANGE = 4;

  protected TpmacAssemblyAxDrMo secondCrystalRollDrive;
  protected double originalSecondCrystalRollDrivePosition;
  protected EpicsMotor bpmStageHorizMotor;
  protected EpicsMotor bpmStageVertMotor;
  protected double originalBpmStageHorizMotorPosition;
  protected double originalBpmStageVertMotorPosition;
  protected EpicsPv bpmAutoRangePv;
  protected EpicsPv bpmAutoRangeRqsPv;
  protected boolean originalBpmAutoRange;
  protected EpicsPv bpmRangePv;
  protected EpicsPv bpmRangeRqsPv;
  protected int originalBpmRange;

  @Override
  protected void initialize()
  throws InterruptedException, EpicsCaException {
    super.initialize();
    secondCrystalRollDrive = new TpmacAssemblyAxDrMo(
      Resources.getDefault().getSecondCrystalTuningAssemblyPvName(),
      Resources.getDefault().getSecondCrystalRollDrivePvName(), 1.0,
      "mrad");
    originalSecondCrystalRollDrivePosition =
      secondCrystalRollDrive.getPosition();
    bpmStageHorizMotor = new EpicsMotor(
      Resources.getDefault().getBpm3StageHorizMotorPvName(), "\u03BCm",
      2);
    originalBpmStageHorizMotorPosition = bpmStageHorizMotor.getPosition();
    bpmStageVertMotor = new EpicsMotor(
      Resources.getDefault().getBpm3StageVertMotorPvName(), "\u03BCm",
      2);
    originalBpmStageVertMotorPosition = bpmStageVertMotor.getPosition();
    bpmAutoRangePv = new EpicsPv(
      Resources.getDefault().getBpm3AutoRangePvName(), false, false,
      new ShortToBooleanEpicsCaValueConverter((short)1, (short)0));
    bpmAutoRangeRqsPv = new EpicsPv(
      Resources.getDefault().getBpm3AutoRangeRqsPvName(), false, false,
      new ShortToBooleanEpicsCaValueConverter((short)1, (short)0));
    originalBpmAutoRange = (Boolean)bpmAutoRangePv.getValue();
    bpmRangePv = new EpicsPv(
      Resources.getDefault().getBpm3RangePvName(), false, false);
    bpmRangeRqsPv = new EpicsPv(
      Resources.getDefault().getBpm3RangeRqsPvName(), false, false);
    originalBpmRange = ((Number)bpmRangePv.getValue()).intValue();
  }

  @Override
  public void release() {
    if (secondCrystalRollDrive != null) secondCrystalRollDrive.release();
    if (bpmStageHorizMotor != null) bpmStageHorizMotor.release();
    if (bpmStageVertMotor != null) bpmStageVertMotor.release();
    if (bpmAutoRangePv != null) bpmAutoRangePv.release();
    if (bpmAutoRangeRqsPv != null) bpmAutoRangeRqsPv.release();
    if (bpmRangePv != null) bpmRangePv.release();
    if (bpmRangeRqsPv != null) bpmRangeRqsPv.release();
    super.release();
  }

  @Override
  protected void go()
  throws InterruptedException, TimeoutException, EpicsCaException {
    setStatus("Configuring beamline for optimization");
    disableAutobp();
    moveBpmStageToScanPosition();
    configureBpmRange();
    setProgress(10);

    setStatus("Scanning Second Crystal Roll");
    scanSecondCrystalRoll();
    setProgress(60);

    setStatus("Fitting scan data to Gaussian function");
    GaussianFunctionCurveFitter fitter = null;
    try {
      fitter = new GaussianFunctionCurveFitter(measuredXYSeries.toArray());
    } catch (Exception e) {}
    if (fitter != null) {
      try {
        setFitDataSeriesInEventDispatchThreadAndWait(fitter);
      } catch (InvocationTargetException e) {
        throw new RuntimeException("Failed to add fit data to " +
          "XYSeriesCollection", e);
      }
    }
    setProgress(70);

    double measuredMax = getXWithMaximumY(measuredXYSeries);
    double fitMax = 0.0;

    boolean shouldMoveToMeasuredMax = false;
    boolean shouldMoveToFitMax = false;

    DecimalFormat positionFormat = new DecimalFormat("0.000");
    if (fitter == null) {
      setStatus("Waiting for measured choice");
      int choice = JOptionPane.showOptionDialog(UiUtilities.getRootFrame(),
        StringUtilities.wordWrap("The Second Crystal Roll scan has " +
          "finished. However, the measured data could not be fit to a " +
          "Gaussian function. Would you like to move Second Crystal Roll " +
          "to the measured maximum at " +
          positionFormat.format(measuredMax) + " " +
          getSecondCrystalRollUnitOfMeasure() +
          "?",
          78, "\n"),
        "Move to measured maximum?", JOptionPane.YES_NO_OPTION,
        JOptionPane.QUESTION_MESSAGE, null, new String[] {"Yes", "Cancel"},
        "Yes");
      if (choice == JOptionPane.YES_OPTION) shouldMoveToMeasuredMax = true;
    } else {
      fitMax = fitter.getC();
      setStatus("Waiting for measured or fit choice");
      int choice = JOptionPane.showOptionDialog(UiUtilities.getRootFrame(),
        StringUtilities.wordWrap("The Second Crystal Roll scan has " +
          "finished. The measured maximum is at " +
          positionFormat.format(measuredMax) + " " +
          getSecondCrystalRollUnitOfMeasure() +
          ", and the fit maximum is at " +
          positionFormat.format(fitMax) + " " +
          getSecondCrystalRollUnitOfMeasure() +
          ". To which maximum would you like to move Second Crystal Roll?",
          78, "\n"),
        "Move to measured or fit?", JOptionPane.YES_NO_CANCEL_OPTION,
        JOptionPane.QUESTION_MESSAGE, null, new String[] {"Measured", "Fit",
        "Cancel"}, "Fit");
      if (choice == JOptionPane.YES_OPTION) shouldMoveToMeasuredMax = true;
      else if (choice == JOptionPane.NO_OPTION) shouldMoveToFitMax = true;
    }
    setProgress(75);
    if (shouldMoveToMeasuredMax) {
      setStatus("Moving to measured maximum");
      moveSecondCrystalRoll(measuredMax);
    } else if (shouldMoveToFitMax) {
      setStatus("Moving to fit maximum");
      moveSecondCrystalRoll(fitMax);
    } else {
      setStatus("Canceling");
      moveSecondCrystalRoll(originalSecondCrystalRollDrivePosition);
      moveBpmStageToOriginalPosition();
      restoreOriginalBpmRange();
      resetMonoAutobpAverageIntensity();
      enableAutobp();
      setStatus("Canceled");
      return;
    }
    setProgress(90);

    setStatus("Configuring beamline for use");
    moveBpmStageToOriginalPosition();
    restoreOriginalBpmRange();
    resetMonoAutobpAverageIntensity();
    enableAutobp();

    setProgressAndStatusToComplete();
  }

  @Override
  public void cleanUp() {
    Exception caughtException = null;

    try {
      scan.stop();
    } catch (Exception e) {
      if (caughtException == null) caughtException = e;
    }
    try {
      secondCrystalRollDrive.stopAndWait();
    } catch (Exception e) {
      if (caughtException == null) caughtException = e;
    }
    try {
      secondCrystalRollDrive.setDesiredPosition(
        originalSecondCrystalRollDrivePosition);
      secondCrystalRollDrive.moveAndWait();
    } catch (Exception e) {
      if (caughtException == null) caughtException = e;
    }
    try {
      bpmStageHorizMotor.stop();
    } catch (Exception e) {
      if (caughtException == null) caughtException = e;
    }
    try {
      bpmStageHorizMotor.setDesiredPosition(
        originalBpmStageHorizMotorPosition);
      bpmStageHorizMotor.moveAndWait();
    } catch (Exception e) {
      if (caughtException == null) caughtException = e;
    }
    try {
      bpmStageVertMotor.stop();
    } catch (Exception e) {
      if (caughtException == null) caughtException = e;
    }
    try {
      bpmStageVertMotor.setDesiredPosition(originalBpmStageVertMotorPosition);
      bpmStageVertMotor.moveAndWait();
    } catch (Exception e) {
      if (caughtException == null) caughtException = e;
    }
    try {
      if (originalBpmAutoRange) {
        bpmAutoRangeRqsPv.setValueAndWait(Boolean.TRUE);
      } else if (originalBpmRange != SECOND_XTAL_ROLL_BPM_RANGE) {
        bpmRangeRqsPv.setValueAndWait(originalBpmRange);
      }
    } catch (Exception e) {
      if (caughtException == null) caughtException = e;
    }
    try {
      resetMonoAutobpAverageIntensity();
    } catch (Exception e) {
      if (caughtException == null) caughtException = e;
    }
    try {
      enableAutobp();
    } catch (Exception e) {
      if (caughtException == null) caughtException = e;
    }

    try {
      super.cleanUp();
    } catch (Exception e) {
      if (caughtException == null) caughtException = e;
    }

    if (caughtException != null) throw new RuntimeException(caughtException);
  }

  protected void scanSecondCrystalRoll()
  throws EpicsCaException, InterruptedException {
    performSecondCrystalRollScan();
    if (!scan.completedSuccessfully())
      throw new RuntimeException(
        "Second Crystal Roll scan did not complete successfully");
    try {
      setMeasuredDataSeriesInEventDispatchThreadAndWait(getData(scan));
    } catch (InvocationTargetException e) {
      throw new RuntimeException("Failed to update measured data series " +
        "with complete data from scan", e);
    }
  }

  protected void performSecondCrystalRollScan() throws EpicsCaException,
  InterruptedException {
    setXYDatasetTitle("Second Crystal Roll optimization scan");
    setXDataLabel("Second Crystal Roll (" +
      getSecondCrystalRollUnitOfMeasure() + ")");
    setYDataLabel("Sample BPM intensity (counts)");
    secondCrystalRollDrive.setAutoStartEnabled(true);
    scanResetter.reset();
    scan.setBeforeScanPvLink(
      Resources.getDefault().getBpm3IntensityAcquisitionTimePvName());
    scan.setBeforeScanCommandData(1.0f);
    scan.setPositionerPvLink(0,
      Resources.getDefault().
        getDesiredSecondCrystalRollPositionPvName());
    scan.setPositionerReadbackPvLink(0,
      Resources.getDefault().getSecondCrystalRollPositionPvName());
    scan.setPositionerReadbackDelta(0, 0.03);
    scan.setPositionerCenteredLinearMode(0, getSecondCrystalRollPosition(),
      0.015, 40);
    scan.setPositionerSettlingTime(0.2f);
    scan.setDetectorTriggerPvLink(0,
      Resources.getDefault().getBpm3IntensityAcquisitionControlPvName());
    scan.setDetectorTriggerCommandData(0, 1.0f);
    scan.setDetectorPvLink(0,
      Resources.getDefault().getBpm3IntensityAcquisitionSumPvName());
    scan.go();
    scan.waitToFinish();
  }

  protected double getSecondCrystalRollPosition()
  throws InterruptedException, EpicsCaException {
    return secondCrystalRollDrive.getPosition();
  }

  protected String getSecondCrystalRollUnitOfMeasure() {
    return secondCrystalRollDrive.getUnitOfMeasure();
  }

  protected void moveSecondCrystalRoll(double desiredPosition)
  throws InterruptedException, TimeoutException, EpicsCaException {
    secondCrystalRollDrive.setDesiredPosition(desiredPosition);
    secondCrystalRollDrive.moveAndWait();
  }

  protected void moveBpmStageToScanPosition()
  throws InterruptedException, TimeoutException, EpicsCaException {
    moveBpmStageHoriz(
      originalBpmStageHorizMotorPosition +
        Resources.getDefault().
          getSecondCrystalRollOptimizationBpm3StageHorizOffset());
    moveBpmStageVert(
      originalBpmStageVertMotorPosition +
        Resources.getDefault().
          getSecondCrystalRollOptimizationBpm3StageVertOffset());
  }

  protected void moveBpmStageToOriginalPosition()
  throws InterruptedException, TimeoutException, EpicsCaException {
    moveBpmStageHoriz(originalBpmStageHorizMotorPosition);
    moveBpmStageVert(originalBpmStageVertMotorPosition);
  }

  protected void moveBpmStageHoriz(double desiredPosition)
  throws InterruptedException, TimeoutException, EpicsCaException {
    bpmStageHorizMotor.setDesiredPosition(desiredPosition);
    bpmStageHorizMotor.moveAndWait();
  }

  protected void moveBpmStageVert(double desiredPosition)
  throws InterruptedException, TimeoutException, EpicsCaException {
    bpmStageVertMotor.setDesiredPosition(desiredPosition);
    bpmStageVertMotor.moveAndWait();
  }

  protected void configureBpmRange()
  throws InterruptedException, EpicsCaException {
    bpmAutoRangeRqsPv.setValueAndWait(Boolean.FALSE);
    bpmRangeRqsPv.setValueAndWait(SECOND_XTAL_ROLL_BPM_RANGE);
  }

  protected void restoreOriginalBpmRange()
  throws InterruptedException, EpicsCaException {
    if (originalBpmAutoRange) {
      bpmAutoRangeRqsPv.setValueAndWait(Boolean.TRUE);
    } else if (originalBpmRange != SECOND_XTAL_ROLL_BPM_RANGE) {
      bpmRangeRqsPv.setValueAndWait(originalBpmRange);
    }
  }
}
