package imcacat.imcabc;

import imcacat.epicsca4j.ca.EpicsCaException;
import imcacat.epicsca4j.model.TpmacAssemblyAxDrMo;
import imcacat.imcabc.util.PhysicsUtilities;

/**
 * I am a tpmac Energy axis. I support reporting and setting my position in
 * eV and Angstroms.
 * <p>
 * I use {@link PhysicsUtilities#convertToEnergy} and
 * {@link PhysicsUtilities#convertToWavelength} to convert between Energy and
 * Wavelength.
 *
 * @see PhysicsUtilities
 */
public class TpmacEnergyAxis extends TpmacAssemblyAxDrMo {
  private static final long serialVersionUID = 1;

  protected static final String EV_UNIT_OF_MEASURE = "eV";

  public static final String PROPNAME_WAVELENGTH_POSITION =
    "wavelengthPosition";
  public static final String PROPNAME_WAVELENGTH_DESIRED_POSITION =
    "wavelengthDesiredPosition";

  protected double wavelengthPosition;
  protected double wavelengthDesiredPosition;

  public TpmacEnergyAxis(String assemblyName, String energyAxisName,
      double positionToEpicsScaleFactor)
      throws InterruptedException, EpicsCaException {
    super(assemblyName, energyAxisName, positionToEpicsScaleFactor,
      EV_UNIT_OF_MEASURE);
  }

  public TpmacEnergyAxis(String assemblyName, String energyAxisName,
      double positionToEpicsScaleFactor, int precision)
      throws InterruptedException, EpicsCaException {
    super(assemblyName, energyAxisName, positionToEpicsScaleFactor,
      EV_UNIT_OF_MEASURE, precision);
  }

  public TpmacEnergyAxis(String assemblyName, String energyAxisName,
      double positionToEpicsScaleFactor, int precision,
      Double desiredPositionLowLimit, Double desiredPositionHighLimit)
      throws InterruptedException, EpicsCaException {
    super(assemblyName, energyAxisName, positionToEpicsScaleFactor,
      EV_UNIT_OF_MEASURE, precision, desiredPositionLowLimit,
      desiredPositionHighLimit);
  }

  public TpmacEnergyAxis(String assemblyName, String energyAxisName,
      double positionToEpicsScaleFactor, int precision,
      Double desiredPositionLowLimit, Double desiredPositionHighLimit,
      String positionText)
      throws InterruptedException, EpicsCaException {
    super(assemblyName, energyAxisName, positionToEpicsScaleFactor,
      EV_UNIT_OF_MEASURE, precision, desiredPositionLowLimit,
      desiredPositionHighLimit, positionText);
  }

  /*
   * @return  position in Angstroms
   */
  public synchronized double getWavelengthPosition() {
    return wavelengthPosition;
  }

  protected void setWavelengthPosition(double newValue) {
    double oldValue = basicSetWavelengthPosition(newValue);
    if (oldValue != newValue)
      firePropertyChange(PROPNAME_WAVELENGTH_POSITION, oldValue, newValue);
  }

  protected synchronized double basicSetWavelengthPosition(double newValue) {
    double oldValue = wavelengthPosition;
    wavelengthPosition = newValue;
    return oldValue;
  }

  @Override
  protected void setPosition(double newValue) {
    super.setPosition(newValue);
    setWavelengthPosition(PhysicsUtilities.convertToWavelength(newValue));
  }

  @Override
  public void setDesiredPosition(double newValue) {
    super.setDesiredPosition(newValue);
    double newWavelength = PhysicsUtilities.convertToWavelength(newValue);
    double oldWavelength = basicSetWavelengthDesiredPosition(newWavelength);
    if (oldWavelength != newWavelength)
      firePropertyChange(PROPNAME_WAVELENGTH_DESIRED_POSITION, oldWavelength,
        newWavelength);
  }

  /**
   * @return  desired position in Angstroms
   */
  public synchronized double getWavelengthDesiredPosition() {
    return wavelengthDesiredPosition;
  }

  /**
   * @param newValue  desired position in Angstroms
   */
  public void setWavelengthDesiredPosition(double newValue) {
    double oldValue = basicSetWavelengthDesiredPosition(newValue);
    if (oldValue != newValue)
      firePropertyChange(PROPNAME_WAVELENGTH_DESIRED_POSITION, oldValue,
        newValue);
    super.setDesiredPosition(PhysicsUtilities.convertToEnergy(newValue));
  }

  protected synchronized double basicSetWavelengthDesiredPosition(
      double newValue) {
    double oldValue = wavelengthDesiredPosition;
    wavelengthDesiredPosition = newValue;
    return oldValue;
  }
}
