package imcacat.imcabc;

import imcacat.epicsca4j.util.ObjectUtilities;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import javax.swing.AbstractAction;
import javax.swing.Action;

/**
 * I am a two state button <code>Action</code>.
 * <p>
 * Say, for example, there is a GUI for moving a motor with a button that says
 * "Start". When you click the button, its label changes to "Stop". If you
 * click the button again while it says "Stop", the motor will stop and its
 * label will change to "Start" again. This two state button behavior is what
 * I provide.
 * <p>
 * Normally, I compare (using <tt>equals</tt>) the read state value with my
 * state 1 and state 2 values to determine which state I'm in. However, if
 * my state 1 or state 2 value is a {@link Collection}, I will consider the
 * read state value to match my state 1 or state 2 value if my state 1 or
 * state 2 value that is a <tt>Collection</tt> contains (using
 * <tt>contains</tt>) the read state value.
 * <p>
 * Subclasses must call <code>initialize</code> in their constructor.
 */
public abstract class TwoStateButtonAction extends AbstractAction
implements PropertyChangeListener {
  private static final long serialVersionUID = 1;

  protected static final int STATE_NONE = 0;
  protected static final int STATE_1 = 1;
  protected static final int STATE_2 = 2;

  protected static final String STATE_NONE_COMMAND = "0";
  protected static final String STATE_1_COMMAND = "1";
  protected static final String STATE_2_COMMAND = "2";

  public static final String PROPNAME_ACTION_PERFORMED_STATE =
    "actionPerformedState";
  public static final String PROPNAME_ACTION_COMMAND =
    "actionCommand";

  protected String state1Label;
  protected String state2Label;
  protected Object state1Value;
  protected boolean state1ValueIsCollection;
  protected Object state2Value;
  protected boolean state2ValueIsCollection;

  protected int actionPerformedState;
  protected String mostRecentCommandPerformed;
  protected boolean shouldEnforceOnePressPerState;

  protected List<ActionListener> actionListeners =
    new ArrayList<ActionListener>();

  public TwoStateButtonAction(String state1Label, String state2Label) {
    this(state1Label, state2Label, Boolean.FALSE, Boolean.TRUE, true);
  }

  public TwoStateButtonAction(String state1Label, String state2Label,
      Object state1Value, Object state2Value,
      boolean shouldEnforceOnePressPerState) {
    super();
    this.state1Label = state1Label;
    this.state2Label = state2Label;
    this.state1Value = state1Value;
    state1ValueIsCollection = state1Value instanceof Collection;
    this.state2Value = state2Value;
    state2ValueIsCollection = state2Value instanceof Collection;
    this.shouldEnforceOnePressPerState = shouldEnforceOnePressPerState;
    mostRecentCommandPerformed = STATE_NONE_COMMAND;
  }

  protected void initialize() {
    updateFromStateValue(getStateValue());
    registerStateValueListener();
  }

  protected void finalize() throws Throwable {
    unregisterStateValueListener();
    super.finalize();
  }

  public void actionPerformed(ActionEvent e) {
    String command = e.getActionCommand();

    if (shouldEnforceOnePressPerState &&
        mostRecentCommandPerformed.equals(command)) return;

    fireActionPerformed(e);

    String oldCommand = mostRecentCommandPerformed;
    mostRecentCommandPerformed = command;

    if (STATE_1_COMMAND.equals(command)) {
      performState1Action();
    } else if (STATE_2_COMMAND.equals(command)) {
      performState2Action();
    } else {
      throw new IllegalArgumentException("Unexpected Action command: " +
        command);
    }

    if (!ObjectUtilities.areEqual(oldCommand, command))
      firePropertyChange(PROPNAME_ACTION_COMMAND, oldCommand, command);
  }

  public void propertyChange(PropertyChangeEvent e) {
    updateFromStateValue(e.getNewValue());
  }

  public void updateFromStateValue(Object stateValue) {
    if (isState1ValueMatch(stateValue)) {
      mostRecentCommandPerformed = STATE_2_COMMAND;
      changeToState1();
    } else if (isState2ValueMatch(stateValue)) {
      mostRecentCommandPerformed = STATE_1_COMMAND;
      changeToState2();
    }
  }

  protected boolean isState1ValueMatch(Object stateValue) {
    if (stateValue == null) return false;
    if (state1ValueIsCollection)
      return ((Collection<?>)state1Value).contains(stateValue);
    else
      return state1Value.equals(stateValue);
  }

  protected boolean isState2ValueMatch(Object stateValue) {
    if (stateValue == null) return false;
    if (state2ValueIsCollection)
      return ((Collection<?>)state2Value).contains(stateValue);
    else
      return state2Value.equals(stateValue);
  }

  /**
   * Performs initialization related to being notified when the state that I
   * model has changed. Called at instantiation time. Currently does nothing,
   * but may be overridden in a subclass to do something useful like add a
   * property change listener.
   */
  public void registerStateValueListener() {
  }

  /**
   * Performs cleanup tasks related to being notified when the state that I
   * model has changed. Called at object destruction time from the finalize
   * method. Currently does nothing, but may be overridden in a subclass to do
   * something useful like remove a property change listener.
   */
  public void unregisterStateValueListener() {
  }

  /**
   * Gets state value of the current state. Subclasses may override to provide
   * different behavior.
   */
  public Object getStateValue() {
    return mostRecentCommandPerformed.equals(STATE_2_COMMAND) ?
      state2Value : state1Value;
  }

  /**
   * Performs state 1 action. Subclasses may override to provide different
   * behavior.
   */
  public void performState1Action() {
    setActionPerformedState(STATE_1);
  }

  /**
   * Performs state 2 action. Subclasses may override to provide different
   * behavior.
   */
  public void performState2Action() {
    setActionPerformedState(STATE_2);
  }

  /**
   * Called to change to state 1. Subclasses may override to provide added
   * behavior.
   */
  public void changeToState1() {
    putValue(Action.ACTION_COMMAND_KEY, STATE_1_COMMAND);
    putValue(Action.NAME, state1Label);
  }

  /**
   * Called to change to state 2. Subclasses may override to provide added
   * behavior.
   */
  public void changeToState2() {
    putValue(Action.ACTION_COMMAND_KEY, STATE_2_COMMAND);
    putValue(Action.NAME, state2Label);
  }

  public int getActionPerformedState() {
    return actionPerformedState;
  }

  protected void setActionPerformedState(int newValue) {
    int oldValue = actionPerformedState;
    actionPerformedState = newValue;
    if (oldValue != newValue)
      firePropertyChange(PROPNAME_ACTION_PERFORMED_STATE, oldValue, newValue);
  }

  protected synchronized void fireActionPerformed(ActionEvent event) {
    for (ActionListener each : actionListeners)
      each.actionPerformed(event);
  }

  protected synchronized void addActionListener(ActionListener listener) {
    actionListeners.add(listener);
  }

  protected synchronized void removeActionListener(ActionListener listener) {
    actionListeners.remove(listener);
  }
}
