package imcacat.imcabc.util;

import org.apache.commons.math.FunctionEvaluationException;
import org.apache.commons.math.MathRuntimeException;
import org.apache.commons.math.analysis.UnivariateRealFunction;

/**
 * I am the derivative of {@link GaussianFunction}.  Specifically:
 * <p>
 * <tt>f'(x) = (-b / (d^2)) * (x - c) * exp(-((x - c)^2) / (2*(d^2)))</tt>
 */
public class GaussianDerivativeFunction implements UnivariateRealFunction {
  protected double b;
  protected double c;
  protected double d;

  /**
   * Constructs an instance of me with the specified parameters.
   *
   * @param b <tt>b</tt> parameter value
   * @param c <tt>c</tt> parameter value
   * @param d <tt>d</tt> parameter value
   *
   * @throws IllegalArgumentException if <code>d</code> is 0
   */
  public GaussianDerivativeFunction(double b, double c, double d) {
    if (d == 0.0)
      throw MathRuntimeException.createIllegalArgumentException(
        "parameter d == {0} results in division by zero", 0);
    this.b = b;
    this.c = c;
    this.d = d;
  }

  /**
   * Constructs an instance of me with the specified parameters.
   *
   * @param parameters <tt>b</tt>, <tt>c</tt>, and <tt>d</tt> parameter values
   *
   * @throws IllegalArgumentException if <code>parameters</code> is null,
   *         <code>parameters</code> length is not 3, or if
   *         <code>parameters[2]</code> is 0
   */
  public GaussianDerivativeFunction(double[] parameters) {
    if (parameters == null)
      throw MathRuntimeException.createIllegalArgumentException(
        "parameters array == {0}", (Object)null);
    if (parameters.length != 3)
      throw MathRuntimeException.createIllegalArgumentException(
        "parameters array length != {0}", 3);
    if (parameters[2] == 0.0)
      throw MathRuntimeException.createIllegalArgumentException(
        "parameter[{0}] == {1} results in division by zero", 2, 0);
    this.b = parameters[0];
    this.c = parameters[1];
    this.d = parameters[2];
  }

  public double value(double x) throws FunctionEvaluationException {
    return (-b / (d * d)) * (x - c) *
      Math.exp(-((x - c) * (x - c)) / (2.0 * (d * d)));
  }
}
