package imcacat.imcabc.util;

import org.apache.commons.math.FunctionEvaluationException;
import org.apache.commons.math.optimization.
       DifferentiableMultivariateVectorialOptimizer;
import org.apache.commons.math.optimization.OptimizationException;
import org.apache.commons.math.optimization.fitting.CurveFitter;
import org.apache.commons.math.optimization.fitting.WeightedObservedPoint;

/**
 * I fit my points to a Gaussian function.
 *
 * @see ParametricGaussianFunction
 */
public class GaussianFitter {
  protected CurveFitter fitter;

  /**
   * Constructs an instance of me using the specified optimizer.
   *
   * @param optimizer optimizer to use for the fitting
   */
  public GaussianFitter(
      DifferentiableMultivariateVectorialOptimizer optimizer) {
    this.fitter = new CurveFitter(optimizer);
  }

  /**
   * Adds point (<code>x</code>, <code>y</code>) to my list of observed points
   * with a weight of 1.0.
   *
   * @param x <tt>x</tt> point value
   * @param y <tt>y</tt> point value
   */
  public void addObservedPoint(double x, double y) {
    addObservedPoint(1.0, x, y);
  }

  /**
   * Adds point (<code>x</code>, <code>y</code>) to my list of observed points
   * with a weight of <code>weight</code>.
   *
   * @param weight weight assigned to point
   * @param x <tt>x</tt> point value
   * @param y <tt>y</tt> point value
   */
  public void addObservedPoint(double weight, double x, double y) {
    fitter.addObservedPoint(weight, x, y);
  }

  /**
   * Fits Gaussian function to the observed points.
   *
   * @return Gaussian function best fitting the observed points
   *
   * @throws FunctionEvaluationException if <code>CurveFitter.fit</code>
   *         throws it
   * @throws OptimizationException if <code>CurveFitter.fit</code> throws it
   * @throws IllegalArgumentException if <code>CurveFitter.fit</code> throws
   *         it
   *
   * @see CurveFitter
   */
  public GaussianFunction fit()
  throws FunctionEvaluationException, OptimizationException {
    return new GaussianFunction(fitter.fit(new ParametricGaussianFunction(),
      createParametersGuesser(fitter.getObservations()).guess()));
  }

  protected GaussianParametersGuesser createParametersGuesser(
      WeightedObservedPoint[] observations) {
    return new GaussianParametersGuesser(observations);
  }
}
