package imcacat.imcabc.util;

import org.apache.commons.math.FunctionEvaluationException;
import org.apache.commons.math.MathRuntimeException;
import org.apache.commons.math.analysis.DifferentiableUnivariateRealFunction;
import org.apache.commons.math.analysis.UnivariateRealFunction;

/**
 * I am a Gaussian function.  Specifically:
 * <p>
 * <tt>f(x) = a + b*exp(-((x - c)^2 / (2*d^2)))</tt>
 *
 * @see GaussianDerivativeFunction
 * @see ParametricGaussianFunction
 */
public class GaussianFunction
implements DifferentiableUnivariateRealFunction {
  protected double a;
  protected double b;
  protected double c;
  protected double d;

  /**
   * Constructs an instance of me with the specified parameters.
   *
   * @param a <tt>a</tt> parameter value
   * @param b <tt>b</tt> parameter value
   * @param c <tt>c</tt> parameter value
   * @param d <tt>d</tt> parameter value
   *
   * @throws IllegalArgumentException if <code>d</code> is 0
   */
  public GaussianFunction(double a, double b, double c, double d) {
    if (d == 0.0)
      throw MathRuntimeException.createIllegalArgumentException(
        "parameter d == {0} results in division by zero", 0);
    this.a = a;
    this.b = b;
    this.c = c;
    this.d = d;
  }

  /**
   * Constructs an instance of me with the specified parameters.
   *
   * @param parameters <tt>a</tt>, <tt>b</tt>, <tt>c</tt>, and <tt>d</tt>
   *        parameter values
   *
   * @throws IllegalArgumentException if <code>parameters</code> is null,
   *         <code>parameters</code> length is not 4, or if
   *         <code>parameters[3]</code> is 0
   */
  public GaussianFunction(double[] parameters) {
    if (parameters == null)
      throw MathRuntimeException.createIllegalArgumentException(
        "parameters array == {0}", (Object)null);
    if (parameters.length != 4)
      throw MathRuntimeException.createIllegalArgumentException(
        "parameters array length != {0}", 4);
    if (parameters[3] == 0.0)
      throw MathRuntimeException.createIllegalArgumentException(
        "parameter[{0}] == {1} results in division by zero", 3, 0);
    this.a = parameters[0];
    this.b = parameters[1];
    this.c = parameters[2];
    this.d = parameters[3];
  }

  public UnivariateRealFunction derivative() {
    return new GaussianDerivativeFunction(b, c, d);
  }

  public double value(double x) throws FunctionEvaluationException {
    return a + b * Math.exp(-((x - c) * (x - c)) / (2.0 * (d * d)));
  }

  /**
   * Gets <tt>a</tt> parameter value.
   *
   * @return <tt>a</tt> parameter value
   */
  public double getA() {
    return a;
  }

  /**
   * Gets <tt>b</tt> parameter value.
   *
   * @return <tt>b</tt> parameter value
   */
  public double getB() {
    return b;
  }

  /**
   * Gets <tt>c</tt> parameter value.
   *
   * @return <tt>c</tt> parameter value
   */
  public double getC() {
    return c;
  }

  /**
   * Gets <tt>d</tt> parameter value.
   *
   * @return <tt>d</tt> parameter value
   */
  public double getD() {
    return d;
  }
}
