package imcacat.imcabc.util;

import java.util.Arrays;
import java.util.Comparator;

import org.apache.commons.math.MathException;
import org.apache.commons.math.MathRuntimeException;
import org.apache.commons.math.optimization.fitting.WeightedObservedPoint;

/**
 * I guess the parameters (<tt>a</tt>, <tt>b</tt>, <tt>c</tt>, and <tt>d</tt>)
 * of {@link ParametricGaussianFunction} based on the specified observed
 * points.
 */
public class GaussianParametersGuesser {
  protected WeightedObservedPoint[] observations;
  protected double[] parameters;

  /**
   * Constructs instance of me with the specified observed points.
   *
   * @param observations observed points upon which I should base my guess
   */
  public GaussianParametersGuesser(WeightedObservedPoint[] observations) {
    if (observations == null)
      throw MathRuntimeException.createIllegalArgumentException(
        "observations array == {0}", (Object)null);
    if (observations.length < 3)
      throw MathRuntimeException.createIllegalArgumentException(
        "observations array length < {0}", 3);
    this.observations = observations.clone();
  }

  /**
   * Guesses the parameters based on the observed points.
   *
   * @return guessed parameters array <code>{a, b, c, d}</code>
   */
  public double[] guess() {
    if (parameters == null) parameters = basicGuess(observations);
    return parameters;
  }

  protected double[] basicGuess(WeightedObservedPoint[] points) {
    Arrays.sort(points, createWeightedObservedPointComparator());
    double[] params = new double[4];

    int minYIdx = findMinY(points);
    params[0] = points[minYIdx].getY();

    int maxYIdx = findMaxY(points);
    params[1] = points[maxYIdx].getY();
    params[2] = points[maxYIdx].getX();

    double fwhmApprox;
    try {
      double halfY = params[0] + ((params[1] - params[0]) / 2.0);
      double fwhmX1 = interpolateXAtY(points, maxYIdx, -1, halfY);
      double fwhmX2 = interpolateXAtY(points, maxYIdx, +1, halfY);
      fwhmApprox = fwhmX2 - fwhmX1;
    } catch (MathException e) {
      fwhmApprox = points[points.length - 1].getX() - points[0].getX();
    }
    params[3] = fwhmApprox / (2.0 * Math.sqrt(2.0 * Math.log(2.0)));

    return params;
  }

  protected int findMinY(WeightedObservedPoint[] points) {
    int minYIdx = 0;
    for (int i = 1; i < points.length; i++)
      if (points[i].getY() < points[minYIdx].getY()) minYIdx = i;
    return minYIdx;
  }

  protected int findMaxY(WeightedObservedPoint[] points) {
    int maxYIdx = 0;
    for (int i = 1; i < points.length; i++)
      if (points[i].getY() > points[maxYIdx].getY()) maxYIdx = i;
    return maxYIdx;
  }

  protected double interpolateXAtY(WeightedObservedPoint[] points,
      int startIdx, int idxStep, double y)
  throws MathException {
    if (idxStep == 0)
      throw MathRuntimeException.createIllegalArgumentException(
        "idxStep == {0}; should be positive or negative", 0);
    WeightedObservedPoint[] twoPoints = getInterpolationPointsForY(points,
      startIdx, idxStep, y);
    WeightedObservedPoint pointA = twoPoints[0];
    WeightedObservedPoint pointB = twoPoints[1];
    if (pointA.getY() == y) return pointA.getX();
    if (pointB.getY() == y) return pointB.getX();
    return pointA.getX() +
      (((y - pointA.getY()) * (pointB.getX() - pointA.getX())) /
        (pointB.getY() - pointA.getY()));
  }

  protected WeightedObservedPoint[] getInterpolationPointsForY(
      WeightedObservedPoint[] points, int startIdx, int idxStep, double y)
  throws MathException {
    if (idxStep == 0)
      throw MathRuntimeException.createIllegalArgumentException(
        "idxStep == {0}; should be positive or negative", 0);
    for (int i = startIdx;
         idxStep < 0 ? i + idxStep >= 0 : i + idxStep < points.length;
         i += idxStep)
      if (isBetween(y, points[i].getY(), points[i+idxStep].getY()))
        return idxStep < 0
          ? new WeightedObservedPoint[] {points[i+idxStep], points[i]}
          : new WeightedObservedPoint[] {points[i], points[i+idxStep]};
    throw new MathException("y ({0}) not within observed points", y);
  }

  protected boolean isBetween(double value, double boundary1,
      double boundary2) {
    return (value >= boundary1 && value <= boundary2) ||
      (value >= boundary2 && value <= boundary1);
  }

  protected Comparator<WeightedObservedPoint>
  createWeightedObservedPointComparator() {
    return new Comparator<WeightedObservedPoint>() {
      public int compare(WeightedObservedPoint p1, WeightedObservedPoint p2) {
        if (p1 == null && p2 == null) return 0;
        if (p1 == null) return -1;
        if (p2 == null) return 1;
        if (p1.getX() < p2.getX()) return -1;
        if (p1.getX() > p2.getX()) return 1;
        if (p1.getY() < p2.getY()) return -1;
        if (p1.getY() > p2.getY()) return 1;
        if (p1.getWeight() < p2.getWeight()) return -1;
        if (p1.getWeight() > p2.getWeight()) return 1;
        return 0;
      }
    };
  }
}
