package imcacat.imcabc.util;

import org.apache.commons.math.FunctionEvaluationException;
import org.apache.commons.math.MathRuntimeException;
import org.apache.commons.math.optimization.fitting.ParametricRealFunction;

/**
 * I am a Gaussian function.  Specifically:
 * <p>
 * <tt>f(x) = a + b*exp(-((x - c)^2 / (2*d^2)))</tt>
 * <p>
 * My parameters have the following meaning:
 * <ul>
 * <li><tt>a</tt> is a constant offset that shifts <tt>f(x)</tt> up or down
 * <li><tt>b</tt> is the height of my peak
 * <li><tt>c</tt> is the position of the center of my peak
 * <li><tt>d</tt> is related to my FWHM by <tt>FWHM = 2*sqrt(2*ln(2))*d</tt>
 * </ul>
 * Notation key:
 * <ul>
 * <li><tt>x^n</tt>: <tt>x</tt> raised to the power of <tt>n</tt>
 * <li><tt>x*y</tt>: <tt>x</tt> multiplied by <tt>y</tt>
 * <li><tt>exp(x)</tt>: <i>e</i><tt>^x</tt>
 * <li><tt>sqrt(x)</tt>: the square root of <tt>x</tt>
 * <li><tt>ln(x)</tt>: the natural logarithm of <tt>x</tt>
 * </ul>
 * References:
 * <ul>
 * <li><a href="http://en.wikipedia.org/wiki/Gaussian_function">Wikipedia:
 *   Gaussian function</a>
 * </ul>
 */
public class ParametricGaussianFunction implements ParametricRealFunction {
  /**
   * Constructs an instance of me.
   */
  public ParametricGaussianFunction() {
    super();
  }

  /**
   * Computes value of my function <tt>f(x)</tt> for the specified <tt>x</tt>
   * and parameters <tt>a</tt>, <tt>b</tt>, <tt>c</tt>, and <tt>d</tt>.
   *
   * @param x <tt>x</tt> value
   * @param parameters values of <tt>a</tt>, <tt>b</tt>, <tt>c</tt>, and
   *        <tt>d</tt>
   *
   * @return value of <tt>f(x)</tt> evaluated at <tt>x</tt> with the specified
   *         parameters
   *
   * @throws IllegalArgumentException if <code>parameters</code> is invalid as
   *         determined by {@link #validateParameters(double[])}
   * @throws FunctionEvaluationException if <code>parameters</code> values are
   *         invalid as determined by {@link #validateParameters(double[])}
   */
  public double value(double x, double[] parameters)
      throws FunctionEvaluationException {
    validateParameters(parameters);
    double a = parameters[0];
    double b = parameters[1];
    double c = parameters[2];
    double d = parameters[3];
    return a + b * Math.exp(-((x - c) * (x - c)) / (2.0 * (d * d)));
  }

  /**
   * Computes the gradient vector for a four variable version of my function
   * where my parameters, <tt>a</tt>, <tt>b</tt>, <tt>c</tt>, and <tt>d</tt>,
   * are considered the variables, not <tt>x</tt>.  That is, instead of
   * computing the gradient vector for my function <tt>f(x)</tt> (which would
   * just be the derivative of <tt>f(x)</tt> with respect to <tt>x</tt> since
   * it's a one-dimensional function), computes the gradient vector for the
   * function <tt>f(a, b, c, d) = a + b*exp(-((x - c)^2 / (2*d^2)))</tt>
   * treating the specified <tt>x</tt> as a constant.
   * <p>
   * The components of the computed gradient vector are the partial
   * derivatives of <tt>f(a, b, c, d)</tt> with respect to each variable.
   * That is, the partial derivative of <tt>f(a, b, c, d)</tt> with respect to
   * <tt>a</tt>, the partial derivative of <tt>f(a, b, c, d)</tt> with respect
   * to <tt>b</tt>, the partial derivative of <tt>f(a, b, c, d)</tt> with
   * respect to <tt>c</tt>, and the partial derivative of <tt>f(a, b, c,
   * d)</tt> with respect to <tt>d</tt>.
   *
   * @param x <tt>x</tt> value to be used as constant in <tt>f(a, b, c,
   *        d)</tt>
   * @param parameters values of <tt>a</tt>, <tt>b</tt>, <tt>c</tt>, and
   *        <tt>d</tt> for computation of gradient vector of <tt>f(a, b, c,
   *        d)</tt>
   *
   * @return gradient vector of <tt>f(a, b, c, d)</tt>
   *
   * @throws IllegalArgumentException if <code>parameters</code> is invalid as
   *         determined by {@link #validateParameters(double[])}
   * @throws FunctionEvaluationException if <code>parameters</code> values are
   *         invalid as determined by {@link #validateParameters(double[])}
   */
  public double[] gradient(double x, double[] parameters)
  throws FunctionEvaluationException {
    validateParameters(parameters);
    double b = parameters[1];
    double c = parameters[2];
    double d = parameters[3];
    return new double[] {
      1.0,
      Math.exp(-((x - c) * (x - c)) / (2.0 * (d * d))),
      b * Math.exp(-((x - c) * (x - c)) / (2.0 * (d * d))) *
        (1.0 / (d * d)) * (x - c),
      (b * Math.exp(-((x - c) * (x - c)) / (2.0 * (d * d))) *
        ((x - c) * (x - c))) / (d * d * d)
    };
  }

  /**
   * Validates parameters to ensure they are appropriate for the evaluation of
   * my <code>value</code> and <code>gradient</code> methods.
   *
   * @param parameters values of <tt>a</tt>, <tt>b</tt>, <tt>c</tt>, and
   *        <tt>d</tt>
   *
   * @throws IllegalArgumentException if <code>parameters</code> is
   *         <code>null</code> or if <code>parameters</code> does not have
   *         length == 4
   * @throws FunctionEvaluationException if <code>parameters[3]</code>
   *         (<tt>d</tt>) is 0
   */
  protected void validateParameters(double[] parameters)
  throws FunctionEvaluationException {
    if (parameters == null)
      throw MathRuntimeException.createIllegalArgumentException(
        "parameters == {0}", (Object)null);
    if (parameters.length != 4)
      throw MathRuntimeException.createIllegalArgumentException(
        "parameters.length != {0}", 4);
    if (parameters[3] == 0.0)
      throw new FunctionEvaluationException(parameters,
        "parameters[{0}] == {1} results in division by zero", 3, 0);
  }
}
