package imcacat.imcabc.util;

/**
 * I provide utility methods related to physics.
 * <p>
 * The constants I use for converting between eV and Angstroms are from
 * <a href="http://physics.nist.gov/cuu/Constants/">2010 CODATA recommended
 * values</a>.
 */
public class PhysicsUtilities {
  /* Planck constant (eV s) */
  protected static final double H = 4.135667516e-15;
  /* Speed of light in vacuum (m/s) */
  protected static final double C = 2.99792458e8;
  /* Angstrom-meter relationship (angstroms/m) */
  protected static final double ANGSTROMS_IN_METER = 1e10;

  private PhysicsUtilities() {}

  /**
   * Converts wavelength in angstroms to energy in eV.
   * 
   * @throws ArithmeticException if wavelength is 0
   */
  public static double convertToEnergy(double wavelength) {
    /*
     * The Planck relation equation relates energy (E) and frequency (v); h
     * is the Planck constant:
     * 
     * E = hv
     * 
     * The wave equation relates frequency (v) and wavelength (L); c is the
     * speed of light in a vacuum:
     * 
     * Lv = c
     * 
     * Substituting for frequency (v) in the Planck relation equation gives:
     * 
     * E = h(c/L)
     */
    return H * C * ANGSTROMS_IN_METER / wavelength;
  }

  /**
   * Converts wavelength in angstroms to energy in eV.
   * 
   * @throws ArithmeticException if wavelength is 0
   */
  public static Double convertToEnergy(Double wavelength) {
    /*
     * See the comment in convertToEnergy(double) for an explanation of the
     * equation for this conversion.
     */
    if (wavelength == null) return null;
    return H * C * ANGSTROMS_IN_METER / wavelength;
  }

  /**
   * Converts energy in eV to wavelength in angstroms.
   * 
   * @throws ArithmeticException if energy is 0
   */
  public static double convertToWavelength(double energy) {
    /*
     * See the comment in convertToEnergy(double) for an explanation of the
     * equation for this conversion.
     */
    return H * C * ANGSTROMS_IN_METER / energy;
  }

  /**
   * Converts energy in eV to wavelength in angstroms.
   * 
   * @throws ArithmeticException if energy is 0
   */
  public static Double convertToWavelength(Double energy) {
    /*
     * See the comment in convertToEnergy(double) for an explanation of the
     * equation for this conversion.
     */
    if (energy == null) return null;
    return H * C * ANGSTROMS_IN_METER / energy;
  }
}
