package imcacat.imcabc.util;

import static org.junit.Assert.assertEquals;

import com.cosylab.epics.caj.CAJContext;

import imcacat.epicsca4j.ca.CaGetListener;
import imcacat.epicsca4j.ca.CaPutListener;
import imcacat.epicsca4j.convert.BasicEpicsChannelAccessValueConverter;
import imcacat.epicsca4j.convert.EpicsCaValueConverter;
import imcacat.epicsca4j.event.QueuedEventDispatcher;
import imcacat.epicsca4j.util.EpicsCaUtilities;
import gov.aps.jca.CAException;
import gov.aps.jca.CAStatus;
import gov.aps.jca.Channel;
import gov.aps.jca.Context;
import gov.aps.jca.JCALibrary;
import gov.aps.jca.TimeoutException;
import gov.aps.jca.configuration.Configuration;
import gov.aps.jca.configuration.DefaultConfiguration;
import gov.aps.jca.dbr.DBR;

/**
 * I provide utility methods for unit tests.
 */
public class TestUtilities {
  protected static final double CA_PEND_IO_TIMEOUT = 5.0;
  protected static final long CA_WAIT_TIMEOUT = 5000;
  protected static Context CONTEXT;
  protected static Configuration JCA_CONFIGURATION;
  protected static final EpicsCaValueConverter EPICS_CA_VALUE_CONVERTER =
    new BasicEpicsChannelAccessValueConverter();

  private TestUtilities() {}

  public static Object caget(String pvName)
  throws CAException, TimeoutException {
    return caget(pvName, false);
  }

  public static Object caget(String pvName, boolean isArray)
  throws CAException, TimeoutException {
    Channel channel = null;
    try {
      channel = createCaChannel(pvName);
      DBR dbrValue = channel.get();
      channel.getContext().pendIO(TestUtilities.CA_PEND_IO_TIMEOUT);
      return EPICS_CA_VALUE_CONVERTER.convertToJava(dbrValue, isArray);
    } finally {
      if (channel != null) channel.destroy();
    }
  }

  public static Object cagetAndWait(String pvName)
  throws CAException, TimeoutException, InterruptedException {
    return cagetAndWait(pvName, false);
  }

  public static Object cagetAndWait(String pvName, boolean isArray)
  throws CAException, TimeoutException, InterruptedException {
    Channel channel = null;
    try {
      channel = createCaChannel(pvName);
      CaGetListener listener = new CaGetListener();
      channel.get(listener);
      synchronized (listener) {
        channel.getContext().flushIO();
        long timeout = millisecondsToNanoseconds(CA_WAIT_TIMEOUT);
        long startTime = System.nanoTime();
        while (listener.getStatus() == null) {
          listener.wait(CA_WAIT_TIMEOUT);
          if ((System.nanoTime() - startTime) > timeout)
            throw new TimeoutException(pvName);
        }
      }
      assertEquals(CAStatus.NORMAL, listener.getStatus());
      return EPICS_CA_VALUE_CONVERTER.convertToJava(listener.getValue(),
        isArray);
    } finally {
      if (channel != null) channel.destroy();
    }
  }

  public static void caput(String pvName, Object value)
  throws CAException, TimeoutException {
    Channel channel = null;
    try {
      channel = createCaChannel(pvName);
      EpicsCaUtilities.put(channel,
        EPICS_CA_VALUE_CONVERTER.convertToChannelAccess(value));
      channel.getContext().pendIO(TestUtilities.CA_PEND_IO_TIMEOUT);
    } finally {
      if (channel != null) channel.destroy();
    }
  }

  public static void caputAndWait(String pvName, Object value)
  throws CAException, TimeoutException, InterruptedException {
    Channel channel = null;
    try {
      channel = createCaChannel(pvName);
      CaPutListener listener = new CaPutListener();
      EpicsCaUtilities.put(channel,
        EPICS_CA_VALUE_CONVERTER.convertToChannelAccess(value), listener);
      synchronized (listener) {
        channel.getContext().flushIO();
        long timeout = millisecondsToNanoseconds(CA_WAIT_TIMEOUT);
        long startTime = System.nanoTime();
        while (listener.getStatus() == null) {
          listener.wait(CA_WAIT_TIMEOUT);
          if ((System.nanoTime() - startTime) > timeout)
            throw new TimeoutException(pvName);
        }
      }
      assertEquals(CAStatus.NORMAL, listener.getStatus());
    } finally {
      if (channel != null) channel.destroy();
    }
  }

  public static Configuration createJcaConfiguration() {
    DefaultConfiguration conf = new DefaultConfiguration("imcabc-test");
    imcacat.epicsca4j.util.Resources.setJcaConfigurationPropertiesFromEnv(
      conf);
    conf.setAttribute("class", CAJContext.class.getCanonicalName());
    DefaultConfiguration eventDispatcherConfiguration =
      new DefaultConfiguration("event_dispatcher");
    eventDispatcherConfiguration.setAttribute("class",
      QueuedEventDispatcher.class.getCanonicalName());
    conf.addChild(eventDispatcherConfiguration);
    return conf;
  }

  protected static Channel createCaChannel(String pvName) throws CAException {
    Channel channel = null;
    try {
      channel = getCaContext().createChannel(pvName);
      getCaContext().pendIO(TestUtilities.CA_PEND_IO_TIMEOUT);
      assert(Channel.CONNECTED.equals(channel.getConnectionState()));
      return channel;
    } catch (Exception e) {
      try { channel.destroy(); } catch (Exception e2) {}
      throw new RuntimeException(e);
    }
  }

  protected synchronized static Context getCaContext() throws CAException {
    if (CONTEXT == null)
      CONTEXT = JCALibrary.getInstance().createContext(getJcaConfiguration());
    return CONTEXT;
  }

  protected synchronized static Configuration getJcaConfiguration() {
    if (JCA_CONFIGURATION == null)
      JCA_CONFIGURATION = createJcaConfiguration();
    return JCA_CONFIGURATION;
  }

  protected static long millisecondsToNanoseconds(long milliseconds) {
    return milliseconds * 1000000L;
  }
}
